(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Equational_Reasoning 
imports
  Trs
  Option_Monad
begin

text {*
The following datatype represents proofs in equational logic
using the following rules:
\begin{itemize}
  \item reflexivity: $E \vdash t = t$
  \item symmetry: $E \vdash s = t$ implies $E \vdash t = s$
  \item transitivity: $E \vdash s = t$ and $E \vdash t = u$
    implies $E \vdash s = u$
  \item assumption: for every $(s, t) \in E$ and substitution
    $\sigma$, we have $E \vdash s\sigma = t\sigma$ (Note that this
    rules combines the instantiation and assumption rules that are
    for example found in \emph{Term Rewriting and All That}.)
  \item congruence: for every $n$-ary function symbol $f$ and
    $E \vdash s_1 = t_1$, \ldots, $E \vdash s_n = t_n$, we have
    $E \vdash f(s_1, \ldots, s_n) = f(t_1, \ldots, t_n)$
\end{itemize}
*}
datatype ('f, 'v) eq_proof = 
  Refl "('f, 'v) term" 
| Sym "('f, 'v) eq_proof" 
| Trans "('f, 'v) eq_proof" "('f, 'v) eq_proof"
| Assm "('f, 'v) rule" "('f, 'v) subst"
| Cong 'f "('f, 'v) eq_proof list"

lemma
  fixes P :: "('f, 'v) eq_proof \<Rightarrow> bool"
  assumes "\<And>t. P (Refl t)" and "\<And>p. P p \<Longrightarrow> P (Sym p)"
  and "\<And>p1 p2. P p1 \<Longrightarrow> P p2 \<Longrightarrow> P (Trans p1 p2)"
  and "\<And>l r \<sigma>. P (Assm (l, r) \<sigma>)"
  and "\<And>f ps. (\<And>p. p \<in> set ps \<Longrightarrow> P p) \<Longrightarrow> P (Cong f ps)"
  shows eq_proof_induct[case_names Refl Sym Trans Assm Cong, induct type: eq_proof]:
    "P p" and "\<And>p. p \<in> set ps \<Longrightarrow> P p"
  using assms by (induct p and ps) auto

type_synonym ('f, 'v) equation = "('f, 'v) rule"

text {*
Given an equational system @{term E} and an equational logic proof @{term p},
@{term "proves E p"} obtains the equation which is proved, simultaneously checking
the correctness of @{term p}. If @{term p} is not correct the result is @{term None},
otherwise the result is the proved equation.
*}
fun
  proves :: "('f, 'v) trs \<Rightarrow> ('f, 'v) eq_proof \<Rightarrow> ('f, 'v) equation option"
where
  "proves E (Refl s) = Some (s, s)"
| "proves E (Sym p) = do {
    (s, t) \<leftarrow> proves E p;
    Some (t, s) }"
| "proves E (Trans p1 p2) = do {
    (s, t)  \<leftarrow> proves E p1;
    (t', u) \<leftarrow> proves E p2;
    guard (t = t');
    Some (s, u)
  }"
| "proves E (Assm (l, r) \<sigma>) = do {
    guard ((l, r) \<in> E);
    Some (l \<cdot> \<sigma>, r \<cdot> \<sigma>)
  }"
| "proves E (Cong f ps) = do {
    sts \<leftarrow> mapM (proves E) ps;
    Some (Fun f (map fst sts), Fun f (map snd sts))
  }"

text {*
The \emph{equational theory} of an equational system @{term E} is the set
of all provable equations.
*}
definition eq_theory :: "('f, 'v) trs \<Rightarrow> ('f, 'v) equation set" where
  "eq_theory E \<equiv> {e. \<exists>p. proves E p = Some e}"

text {*
In contrast to applying rules of term rewriting, for applying equations
the direction does not matter.
*}
definition estep :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs" where
  "estep E \<equiv> rstep (E \<union> E^-1)"
lemma estep_sym_closure_conv: "estep E = (rstep E)\<^sup>\<leftrightarrow>"
  unfolding estep_def rstep_union reverseTrs ..

lemma join_conversion: "join (rstep R) \<subseteq> (estep R)^*"
proof
  fix a b
  assume "(a,b) \<in> join (rstep R)"
  hence steps: "(a,b) \<in> (rstep R)^* O ((rstep R)^*)^-1" by auto
  show "(a,b) \<in> (estep R)^*"
    by (rule set_mp[OF _ steps], unfold rtrancl_converse[symmetric] estep_sym_closure_conv, regexp)
qed

lemma esteps_is_conversion: "(estep E)^* = conversion (rstep E)"
  unfolding conversion_def estep_sym_closure_conv ..

lemma sym_estep: "sym (estep E)" 
  unfolding estep_sym_closure_conv sym_def by auto

lemma sym_esteps: "sym ((estep E)^*)"
  by (rule sym_rtrancl, rule sym_estep)

lemma subst_closed_estep: "subst_closed (estep E)"
  unfolding estep_def rstep_union
  by (rule subst_closed_union, insert subst_closed_rstep, auto)

lemma ctxt_closed_estep[intro]: "ctxt_closed (estep E)"
  unfolding estep_def rstep_union
  by (rule ctxt_closed_union, insert ctxt_closed_rstep, auto)

lemma sym_esteps_pair: "(s,t) \<in> (estep E)^* \<Longrightarrow> (t,s) \<in> (estep E)^*"
  using sym_esteps[unfolded sym_def] by auto

lemma all_ctxt_closed_esteps: "all_ctxt_closed ((estep E)^*)"
  unfolding estep_def
  by (rule rstep_rtrancl_all_ctxt_closed)

lemma ctxt_closed_eq_theory: "ctxt_closed (eq_theory E)"
proof (rule one_imp_ctxt_closed)
  fix f bef s t aft
  assume "(s,t) \<in> eq_theory E"
  from this[unfolded eq_theory_def] obtain p where p: "proves E p = Some (s,t)" by auto
  show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> eq_theory E" (is "?pair \<in> _")
    unfolding eq_theory_def
  proof(rule, rule exI)
    have id: "mapM (proves E) (map Refl bef @ p # map Refl aft) = (Some (map (\<lambda> b. (b,b)) bef @ (s,t) # map (\<lambda> a. (a,a)) aft))"
    proof (induct bef)
      case (Cons b bef)
      thus ?case by auto
    next
      case Nil
      have "mapM (proves E) (map Refl aft) = Some (map (\<lambda> a. (a,a)) aft)"
        by (induct aft, auto)
      thus ?case unfolding map.simps append_Nil mapM.simps p by simp
    qed
    thus "proves E (Cong f (map (\<lambda> b. Refl b) bef @ p # map (\<lambda> a. Refl a) aft)) = Some ?pair" by (auto simp: o_def)
  qed
qed

text {*
For every sentence of the equational theory of @{term E},
the left-hand side can be transformed into the right-hand side by
applying equations from @{term E}.
*}
lemma eq_theory_is_esteps: "eq_theory E = (estep E)^*" (is "?P = ?E")
proof -
  let ?p = "proves E"
  note conv = esteps_is_conversion conversion_def
  {
    fix s t
    assume "(s,t) \<in> eq_theory E"
    from this[unfolded eq_theory_def] obtain p where
      "proves E p = Some (s,t)" by auto
    hence "(s,t) \<in> ?E"
    proof (induct p arbitrary: s t)
      case (Refl s)
      thus ?case unfolding conv by auto
    next
      case (Sym p)
      from Sym(2) have p: "?p p = Some (t,s)" by (cases "?p p", auto)
      from Sym(1)[OF p] have ts: "(t,s) \<in> ?E" .
      from sym_esteps[of E] ts
      show ?case unfolding sym_def by auto
    next
      case (Trans p1 p2)
      from Trans(3) obtain s1 t1 where p1: "?p p1 = Some (s1,t1)" by (cases "?p p1", auto)
      from Trans(3) obtain s2 t2 where p2: "?p p2 = Some (s2,t2)" using p1 by (cases "?p p2", auto)
      from Trans(3) have t1s2: "t1 = s2" using p1 p2 by (cases "t1 = s2", auto)
      from Trans(3) p1 p2 t1s2 have p1: "?p p1 = Some (s,t1)" and p2: "?p p2 = Some (t1,t)" by auto
      from Trans(1)[OF p1] Trans(2)[OF p2] show ?case by auto
    next
      case (Assm l r \<sigma>)
      from Assm have mem: "(l,r) \<in> E" by (cases "(l,r) \<in> E", auto)
      from Assm mem have id: "s = l \<cdot> \<sigma>" "t = r \<cdot> \<sigma>" by auto
      have "(s,t) \<in> rstep E" unfolding id using mem by auto
      thus ?case unfolding conv by auto
    next
      case (Cong f ps)
      from Cong(2) obtain sts where sts: "mapM ?p ps = Some sts" by (cases "mapM ?p ps", auto)
      let ?ss = "map fst sts"
      let ?ts = "map snd sts"
      from Cong(2) sts
      have id: "s = Fun f ?ss" "t = Fun f ?ts" by auto
      note all = all_ctxt_closed_esteps[of E, unfolded all_ctxt_closed_def, rule_format, of ?ss ?ts f]
      note sts' = mapM_Some[OF sts]
      show ?case unfolding id
      proof (rule all, intro conjI, simp, unfold length_map, intro allI impI)
        fix i
        assume i: "i < length sts"
        have psi: "ps ! i \<in> set ps" using sts sts' i by auto
        show "(?ss ! i, ?ts ! i) \<in> ?E" unfolding nth_map[OF i]
        proof (rule Cong(1)[OF psi])
          from sts' psi have "?p (ps ! i) \<noteq> None" by auto
          thus "?p (ps ! i) = Some (fst (sts ! i), snd (sts ! i))"
            using sts' i by (cases "?p (ps ! i)", auto)
        qed
      qed
    qed
  }
  hence "?P \<subseteq> ?E" by auto
  moreover
  {
    fix s t
    assume "(s,t) \<in> ?E"
    hence "(s,t) \<in> ?P" unfolding eq_theory_def
    proof (induct)
      case base
      have "proves E (Refl s) = Some (s,s)" by simp
      thus ?case by blast
    next
      case (step t u)
      from step(3) obtain p where st: "?p p = Some (s,t)" by auto      
      obtain l r C \<sigma> where lr: "(l,r) \<in> E^<->" and t: "t = C\<langle>l \<cdot> \<sigma>\<rangle>" and u: "u = C\<langle>r \<cdot> \<sigma>\<rangle>" by (rule rstepE[OF step(2)[unfolded estep_def]])
      from lr
      have "\<exists> p. ?p p = Some (l \<cdot> \<sigma>, r \<cdot> \<sigma>)"
      proof
        assume "(l,r) \<in> E"
        hence "?p (Assm (l,r) \<sigma>) = Some (l \<cdot> \<sigma>, r \<cdot> \<sigma>)" by simp
        thus ?thesis ..
      next
        assume "(l,r) \<in> E^-1"
        hence "(r,l) \<in> E" by simp
        hence "?p (Sym (Assm (r,l) \<sigma>)) = Some (l \<cdot> \<sigma>, r \<cdot> \<sigma>)" by simp
        thus ?thesis ..
      qed
      hence "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ?P" unfolding eq_theory_def by auto
      from ctxt_closed_imp_ctxt[OF ctxt_closed_eq_theory this, of C]
      obtain p2 where tu: "?p p2 = Some (t,u)" unfolding t u unfolding eq_theory_def by auto
      from st tu have "?p (Trans p p2) = Some (s,u)" by simp
      thus ?case by blast
    qed
  }
  hence "?E \<subseteq> ?P" by auto
  ultimately show ?thesis by auto
qed

type_synonym ('f, 'a) eq_inter = "'f \<Rightarrow> 'a list \<Rightarrow> 'a"
type_synonym ('v, 'a) eq_assign = "'v \<Rightarrow> 'a"

definition eq_inter :: "'a set \<Rightarrow> ('f, 'a) eq_inter \<Rightarrow> bool" where
  "eq_inter U I \<equiv> \<forall>f as. set as \<subseteq> U \<longrightarrow> I f as \<in> U"

definition eq_assign :: "'a set \<Rightarrow> ('v, 'a) eq_assign \<Rightarrow> bool" where
  "eq_assign U \<alpha> \<equiv> range \<alpha> \<subseteq> U"

text {*
We evaluate terms w.r.t\ an interpretation @{term I} on function symbols
and an assignment @{term \<alpha>} on variables.
*}
fun
  eq_eval :: "('f, 'a) eq_inter \<Rightarrow> ('v, 'a) eq_assign \<Rightarrow> ('f, 'v) term \<Rightarrow> 'a"
where
  "eq_eval I \<alpha> (Var x) = \<alpha> x"
| "eq_eval I \<alpha> (Fun f ts) = I f (map (eq_eval I \<alpha>) ts)"

fun
  eq_model_eqn :: "'a set \<Rightarrow> ('f, 'a) eq_inter \<Rightarrow> ('f, 'v) equation \<Rightarrow> bool"
where
  "eq_model_eqn U I (l, r) = (\<forall>\<alpha>. eq_assign U \<alpha> \<longrightarrow> eq_eval I \<alpha> l = eq_eval I \<alpha> r)"

text {*
The algebra given by the carrier @{term U} and where function symbols are interpreted
by @{term I} is a \emph{model} of an equational system @{term E} if all its equations
are satisfied for arbitrary assignments.
*}
definition models :: "'a set \<Rightarrow> ('f, 'a) eq_inter \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool" where
  "models U I E \<equiv> (\<forall>e\<in>E. eq_model_eqn U I e)"

text {*
The equational system @{term E} \emph{entails} the equation @{term e}, if every
model of @{term E} is also a model of @{term e}.
*}
definition entails :: "'a \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) equation \<Rightarrow> bool" where
  "entails a E e \<equiv> \<forall>U::'a set. \<forall>I. eq_inter U I \<longrightarrow> models U I E \<longrightarrow> eq_model_eqn U I e"

lemma eq_subst_lemma: "eq_eval I \<alpha> (t \<cdot> \<sigma>) = eq_eval I (\<lambda>x. eq_eval I \<alpha> (\<sigma> x)) t" 
proof (induct t)
  case (Var x) thus ?case by simp
next
  case (Fun f ts)
  have id: "map (\<lambda>x. eq_eval I \<alpha> (x \<cdot> \<sigma>)) ts =
    map (eq_eval I (\<lambda>x. eq_eval I \<alpha> (\<sigma> x))) ts"
    using Fun by (induct ts, auto)
  show ?case unfolding eq_eval.simps id[symmetric]
    by (simp add: o_def)
qed

lemma eq_inter_assign:
  assumes I: "eq_inter U I" and \<alpha>: "eq_assign U \<alpha>"
  shows "eq_eval I \<alpha> t \<in> U"
proof (induct t)
  case (Var x)
  show ?case using \<alpha>[unfolded eq_assign_def] by auto
next
  case (Fun f ts)
  hence "eq_eval I \<alpha> ` set ts \<subseteq> U"  by auto
  with I[unfolded eq_inter_def] show ?case by auto
qed

text {*
Every equation that is provable from @{term E} is also
a semantic consequence of @{term E}.
*}
lemma eq_theory_imp_entails: 
  fixes E :: "('f, 'v) trs" and a :: "'a"
  assumes "(s, t) \<in> eq_theory E"
  shows "entails a E (s, t)"
proof -
  let ?p = "proves E"
  from assms[unfolded eq_theory_def] obtain p 
    where eq: "?p p = Some (s,t)" by auto
  show ?thesis unfolding entails_def
  proof (intro impI allI)
    fix U :: "'a set" and I
    assume I: "models U I E" "eq_inter U I"
    let ?J = "eq_eval I"
    show "eq_model_eqn U I (s,t)"
      unfolding eq_model_eqn.simps
    proof (intro impI allI)
      fix \<alpha> :: "('v,'a)eq_assign"
      assume \<alpha>: "eq_assign U \<alpha>"
      from eq
      show "?J \<alpha> s = ?J \<alpha> t" (is "?I s = _")
      proof (induct p arbitrary: s t)
        case (Refl u s t)
        thus ?case by auto
      next
        case (Sym p)
        from Sym(2) have p: "?p p = Some (t,s)" by (cases "?p p", auto)
        from Sym(1)[OF p] show ?case by simp
      next
        case (Trans p1 p2)
        from Trans(3) obtain s1 t1 where p1: "?p p1 = Some (s1,t1)" by (cases "?p p1", auto)
        from Trans(3) obtain s2 t2 where p2: "?p p2 = Some (s2,t2)" using p1 by (cases "?p p2", auto)
        from Trans(3) have t1s2: "t1 = s2" using p1 p2 by (cases "t1 = s2", auto)
        from Trans(3) p1 p2 t1s2 have p1: "?p p1 = Some (s,t1)" and p2: "?p p2 = Some (t1,t)" by auto
        from Trans(1)[OF p1] Trans(2)[OF p2] show ?case by auto
      next
        case (Assm l r \<sigma>)
        from Assm have mem: "(l,r) \<in> E" by (cases "(l,r) \<in> E", auto)
        from Assm mem have id: "s = l \<cdot> \<sigma>" "t = r \<cdot> \<sigma>" by auto
        from I[unfolded models_def] mem have "eq_model_eqn U I (l,r)" by auto
        from this[unfolded eq_model_eqn.simps] have lr: "\<And> \<alpha>. eq_assign U \<alpha> \<Longrightarrow> ?J \<alpha> l = ?J \<alpha> r" by simp
        show ?case unfolding id eq_subst_lemma 
        proof (rule lr, unfold eq_assign_def, rule)
          fix a
          assume "a \<in> range (\<lambda> x. ?I (\<sigma> x))"
          then obtain t where a: "a = ?I t" by auto
          show "a \<in> U" unfolding a
            by (rule eq_inter_assign[OF I(2) \<alpha>])
        qed
      next
        case (Cong f ps)
        from Cong(2) obtain sts where sts: "mapM ?p ps = Some sts" by (cases "mapM ?p ps", auto)
        let ?ss = "map fst sts"
        let ?ts = "map snd sts"
        from Cong(2) sts
        have id: "s = Fun f ?ss" "t = Fun f ?ts" by auto
        note sts' = mapM_Some[OF sts]
        show ?case unfolding id eq_eval.simps
        proof (rule arg_cong[where f = "I f"], rule nth_map_conv, simp, 
            unfold length_map, intro allI impI)
          fix i
          assume i: "i < length sts"
          have psi: "ps ! i \<in> set ps" using sts sts' i by auto
          show "?I (?ss ! i) = ?I (?ts ! i)" unfolding nth_map[OF i]
          proof (rule Cong(1)[OF psi])
            from sts' psi have "?p (ps ! i) \<noteq> None" by auto
            thus "?p (ps ! i) = Some (fst (sts ! i), snd (sts ! i))"
              using sts' i by (cases "?p (ps ! i)", auto)
          qed
        qed
      qed
    qed
  qed
qed

text {*
Every equation that is semantic consequence of @{term E} is
also syntactically provable from @{term E}.
*}
lemma entails_imp_in_eq_theory: 
  fixes E :: "('f, 'v) trs" and a :: "('f, 'v) term set"
  assumes entails: "entails a E (s, t)"
  shows "(s, t) \<in> eq_theory E"
proof -
  let ?E = "(estep E)^*"
  obtain cl where cl: "cl \<equiv> \<lambda> t. {s. (t,s) \<in> ?E}" by auto
  {
    fix s t 
    have "((s,t) \<in> ?E) = (cl s = cl t)"
    proof
      assume st: "(s,t) \<in> ?E"
      from st sym_esteps_pair[OF st]
      show "cl s = cl t" unfolding cl by auto
    next
      assume "cl s = cl t"
      thus "(s,t) \<in> ?E"
        unfolding cl by auto
    qed
  } note esteps_iff_cl = this
  note cl_iff_esteps = esteps_iff_cl[symmetric]
  let ?U = "range cl"
  obtain cl' where cl': "cl' \<equiv> \<lambda> eqc. SOME (t :: ('f,'v)term). t \<in> eqc" by auto 
  obtain a_to_s where a_to_s: "a_to_s \<equiv> \<lambda> \<alpha> :: ('v,('f,'v)term set)eq_assign. \<lambda> x. cl' (\<alpha> x)" by auto   
  obtain I where I: "I \<equiv> \<lambda> f eqs. cl (Fun f (map cl' eqs))" by auto
  let ?I = "eq_eval I"
  have I_inter: "eq_inter ?U I"
    unfolding eq_inter_def cl I by auto
  note entails = entails[unfolded entails_def, rule_format, OF I_inter]
  {
    fix t
    let ?P = "\<lambda> s. (t,s) \<in> ?E"
    let ?s = "SOME s. ?P s"
    have id: "cl' (cl t) = ?s" unfolding cl' cl by auto
    have "cl (cl' (cl t)) = cl t" 
      unfolding id cl_iff_esteps
      by (rule sym_esteps_pair, rule someI[of ?P t], simp)
  } note cl_cl' = this
  {
    fix s and \<alpha> :: "('v,('f,'v)term set)eq_assign"
    assume \<alpha>: "eq_assign ?U \<alpha>"
    have "?I \<alpha> s = cl (s \<cdot> a_to_s \<alpha>)"
    proof (induct s)
      case (Var x)
      from \<alpha>[unfolded eq_assign_def] obtain t where \<alpha>x: "\<alpha> x = cl t" by auto
      show ?case unfolding a_to_s eq_eval.simps subst_apply_term.simps
        unfolding \<alpha>x cl_cl' by simp
    next
      case (Fun f ss)
      from I have id: "I f = (\<lambda> eqs. cl (Fun f (map cl' eqs)))" by auto
      from all_ctxt_closed_esteps
      show ?case unfolding eq_eval.simps subst_apply_term.simps id map_map o_def
        unfolding cl_iff_esteps
      proof (rule all_ctxt_closedE, simp, unfold length_map)
        fix i
        assume i: "i < length ss"
        from Fun[unfolded set_conv_nth, of "ss ! i"] i have id: "?I \<alpha> (ss ! i) = 
          cl (ss ! i \<cdot> a_to_s \<alpha>)" by auto
        show "(map (\<lambda> x. cl' (?I \<alpha> x)) ss ! i, map (\<lambda> t. t \<cdot> a_to_s \<alpha>) ss ! i) \<in> ?E"
          unfolding nth_map[OF i]
          unfolding esteps_iff_cl 
          unfolding id cl_cl' ..
      qed
    qed
  } note eq_eval_id = this
  have model: "models ?U I E"
    unfolding models_def 
  proof(rule, clarify)
    fix s t
    assume st: "(s,t) \<in> E"
    show "eq_model_eqn ?U I (s,t)" unfolding eq_model_eqn.simps
    proof (intro allI impI)
      fix \<alpha> :: "('v,('f,'v)term set)eq_assign"
      assume \<alpha>: "eq_assign ?U \<alpha>"
      from st have "(s \<cdot> a_to_s \<alpha>, t \<cdot> a_to_s \<alpha>) \<in> estep E"
        "(t \<cdot> a_to_s \<alpha>, s \<cdot> a_to_s \<alpha>) \<in> estep E"
        unfolding estep_def by auto
      thus "eq_eval I \<alpha> s = eq_eval I \<alpha> t"
        unfolding eq_eval_id[OF \<alpha>]
        unfolding cl by auto
    qed
  qed
  note entails = entails[OF model, unfolded eq_model_eqn.simps, rule_format]
  obtain \<alpha> where \<alpha>: "\<alpha> \<equiv> \<lambda> x. cl (Var x)" by simp
  have \<alpha>_inter: "eq_assign ?U \<alpha>" unfolding eq_assign_def \<alpha> by auto
  note entails = entails[OF this]
  {
    fix s
    have "cl(s \<cdot> a_to_s \<alpha>) = cl s"
    proof (induct s)
      case (Var x)
      show ?case unfolding a_to_s \<alpha> 
        by (simp add: cl_cl')
    next
      case (Fun f ss)
      from all_ctxt_closed_esteps
      show ?case unfolding subst_apply_term.simps
        unfolding cl_iff_esteps
      proof (rule all_ctxt_closedE, simp, unfold length_map)
        fix i
        assume i: "i < length ss"
        have "map (\<lambda> t. t \<cdot> a_to_s \<alpha>) ss ! i = ss ! i \<cdot> a_to_s \<alpha>" using i by simp
        also have "(ss ! i \<cdot> a_to_s \<alpha>, ss ! i) \<in> ?E" using Fun[unfolded set_conv_nth cl_iff_esteps, of "ss ! i"] i by auto
        finally show "(map (\<lambda> t. t \<cdot> a_to_s \<alpha>) ss ! i, ss ! i) \<in> ?E" .
      qed
    qed
  } note a_to_s = this
  from entails have "?I \<alpha> s = ?I \<alpha> t" .
  from this[unfolded eq_eval_id[OF \<alpha>_inter] a_to_s]  
  show ?thesis unfolding eq_theory_is_esteps esteps_iff_cl .
qed

text {*
Syntactic consequence (provability w.r.t.\ equational logic) and
semantic consequence (entailment) coincide.
*}
lemma birkhoff:
  fixes E :: "('f, 'v) trs" and a :: "('f, 'v) term set" 
  shows "entails a E (s, t) \<longleftrightarrow> (s, t) \<in> eq_theory E"
  using entails_imp_in_eq_theory[of a E s t]
    and eq_theory_imp_entails[of s t E] by auto

text {*
Two equational systems are \emph{equivalent} if they prove
same equations, i.e., have the same equtional theory.
*}
definition equivalent :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool" where
  "equivalent E E' \<equiv> eq_theory E = eq_theory E'"

text {*
The equational system @{term E} \emph{subsumes} the equational system
@{term E'} when all equations that are provable using @{term E'} are
also provable using @{term E}.
*}
definition subsumes :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool" where
  "subsumes E E' \<equiv> eq_theory E' \<subseteq> eq_theory E"

lemma equivalent_subsumes:
  "equivalent E E' \<longleftrightarrow> subsumes E E' \<and> subsumes E' E"
  unfolding subsumes_def equivalent_def by auto

lemma subsumes_via_rule_conversion:
  "subsumes E E' = (\<forall>(s, t)\<in>E'. (s, t) \<in> (estep E)^*)"
proof
  assume subs: "subsumes E E'"
  show "\<forall> (s,t) \<in> E'. (s,t) \<in> (estep E)^*"
  proof(clarify)
    fix s t
    assume "(s,t) \<in> E'"
    hence "(s,t) \<in> (estep E')^*" unfolding estep_def by auto
    with subs show "(s,t) \<in> (estep E)^*" unfolding subsumes_def eq_theory_is_esteps by auto
  qed
next
  assume "\<forall>(s, t)\<in>E'. (s,t) \<in> (estep E)^*"
  hence steps: "\<And>s t. (s, t) \<in> E' \<Longrightarrow> (s, t) \<in> (estep E)^*" by auto
  show "subsumes E E'"
    unfolding subsumes_def eq_theory_is_esteps 
  proof
    fix s t 
    assume "(s,t) \<in> (estep E')^*"
    thus "(s,t) \<in> (estep E)^*"
    proof(induct)
      case base
      show ?case unfolding estep_def by auto
    next
      case (step t u)
      from step(2)[unfolded estep_def] obtain l r C \<sigma> where t: "t = C\<langle>l\<cdot>\<sigma>\<rangle>"
        and u: "u = C\<langle>r \<cdot> \<sigma>\<rangle>" and lr: "(l,r) \<in> E' \<union> E'^-1" by auto
      from lr have steps: "(l,r) \<in> (estep E)^*"
      proof
        assume "(l,r) \<in> E'"
        from steps[OF this] show ?thesis .
      next
        assume "(l,r) \<in> E'^-1" 
        hence "(r,l) \<in> E'" by simp
        from steps[OF this] have "(r,l) \<in> (estep E)^*" .
        with sym_rtrancl[OF sym_estep, of E]
        show ?thesis unfolding sym_def by auto
      qed
      with subst_closed_rtrancl[OF subst_closed_estep[of E], unfolded subst_closed_def] have steps: "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> (estep E)^*" by auto      
      have "(t,u) \<in> (estep E)^*" unfolding t u 
        by (rule ctxt_closed_imp_ctxt[OF _ steps], blast)
      with step(3) show ?case by simp
    qed
  qed
qed

text {*
We say that the TRS @{term R} is a complete rewrite system for the
equational system @{term E} iff their respective equational
theories are the same and @{term R} is convergent, i.e., confluent
and terminating.
*}
definition
  completed_rewrite_system :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool"
where
  "completed_rewrite_system E R \<equiv> equivalent E R \<and> CR (rstep R) \<and> SN (rstep R)"

lemma completion_via_WCR_SN_simulation:
  assumes 1: "subsumes E R" 
    and 2: "subsumes R E"
    and 3: "WCR (rstep R)"
    and 4: "SN (rstep R)"
  shows "completed_rewrite_system E R"
  unfolding completed_rewrite_system_def equivalent_subsumes 
  using Newman[OF 4 3] 1 2 4 by auto

end