(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Linear_Orders
imports
  Labelings
  Term
  "~~/src/HOL/Library/List_lexord"
  "~~/src/HOL/Library/Char_nat"
begin

(* text define own char order, since
  1) Char_ord loads Prod_ord and
  2) Prod_ord is incompatible with Prod_ord in AFP which is required by RBTMap..
*)
instantiation char :: linorder
begin
fun less_eq_char :: "char \<Rightarrow> char \<Rightarrow> bool" where "less_eq_char c d = (nat_of_char c \<le> nat_of_char d)"
fun less_char :: "char \<Rightarrow> char \<Rightarrow> bool" where "less_char c d = (nat_of_char c < nat_of_char d)"

instance
proof
  fix x y :: char
  assume "x \<le> y" and "y \<le> x"
  hence "nat_of_char x = nat_of_char y" (is "?x = ?y") by auto
  hence "char_of_nat ?x = char_of_nat ?y" by simp
  thus "x = y" unfolding char_of_nat_of_char .
qed auto
end


fun lexord_code :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> bool"
  where "lexord_code lt _ [] = False"
    |   "lexord_code lt [] _ = True"
    |   "lexord_code lt (x # xs) (y # ys) = (lt x y \<or> x = y \<and> lexord_code lt xs ys)"

lemma lexord_code: "lexord_code lt xs ys = ((xs,ys) \<in> lexord {(x,y). lt x y})"
proof (induct ys arbitrary: xs)
  case Nil 
  thus ?case by auto
next
  case (Cons y ys xxs) note ys = this
  show ?case
  proof (cases xxs)
    case Nil
    thus ?thesis by auto
  next
    case (Cons x xs)
    with ys[of xs] show ?thesis by auto
  qed
qed

lemma disj_left_cancel: "A = B \<Longrightarrow> C \<or> A \<longleftrightarrow> C \<or> B" by auto

lemma lexord_cong[fundef_cong]:
  assumes "as = as'" and "bs = bs'"
    and i: "\<And>i. \<lbrakk>i < length as'; i < length bs'\<rbrakk> \<Longrightarrow> (as' ! i, bs' ! i) \<in> f \<longleftrightarrow> (as' ! i, bs' ! i) \<in> g"
  shows "(as, bs) \<in> lexord f \<longleftrightarrow> (as', bs') \<in> lexord g"
proof -
  have "(as, bs) \<in> lexord f \<longleftrightarrow> (as',bs') \<in> lexord f" unfolding assms ..
  also have "... \<longleftrightarrow> (as',bs') \<in> lexord g"
    by (cases "length as' \<le> length bs'") (insert i, auto simp: lexord_take_index_conv)
  finally show ?thesis .
qed

lemma lexord_code_cong[fundef_cong]:
  assumes "as = as'" and "bs = bs'"
    and i: "\<And>i. \<lbrakk>i < length as'; i < length bs'\<rbrakk> \<Longrightarrow> f (as' ! i) (bs' ! i) = g (as' ! i) (bs' ! i)" 
  shows "lexord_code f as bs = lexord_code g as' bs'"
  unfolding lexord_code
  by (rule lexord_cong, insert assms, auto)

instantiation lab :: (ord,ord) ord
begin
lemma size_i: assumes "i < length ts" shows "size (ts ! i) < Suc (list_size size ts)" by (rule size_simps(1), insert assms, auto)

function(sequential) less_lab :: "('a :: ord, 'b :: ord)lab \<Rightarrow> ('a,'b)lab \<Rightarrow> bool" 
  where "less_lab (UnLab f) (UnLab g) = (f < g)"
     |  "less_lab (UnLab f) _ = True"
     |  "less_lab (Lab f l) (Lab g l') = (l < l' \<or> l = l' \<and> less_lab f g)"
     |  "less_lab (Lab f l) (UnLab _) = False"
     |  "less_lab (Lab f l) _ = True"
     |  "less_lab (Sharp f) (Sharp g) = (f < g)"
     |  "less_lab (Sharp f) (UnLab _) = False"
     |  "less_lab (Sharp f) (Lab _ _) = False"     
     |  "less_lab (Sharp f) _ = True"     
     |  "less_lab (FunLab f fs) (FunLab g gs) = (less_lab f g \<or> f = g \<and> lexord_code less_lab fs gs)"
     |  "less_lab (FunLab f fs) _ = False"     
  by pat_completeness auto

termination 
proof
  fix f :: "('f,'v)lab" and fs :: "('f,'v)lab list" and g :: "('f,'v)lab" and gs :: "('f,'v)lab list" and i 
  assume i: "i < length fs" 
  show "((fs ! i,gs ! i), (FunLab f fs, FunLab g gs))
    \<in> measures [\<lambda> i. case i of (xs,ys) \<Rightarrow> size xs]"
    using size_i[OF i] by auto
qed auto

fun less_eq_lab :: "('a :: ord, 'b :: ord)lab \<Rightarrow> ('a,'b)lab \<Rightarrow> bool" 
  where "less_eq_lab xs ys = (xs < ys \<or> xs = ys)"
    
instance  ..
end


lemma lab_less_trans: assumes "x < (y :: ('a :: order,'b :: order)lab)" 
  and "y < z"
  shows "x < z"
  using assms 
proof (induct x arbitrary: y z)
  case (UnLab f) note x = this
  show ?case 
  proof (cases z)
    case (UnLab h)
    then obtain g where y: "y = UnLab g" using x by (cases y, auto)
    from x
    show ?thesis unfolding x y UnLab by auto
  qed (auto simp: UnLab)
next
  case (Lab f l) note x = this
  show ?case
  proof (cases z)
    case (Lab h lll) note z = this
    from x(2) x(3) obtain g ll where y: "y = Lab g ll" 
      unfolding z by (cases y, auto)
    note x = x[unfolded y z]
    from x(2) have one: "l < ll \<or> l = ll \<and> f < g" by auto
    from x(3) have two: "ll < lll \<or> ll = lll \<and> g < h" by auto
    hence "l < lll \<or> l = lll \<and> f < h" 
    proof
      assume l2: "ll < lll" 
      from one have "l < ll \<or> l = ll" by auto
      thus ?thesis
      proof
        assume "l < ll"
        with l2 have "l < lll" by auto
        thus ?thesis by simp
      qed (insert l2, auto)
    next
      assume "ll = lll \<and> g < h"
      hence l2: "ll = lll" and gh: "g < h" by auto
      from l2 one x(1)[OF _ gh] show ?thesis by auto
    qed      
    thus ?thesis unfolding z by auto
  qed ((cases y, insert x(2) x(3), auto)+)
next
  case (Sharp f) note x = this
  show ?case
  proof (cases z)
    case (Sharp h) note z = this
    from x(2) x(3) obtain g where y: "y = Sharp g" 
      unfolding z by (cases y, auto)
    show ?thesis using x(2) x(3) x(1) unfolding y z by auto
  qed ((cases y, insert x(2) x(3), auto)+)
next
  case (FunLab f l) note x = this
  from x(3) obtain g ll where y: "y = FunLab g ll" 
    by (cases y, auto)
  from x(4)[unfolded y] obtain h lll where z: "z = FunLab h lll" by (cases z, auto)
  note x = x[unfolded y z]
  let ?lo = "lexord {(x, y). x < y}"
  from x(3) x(4) have one: "f < g \<or> f = g \<and> (l,ll) \<in> ?lo" 
                  and two: "g < h \<or> g = h \<and> (ll,lll) \<in> ?lo"
    by (auto simp: lexord_code)
  have "f < h \<or> f = h \<and> (l,lll) \<in> ?lo"
  proof (cases "g < h")
    case True
    from x(1)[OF _ this] one True show ?thesis by auto
  next
    case False
    with two have gh: "g = h" and two: "(ll,lll) \<in> ?lo" by auto
    show ?thesis
    proof (cases "f < g")
      case True
      with gh show ?thesis by auto
    next
      case False
      with one have fg: "f = g" and one: "(l,ll) \<in> ?lo" by auto
      have main: "(l,lll) \<in> ?lo"
        by (rule lexord_partial_trans[OF _ one two], insert x(2), auto)
      thus ?thesis unfolding fg gh by auto
    qed
  qed
  thus ?case unfolding z by (auto simp: lexord_code)
qed    

instantiation lab :: (order,order)preorder
begin
instance 
proof
  fix x :: "('a,'b)lab" 
  show "x \<le> x" by auto
next
  fix x y :: "('a,'b)lab"
  show "(x < y) = (x \<le> y \<and> \<not> y \<le> x)" (is "?l = ?r") 
  proof
    assume ?r thus ?l by auto
  next
    assume ?l hence one: "x \<le> y" by auto
    from `x < y` have two: "\<not> y \<le> x" 
    proof (induct x arbitrary: y)
      case (UnLab f)
      thus ?case by (cases y, auto)
    next
      case (Lab f l) note x = this
      show ?case 
      proof (cases y)
        case (Lab g ll) note y = this
        with x(1)[of g] x(2) show ?thesis unfolding y by auto
      qed (insert x(2), auto)
    next
      case (Sharp f)
      thus ?case by (cases y, auto)
    next
      case (FunLab f ls) note x = this
      show ?case
      proof (cases y)
        case (FunLab g lls) note y = this
        note x = x[unfolded y]
        let ?lo = "lexord {(x,y). x < y}"
        from x(3)[simplified] have "f < g \<or> f = g \<and> (ls, lls) \<in> ?lo" (is "_ \<or> _ \<and> ?cond") by (auto simp: lexord_code)
        thus ?thesis
        proof
          assume "f < g"
          from x(1)[OF this] show ?thesis unfolding y by auto
        next
          assume "f = g \<and> ?cond" 
          hence fg: "f = g" and cond: ?cond by auto
          from x(1)[of g, unfolded fg] have gg: "\<not> g < g" by auto
          from cond have "(ls,lls) \<in> lexord {(x,y). x < y \<and> x \<in> set ls}"
            (is "_ \<in> lexord ?r")
            unfolding lexord_take_index_conv by force
          {
            fix x
            assume "(x,x) \<in> ?r"
            hence xl: "x \<in> set ls" and xx: "x < x" by auto
            from x(2)[OF xl xx] have False by auto
          }
          hence "\<forall> x. (x,x) \<notin> ?r" by force
          from lexord_irreflexive[OF this] have "(ls,ls) \<notin> lexord ?r" by auto
          hence ls: "(ls,ls) \<notin> ?lo" 
            unfolding lexord_take_index_conv by auto
          with cond have neq: "lls \<noteq> ls" by auto
          {
            assume lt: "(lls,ls) \<in> ?lo"
            have "(ls,ls) \<in> ?lo"
              by (rule lexord_trans[OF cond lt], unfold trans_def, insert lab_less_trans, auto)
            with ls have False ..
          }               
          thus ?thesis unfolding y fg 
            by (auto simp: gg neq lexord_code)
        qed            
      qed (insert x(3), auto)
    qed
    from one two show ?r by auto
  qed 
next
  fix x y z :: "('a,'b)lab"
  assume "x \<le> y" and "y \<le> z"
  thus "x \<le> z" using lab_less_trans[of x y z] by auto
qed
end

instantiation lab :: (order,order) order
begin
instance
proof
  fix x y :: "('a,'b)lab"
  assume "x \<le> y" and "y \<le> x"
  thus "x = y"  
  proof (induct x arbitrary: y)
    case (UnLab f)
    thus ?case by (cases y, auto)
  next
    case (Lab f l) note x = this
    show ?case 
    proof (cases y)
      case (Lab g ll) note y = this
      from x(1)[of g] x(2) x(3) show ?thesis unfolding y by auto
    qed (insert x(2) x(3), auto)
  next
    case (Sharp f)
    thus ?case by (cases y, auto)
  next
    case (FunLab f ls) note x = this
    show ?case
    proof (cases y)
      case (FunLab g lls) note y = this
      note x = x[unfolded y]
      let ?lo = "lexord {(x,y). x < y}"
      from x(3) have one: "f = g \<and> ls = lls \<or> f < g \<or> f = g \<and> (ls,lls) \<in> ?lo" (is "_ \<or> ?cond") by (auto simp: lexord_code)
      from x(4) have two: "g = f \<and> lls = ls \<or> g < f \<or> g = f \<and> (lls, ls) \<in> ?lo" (is "_ \<or> ?cond2") by (auto simp: lexord_code)
      have "f = g \<and> ls = lls" 
      proof (rule ccontr)
        assume not: "\<not> ?thesis"
        from one not have cond: "?cond" by auto
        from two not have cond2: "?cond2" by auto
        from cond cond2 have "f \<le> g" and "g \<le> f" by auto
        from x(1)[OF this] have fg: "f = g" by auto
        with not have not: "\<not> ls = lls" by simp
        from cond[unfolded fg] have one: "(ls,lls) \<in> ?lo" by auto
        from cond2[unfolded fg] have two: "(lls,ls) \<in> ?lo" by auto
        from x(2) one two have "ls = lls"
        proof (induct ls arbitrary: lls)
          case Nil
          thus ?case by auto
        next
          case (Cons x xs yys)
          from Cons(3) obtain y ys where yys: "yys = y # ys" unfolding lexord_def by (cases yys, auto)
          note Cons = Cons[unfolded yys]
          from Cons(3) Cons(4) have "x \<le> y" and "y \<le> x" by auto
          from Cons(2)[OF _ this] have yx: "y = x" by auto
          note Cons = Cons[unfolded yx]
          from Cons(3) Cons(4) have one: "(xs,ys) \<in> ?lo" and two: "(ys,xs) \<in> ?lo" by auto
          have "xs = ys" 
          proof (rule Cons(1)[OF _ one two])
            fix xx y
            assume "xx \<in> set xs" and one: "xx \<le> y" and two: "y \<le> xx"
            hence "xx \<in> set (x # xs)" by auto
            from Cons(2)[OF this one two] show "xx = y" .
          qed
          with yx show ?case unfolding yys by auto
        qed
        with not show False by auto
      qed
      thus ?thesis unfolding y by auto
    qed (insert x(3) x(4), auto)
  qed
qed
end


instantiation lab :: (linorder,linorder) linorder
begin
instance
proof
  fix x y :: "('a,'b)lab"
  show "x \<le> y \<or> y \<le> x" 
  proof (induct x arbitrary: y)
    case (UnLab f)
    thus ?case by (cases y, auto)
  next
    case (Lab f l) note x = this
    show ?case 
    proof (cases y)
      case (Lab g ll) note y = this
      from x(1)[of g] show ?thesis unfolding y by auto
    qed auto
  next
    case (Sharp f)
    thus ?case by (cases y, auto)
  next
    case (FunLab f ls) note x = this
    show ?case
    proof (cases y)
      case (FunLab g lls) note y = this
      note x = x[unfolded y]
      from x(1)[of g] have fg: "f = g \<or> f < g \<or> g < f" by auto
      let ?lo = "lexord {(x,y). x < y}"
      from x(2) have ls: "ls = lls \<or> (ls,lls) \<in> ?lo \<or> (lls,ls) \<in> ?lo" (is "?cond ls lls")
      proof (induct ls arbitrary: lls)
        case Nil
        thus ?case by (cases lls, auto)
      next
        case (Cons x xs yys) note xxs = this
        show ?case
        proof (cases yys)
          case Nil
          thus ?thesis by auto
        next
          case (Cons y ys) note yys = this
          from xxs(2)[of x y] have xy: "x = y \<or> x < y \<or> y < x" by auto
          have xsys: "?cond xs ys" 
            by (rule xxs(1), insert xxs(2), auto)
          from xy xsys show ?thesis unfolding yys by auto
        qed
      qed 
      thus ?thesis using fg unfolding y by (auto simp: lexord_code)
    qed auto
  qed
qed
end

instantiation "term" :: (ord,ord) ord
begin

function(sequential) less_term :: "('a :: ord, 'b :: ord)term \<Rightarrow> ('a,'b)term \<Rightarrow> bool" 
  where "less_term (Var f) (Var g) = (f < g)"
     |  "less_term (Var f) _ = True"
     |  "less_term (Fun f fs) (Fun g gs) = (f < g \<or> f = g \<and> lexord_code less_term fs gs)"
     |  "less_term (Fun f fs) _ = False"     
  by pat_completeness auto

termination 
proof
  fix f :: "'f" and fs :: "('f,'v)term list" and g :: "'f" and gs :: "('f,'v)term list" and i 
  assume i: "i < length fs" 
  show "((fs ! i,gs ! i), (Fun f fs, Fun g gs))
    \<in> measures [\<lambda> i. case i of (xs,ys) \<Rightarrow> size xs]"
    using size_i[OF i] by auto
qed auto

fun less_eq_term :: "('a :: ord, 'b :: ord)term \<Rightarrow> ('a,'b)term \<Rightarrow> bool" 
  where "less_eq_term xs ys = (xs < ys \<or> xs = ys)"
    
instance  ..
end


lemma term_less_trans: assumes "x < (y :: ('a :: order,'b :: order)term)" 
  and "y < z"
  shows "x < z"
  using assms 
proof (induct x arbitrary: y z)
  case (Var f) note x = this
  show ?case 
  proof (cases z)
    case (Var h)
    then obtain g where y: "y = Var g" using x by (cases y, auto)
    from x
    show ?thesis unfolding x y Var by auto
  qed (auto simp: Var)
next
  case (Fun f l) note x = this
  from x(2) obtain g ll where y: "y = Fun g ll" 
    by (cases y, auto)
  from x(3)[unfolded y] obtain h lll where z: "z = Fun h lll" by (cases z, auto)
  note x = x[unfolded y z]
  let ?lo = "lexord {(x, y). x < y}"
  from x(2) x(3) have one: "f < g \<or> f = g \<and> (l,ll) \<in> ?lo" 
                  and two: "g < h \<or> g = h \<and> (ll,lll) \<in> ?lo"
    by (auto simp: lexord_code)
  have "f < h \<or> f = h \<and> (l,lll) \<in> ?lo" 
  proof (cases "g < h")
    case True
    thus ?thesis using one by force
  next
    case False
    with two have gh: "g = h" and two: "(ll,lll) \<in> ?lo" by auto
    show ?thesis
    proof (cases "f < g")
      case True
      with gh show ?thesis by force
    next
      case False
      with one have fg: "f = g" and one: "(l,ll) \<in> ?lo" by auto
      have main: "(l,lll) \<in> ?lo"
        by (rule lexord_partial_trans[OF _ one two], insert x(1), auto)
      thus ?thesis unfolding fg gh by auto
    qed
  qed
  thus ?case unfolding z by (auto simp: lexord_code)
qed    

instantiation "term" :: (order,order)preorder
begin
instance 
proof
  fix x :: "('a,'b)term" 
  show "x \<le> x" by auto
next
  fix x y :: "('a,'b)term"
  show "(x < y) = (x \<le> y \<and> \<not> y \<le> x)" (is "?l = ?r") 
  proof
    assume ?r thus ?l by auto
  next
    assume ?l hence one: "x \<le> y" by auto
    from `x < y` have two: "\<not> y \<le> x" 
    proof (induct x arbitrary: y)
      case (Var f)
      thus ?case by (cases y, auto)
    next
      case (Fun f ls) note x = this
      show ?case
      proof (cases y)
        case (Fun g lls) note y = this
        note x = x[unfolded y]
        let ?lo = "lexord {(x,y). x < y}"
        from x(2)[simplified] have "f < g \<or> f = g \<and> (ls, lls) \<in> ?lo" (is "_ \<or> _ \<and> ?cond") by (auto simp: lexord_code)
        thus ?thesis
        proof
          assume "f < g"
          thus ?thesis unfolding y by force
        next
          assume "f = g \<and> ?cond" 
          hence fg: "f = g" and cond: ?cond by auto
          have gg: "\<not> g < g" by auto
          from cond have "(ls,lls) \<in> lexord {(x,y). x < y \<and> x \<in> set ls}"
            (is "_ \<in> lexord ?r")
            unfolding lexord_take_index_conv by force
          {
            fix x
            assume "(x,x) \<in> ?r"
            hence xl: "x \<in> set ls" and xx: "x < x" by auto
            from x(1)[OF xl xx] have False by auto
          }
          hence "\<forall> x. (x,x) \<notin> ?r" by force
          from lexord_irreflexive[OF this] have "(ls,ls) \<notin> lexord ?r" by auto
          hence ls: "(ls,ls) \<notin> ?lo" 
            unfolding lexord_take_index_conv by auto
          with cond have neq: "lls \<noteq> ls" by auto
          {
            assume lt: "(lls,ls) \<in> ?lo"
            have "(ls,ls) \<in> ?lo"
              by (rule lexord_trans[OF cond lt], unfold trans_def, insert term_less_trans, auto)
            with ls have False ..
          }               
          thus ?thesis unfolding y fg 
            by (auto simp: gg neq lexord_code)
        qed            
      qed (insert x(2), auto)
    qed
    from one two show ?r by auto
  qed 
next
  fix x y z :: "('a,'b)term"
  assume "x \<le> y" and "y \<le> z"
  thus "x \<le> z" using term_less_trans[of x y z] by auto
qed
end

instantiation "term" :: (order,order) order
begin
instance
proof
  fix x y :: "('a,'b)term"
  assume "x \<le> y" and "y \<le> x"
  thus "x = y"  
  proof (induct x arbitrary: y)
    case (Var f)
    thus ?case by (cases y, auto)
  next
    case (Fun f ls) note x = this
    show ?case
    proof (cases y)
      case (Fun g lls) note y = this
      note x = x[unfolded y]
      let ?lo = "lexord {(x,y). x < y}"
      from x(2) have one: "f = g \<and> ls = lls \<or> f < g \<or> f = g \<and> (ls,lls) \<in> ?lo" (is "_ \<or> ?cond") by (auto simp: lexord_code)
      from x(3) have two: "g = f \<and> lls = ls \<or> g < f \<or> g = f \<and> (lls, ls) \<in> ?lo" (is "_ \<or> ?cond2") by (auto simp: lexord_code)
      have "f = g \<and> ls = lls" 
      proof (rule ccontr)
        assume not: "\<not> ?thesis"
        from one not have cond: "?cond" by auto
        from two not have cond2: "?cond2" by auto
        from cond cond2 have "f \<le> g" and "g \<le> f" by auto
        hence fg: "f = g" by auto
        with not have not: "\<not> ls = lls" by simp
        from cond[unfolded fg] have one: "(ls,lls) \<in> ?lo" by auto
        from cond2[unfolded fg] have two: "(lls,ls) \<in> ?lo" by auto
        from x(1) one two have "ls = lls"
        proof (induct ls arbitrary: lls)
          case Nil
          thus ?case by auto
        next
          case (Cons x xs yys)
          from Cons(3) obtain y ys where yys: "yys = y # ys" unfolding lexord_def by (cases yys, auto)
          note Cons = Cons[unfolded yys]
          from Cons(3) Cons(4) have "x \<le> y" and "y \<le> x" by auto
          from Cons(2)[OF _ this] have yx: "y = x" by auto
          note Cons = Cons[unfolded yx]
          from Cons(3) Cons(4) have one: "(xs,ys) \<in> ?lo" and two: "(ys,xs) \<in> ?lo" by auto
          have "xs = ys" 
          proof (rule Cons(1)[OF _ one two])
            fix xx y
            assume "xx \<in> set xs" and one: "xx \<le> y" and two: "y \<le> xx"
            hence "xx \<in> set (x # xs)" by auto
            from Cons(2)[OF this one two] show "xx = y" .
          qed
          with yx show ?case unfolding yys by auto
        qed
        with not show False by auto
      qed
      thus ?thesis unfolding y by auto
    qed (insert x(2) x(3), auto)
  qed
qed
end


instantiation "term" :: (linorder,linorder) linorder
begin
instance
proof
  fix x y :: "('a,'b)term"
  show "x \<le> y \<or> y \<le> x" 
  proof (induct x arbitrary: y)
    case (Var f)
    thus ?case by (cases y, auto)
  next
    case (Fun f ls) note x = this
    show ?case
    proof (cases y)
      case (Fun g lls) note y = this
      note x = x[unfolded y]
      have fg: "f = g \<or> f < g \<or> g < f" by auto
      let ?lo = "lexord {(x,y). x < y}"
      from x(1) have ls: "ls = lls \<or> (ls,lls) \<in> ?lo \<or> (lls,ls) \<in> ?lo" (is "?cond ls lls")
      proof (induct ls arbitrary: lls)
        case Nil
        thus ?case by (cases lls, auto)
      next
        case (Cons x xs yys) note xxs = this
        show ?case
        proof (cases yys)
          case Nil
          thus ?thesis by auto
        next
          case (Cons y ys) note yys = this
          from xxs(2)[of x y] have xy: "x = y \<or> x < y \<or> y < x" by auto
          have xsys: "?cond xs ys" 
            by (rule xxs(1), insert xxs(2), auto)
          from xy xsys show ?thesis unfolding yys by auto
        qed
      qed 
      thus ?thesis using fg unfolding y by (auto simp: lexord_code)
    qed auto
  qed
qed
end

end
