(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)

header {* Term *}

theory Term
imports
  Abstract_Rewriting2
  Position
  Util
begin

text {* A theory on first-order terms. *}

text {*A \emph{signature} is a set of function symbols arity pairs,
where the arity of a function symbol, denotes the number of arguments it expects.*}
type_synonym 'f sig = "('f \<times> nat) set"

text {*(First-order) \emph{terms} are built from function symbols and variables.*}
datatype ('f, 'v) "term" = Var 'v | Fun 'f "('f, 'v) term list"

text {*An alternative induction scheme for terms.*}
lemma
  fixes P :: "('f, 'v) term \<Rightarrow> bool"
  assumes "\<And>x. P(Var x)" and "\<And>f ss. (\<And>s. s \<in> set ss \<Longrightarrow> P s) \<Longrightarrow> P(Fun f ss)"
  shows term_induct[case_names Var Fun, induct type: "term"]:
    "P t" and "\<And>t. t \<in> set ts \<Longrightarrow> P t"
using assms by (induct t and ts) auto

text {*General folds on terms.*}

fun
  fold :: "('v \<Rightarrow> 'a) \<Rightarrow> ('f \<Rightarrow> 'a list \<Rightarrow> 'a) \<Rightarrow> ('f, 'v) term \<Rightarrow> 'a"
where
  "fold var fun (Var x) = var x"
| "fold var fun (Fun f ts) = fun f (map (fold var fun) ts)"

hide_const (open) fold

definition
  map :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('g, 'w) term"
where
  "map var fun = Term.fold (Var \<circ> var) (Fun \<circ> fun)"

hide_const (open) map


subsection {* Variables *}

primrec is_Var :: "('f, 'v) term \<Rightarrow> bool" where
  "is_Var (Var x) = True"
| "is_Var (Fun f ts) = False"
declare is_Var.simps[simp del]

abbreviation is_Fun :: "('f, 'v) term \<Rightarrow> bool" where
  "is_Fun t \<equiv> \<not> is_Var t"

lemma is_VarE[elim]:
  "is_Var t \<Longrightarrow> (\<And>x. t = Var x \<Longrightarrow> P) \<Longrightarrow> P"
  by (cases t) (simp_all add: is_Var.simps)

lemma is_FunE[elim]:
  "is_Fun t \<Longrightarrow> (\<And>f ts. t = Fun f ts \<Longrightarrow> P) \<Longrightarrow> P"
  by (cases t) (auto simp: is_Var.simps)

lemma is_VarI[intro, simp]: "is_Var (Var x)"
  by (simp add: is_Var.simps)

lemma is_FunI[intro, simp]: "is_Fun (Fun f ts)"
  by (simp add: is_Var.simps)

text {*The set of all variables occuring in a term.*}

fun vars_term :: "('f, 'v) term \<Rightarrow> 'v set" where
  "vars_term (Var x) = {x}"
| "vars_term (Fun f ts) = \<Union>set (map vars_term ts)"


text {*The set of all function symbols occuring in a term.*}

fun funs_term :: "('f, 'v) term \<Rightarrow> 'f set" where
  "funs_term (Var _) = {}"
| "funs_term (Fun f ts) = {f} \<union> \<Union>set (map funs_term ts)"

text {*The set of all function symbol and arity pairs occuring in a term.*}

fun funas_term :: "('f, 'v) term \<Rightarrow> 'f sig" where
  "funas_term (Var _) = {}"
| "funas_term (Fun f ts) = {(f, length ts)} \<union> \<Union>set (map funas_term ts)"

fun args :: "('f, 'v) term \<Rightarrow> ('f, 'v) term list" where
  "args (Var x) = []"
| "args (Fun f ts) = ts"

fun num_args :: "('f, 'v) term \<Rightarrow> nat" where
  "num_args (Var x) = 0"
| "num_args (Fun f ts) = length ts"

definition funas_args_term :: "('f, 'v) term \<Rightarrow> 'f sig" where 
  "funas_args_term t = \<Union> set (map funas_term (args t))"

text {*The \emph{root symbol} of a term is defined by:*}
fun root :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) option" where
  "root (Var x) = None"
| "root (Fun f ts) = Some (f, length ts)"

lemma funas_term_conv:
  "funas_term t = Option.set (root t) \<union> funas_args_term t"
  by (cases t) (simp_all add: funas_args_term_def)

lemma finite_funas_term: "finite (funas_term t)"
  by (induct t, auto)

text {*A term is called \emph{ground} if it does not contain any variables.*}
fun ground :: "('f, 'v) term \<Rightarrow> bool" where
  "ground (Var x) = False"
| "ground (Fun f ts) = (\<forall>t\<in>set ts. ground t)"

text {*The \emph{depth} of a term is defined as follows:*}
fun depth :: "('f, 'v) term \<Rightarrow> nat" where
  "depth (Var _) = 0"
| "depth (Fun _ []) = 0"
| "depth (Fun _ ts) = 1 + Max (set (map depth ts))"


subsection {* Subterms *}

text {* The \emph{superterm} relation. *}
inductive_set
  supteq :: "(('f, 'v) term \<times> ('f, 'v) term) set"
where
  refl: "(t, t) \<in> supteq" |
  subt: "u \<in> set ss \<Longrightarrow> (u, t) \<in> supteq \<Longrightarrow> (Fun f ss, t) \<in> supteq"

text {* The \emph{proper superterm} relation. *}
inductive_set
  supt :: "(('f, 'v) term \<times> ('f, 'v) term) set"
where
  arg: "s \<in> set ss \<Longrightarrow> (Fun f ss, s) \<in> supt" |
  subt: "s \<in> set ss \<Longrightarrow> (s, t) \<in> supt \<Longrightarrow> (Fun f ss, t) \<in> supt"

hide_const suptp supteqp
hide_fact
  suptp.arg suptp.cases suptp.induct suptp.intros suptp.subt suptp_supt_eq
hide_fact
  supteqp.cases supteqp.induct supteqp.intros supteqp.refl supteqp.subt supteqp_supteq_eq

hide_fact (open) supt.arg supt.subt supteq.refl supteq.subt

subsubsection {* Syntax for Subterms *}

text {* Infix syntax. *}
abbreviation supt_pred where "supt_pred s t \<equiv> (s, t) \<in> supt"
abbreviation supteq_pred where "supteq_pred s t \<equiv> (s, t) \<in> supteq"
abbreviation (input) subt where "subt s t \<equiv> supt_pred t s"
abbreviation (input) subteq where "subteq s t \<equiv> supteq_pred t s"

notation
  supt ("op |>") and
  supt_pred ("(_/ |> _)" [56, 56] 55) and
  subt (infix "<|" 55) and
  supteq ("op |>=") and
  supteq_pred ("(_/ |>= _)" [56, 56] 55) and
  subteq (infix "<|=" 55)

notation (xsymbols)
  supt ("op \<rhd>") and
  supt_pred ("(_/ \<rhd> _)" [56, 56] 55) and
  subt (infix "\<lhd>" 55) and
  supteq ("op \<unrhd>") and
  supteq_pred ("(_/ \<unrhd> _)" [56, 56] 55) and
  subteq (infix "\<unlhd>" 55)

text {* Quantifier syntax. *}

syntax
  "_All_supteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3ALL _|>=_./ _)" [0, 0, 10] 10)
  "_Ex_supteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3EX _|>=_./ _)" [0, 0, 10] 10)
  "_All_supt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3ALL _|>_./ _)" [0, 0, 10] 10)
  "_Ex_supt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3EX _|>_./ _)" [0, 0, 10] 10)

  "_All_subteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3ALL _<|=_./ _)" [0, 0, 10] 10)
  "_Ex_subteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3EX _<|=_./ _)" [0, 0, 10] 10)
  "_All_subt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3ALL _<|_./ _)" [0, 0, 10] 10)
  "_Ex_subt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3EX _<|_./ _)" [0, 0, 10] 10)

syntax (xsymbols)
  "_All_supteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<forall>_\<unrhd>_./ _)" [0, 0, 10] 10)
  "_Ex_supteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<exists>_\<unrhd>_./ _)" [0, 0, 10] 10)
  "_All_supt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<forall>_\<rhd>_./ _)" [0, 0, 10] 10)
  "_Ex_supt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<exists>_\<rhd>_./ _)" [0, 0, 10] 10)

  "_All_subteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<forall>_\<unlhd>_./ _)" [0, 0, 10] 10)
  "_Ex_subteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<exists>_\<unlhd>_./ _)" [0, 0, 10] 10)
  "_All_subt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<forall>_\<lhd>_./ _)" [0, 0, 10] 10)
  "_Ex_subt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<exists>_\<lhd>_./ _)" [0, 0, 10] 10)

(* for parsing *)
translations
  "ALL x|>=y. P" => "ALL x. x |>= y \<longrightarrow> P"
  "EX x|>=y. P"  => "EX x. x |>= y \<and> P"
  "ALL x|>y. P"  => "ALL x. x |> y \<longrightarrow> P"
  "EX x|>y. P"   => "EX x. x |> y \<and> P"
  
  "ALL x<|=y. P" => "ALL x. x <|= y \<longrightarrow> P"
  "EX x<|=y. P"  => "EX x. x <|= y \<and> P"
  "ALL x<|y. P"  => "ALL x. x <| y \<longrightarrow> P"
  "EX x<|y. P"   => "EX x. x <| y \<and> P"

(* for printing *)
print_translation {*
let
  val All_binder = Mixfix.binder_name @{const_syntax All};
  val Ex_binder = Mixfix.binder_name @{const_syntax Ex};
  val impl = @{const_syntax "implies"};
  val conj = @{const_syntax "conj"};
  val supt = @{const_syntax "supt_pred"};
  val supteq = @{const_syntax "supteq_pred"};

  val trans =
   [((All_binder, impl, supt), ("_All_supt", "_All_subt")),
    ((All_binder, impl, supteq), ("_All_supteq", "_All_subteq")),
    ((Ex_binder, conj, supt), ("_Ex_supt", "_Ex_subt")),
    ((Ex_binder, conj, supteq), ("_Ex_supteq", "_Ex_subteq"))];

  fun matches_bound v t = 
     case t of (Const ("_bound", _) $ Free (v', _)) => (v = v')
              | _ => false
  fun contains_var v = Term.exists_subterm (fn Free (x, _) => x = v | _ => false)
  fun mk v c n P = Syntax.const c $ Syntax_Trans.mark_bound v $ n $ P

  fun tr' q = (q,
    fn [Const ("_bound", _) $ Free (v, _), Const (c, _) $ (Const (d, _) $ t $ u) $ P] =>
      (case AList.lookup (op =) trans (q, c, d) of
        NONE => raise Match
      | SOME (l, g) =>
          if matches_bound v t andalso not (contains_var v u) then mk v l u P
          else if matches_bound v u andalso not (contains_var v t) then mk v g t P
          else raise Match)
     | _ => raise Match);
in [tr' All_binder, tr' Ex_binder] end
*}


subsubsection {* Transitivity Reasoning for Subterms *}

lemma supt_trans[trans]: "s \<rhd> t \<Longrightarrow> t \<rhd> u \<Longrightarrow> s \<rhd> u"
by (induct s t rule: supt.induct) (auto intro: supt.intros)

lemma supteq_trans[trans]: "s \<unrhd> t \<Longrightarrow> t \<unrhd> u \<Longrightarrow> s \<unrhd> u"
by (induct s t rule: supteq.induct) (auto intro: supteq.intros)

text {* Auxiliary lemmas about term size. *}
lemma size_simp5:
  "s \<in> set ss \<Longrightarrow> s \<rhd> t \<Longrightarrow> size t < size s \<Longrightarrow> size t < Suc (list_size size ss)"
by (induct ss) auto
lemma size_simp6:
  "s \<in> set ss \<Longrightarrow> (s, t) \<in> supt \<Longrightarrow> size t < size s \<Longrightarrow> size t < Suc (list_size size ss)"
by (induct ss) auto
lemma size_simp7:
  "s \<in> set ss \<Longrightarrow> s \<unrhd> t \<Longrightarrow> size t \<le> size s \<Longrightarrow> size t \<le> Suc (list_size size ss)"
by (induct ss) auto
lemma size_simp8:
  "s \<in> set ss \<Longrightarrow> (s, t) \<in> supteq \<Longrightarrow> size t \<le> size s \<Longrightarrow> size t \<le> Suc (list_size size ss)"
by (induct ss) auto
lemmas size_simps = size_simp1 size_simp2 size_simp3 size_simp4 size_simp5 size_simp6 size_simp7 size_simp8

declare size_simps[termination_simp]

lemma supt_size: "s \<rhd> t \<Longrightarrow> size s > size t"
by (induct rule: supt.induct) (auto simp: size_simps)

lemma supteq_size: "s \<unrhd> t \<Longrightarrow> size s \<ge> size t"
by (induct rule: supteq.induct) (auto simp: size_simps)

text {* Reflexivity and Asymmetry. *}

declare supteq.refl[iff]

lemma reflcl_supteq[simp]: "supteq^= = supteq" by auto

lemma trancl_supteq[simp]: "supteq^+ = supteq"
by (rule trancl_id) (auto simp: trans_def intro: supteq_trans)

lemma rtrancl_supteq[simp]: "supteq^* = supteq"
unfolding trancl_reflcl[symmetric] by auto

lemma eq_supteq: "s = t \<Longrightarrow> s \<unrhd> t" by auto

lemma supt_neqD: "s \<rhd> t \<Longrightarrow> s \<noteq> t" using supt_size by auto

lemma supteq_Var[simp]: "x \<in> vars_term t \<Longrightarrow> t \<unrhd> Var x"
proof (induct t)
  case (Var y) thus ?case by (cases "x = y") auto
next
  case (Fun f ss) thus ?case by (auto intro: supteq.intros)
qed

lemma term_not_arg[iff]: "Fun f ss \<notin> set ss" (is "?t \<notin> set ss")
proof
  assume "?t \<in> set ss"
  hence "?t \<rhd> ?t" by (auto intro: supt.intros)
  hence "?t \<noteq> ?t" by (auto dest: supt_neqD)
  thus False by simp
qed

lemma supt_Fun[simp]: assumes "s \<rhd> Fun f ss" (is "s \<rhd> ?t") and "s \<in> set ss" shows "False"
proof -
  from `s \<in> set ss` have "?t \<rhd> s" by (auto intro: supt.intros)
  hence "size ?t > size s" by (rule supt_size)
  from `s \<rhd> ?t` have "size s > size ?t" by (rule supt_size)
  with `size ?t > size s` show False by simp
qed

lemma supt_supteq_conv: "s \<rhd> t = (s \<unrhd> t \<and> s \<noteq> t)"
proof
  assume "s \<rhd> t" thus "s \<unrhd> t \<and> s \<noteq> t"
  proof (induct rule: supt.induct)
    case (arg t ss f) show ?case
    proof
      from arg show "Fun f ss \<unrhd> t" by (auto intro: supteq.intros)
    next
      from arg have "size t < size (Fun f ss)" by (auto simp: size_simps)
      thus "Fun f ss \<noteq> t" by auto
    qed
  next
    case (subt s ss t f)
    have "s \<unrhd> s" ..
    from `s \<in> set ss` have "Fun f ss \<unrhd> s" by (auto intro: supteq.intros)
    from `s \<unrhd> t \<and> s \<noteq> t` have "s \<unrhd> t" ..
    with `Fun f ss \<unrhd> s` have first: "Fun f ss \<unrhd> t" by (rule supteq_trans)
    from `s \<in> set ss` and `s \<rhd> t` have "Fun f ss \<rhd> t" ..
    hence second: "Fun f ss \<noteq> t" by (auto dest: supt_neqD)
    from first and second show "Fun f ss \<unrhd> t \<and> Fun f ss \<noteq> t" by auto
  qed
next
  assume "s \<unrhd> t \<and> s \<noteq> t"
  hence "s \<unrhd> t" and "s \<noteq> t" by auto
  thus "s \<rhd> t" by (induct rule: supteq.induct) (auto intro: supt.intros)
qed

lemma supt_not_sym: "s \<rhd> t \<Longrightarrow> \<not> (t \<rhd> s)"
proof
 assume "s \<rhd> t" and "t \<rhd> s" hence "s \<rhd> s" by (rule supt_trans)
 thus False unfolding supt_supteq_conv by simp
qed

lemma supt_irrefl[iff]: "\<not> t \<rhd> t"
using supt_not_sym[of t t] by auto

lemma supt_imp_supteq: "s \<rhd> t \<Longrightarrow> s \<unrhd> t"
unfolding supt_supteq_conv by auto

lemma supt_supteq_not_supteq: "s \<rhd> t = (s \<unrhd> t \<and> \<not> (t \<unrhd> s))"
using supt_not_sym unfolding supt_supteq_conv by auto

lemma supteq_supt_conv: "(s \<unrhd> t) = (s \<rhd> t \<or> s = t)" by (auto simp: supt_supteq_conv)

lemma supteq_antisym:
  assumes "s \<unrhd> t" and "t \<unrhd> s" shows "s = t"
using assms unfolding supteq_supt_conv by (auto simp: supt_not_sym)

text {* The subterm relation is an order on terms. *}
interpretation subterm: order "op \<unlhd>" "op \<lhd>"
by (unfold_locales) (rule supt_supteq_not_supteq,
  auto intro: supteq_trans supteq_antisym supt_supteq_not_supteq)

lemma suptE[elim]: "s \<rhd> t \<Longrightarrow> (s \<unrhd> t \<Longrightarrow> P) \<Longrightarrow> (s \<noteq> t \<Longrightarrow> P) \<Longrightarrow> P"
by (auto simp: supt_supteq_conv)

lemma suptI[intro]: "s \<unrhd> t \<Longrightarrow> s \<noteq> t \<Longrightarrow> s \<rhd> t"
by (induct rule: supteq.induct) (auto intro: supt.intros)

lemma supt_supteq_set_conv: "op \<rhd> = op \<unrhd> \<inter> -Id" using supt_supteq_conv[to_set] by auto

lemma supteq_supt_set_conv: "supteq = supt \<union> Id"
by (auto simp: supt_supteq_conv intro: supteq.intros)

lemmas [intro] = supteq.intros supteq.subt

lemma supteq_imp_vars_subset: "s \<unrhd> t \<Longrightarrow> vars_term t \<subseteq> vars_term s"
  by (induct rule: supteq.induct, auto)

lemma set_supteq_into_supt[simp]:
  assumes "t \<in> set ts" and "t \<unrhd> s"
  shows "Fun f ts \<rhd> s"
proof -
  from `t \<unrhd> s` have "t = s \<or> t \<rhd> s" by auto
  thus ?thesis
  proof
    assume "t = s"
    with `t \<in> set ts` show ?thesis by auto
  next
    assume "t \<rhd> s"
    from supt.subt[OF `t \<in> set ts` this] show ?thesis .
  qed
qed

declare conj_cong[fundef_cong]
fun
  poss :: "('f, 'v) term \<Rightarrow> pos set"
where
  "poss (Var x) = {\<epsilon>}" |
  "poss (Fun f ss) = {\<epsilon>} \<union> {i<#p | i p. i < length ss \<and> p \<in> poss (ss!i)}"
declare conj_cong[fundef_cong del]

text {* The subterm relation is strongly normalizing. *}
lemma SN_supt: "SN (op \<rhd>)" unfolding SN_iff_wf
by (rule wf_subset) (auto intro: supt_size)

text {* The subterm of a term~@{text s} at position~@{text p} is defined as follows: *}
fun
  subt_at :: "('f, 'v) term \<Rightarrow> pos \<Rightarrow> ('f, 'v) term" ("(_ |'_ _)" [60])
where
  "s |_ \<epsilon> = s"
| "Fun f ss |_ i<#p = ss!i|_p"

lemma pos_append_poss: "p \<in> poss t \<Longrightarrow> q \<in> poss (t |_ p) \<Longrightarrow> p <#> q \<in> poss t"
proof (induct t arbitrary: p q)
  case (Var x)
  thus ?case by auto
next
  case (Fun f ts p q)
  show ?case
  proof (cases p)
    case Empty thus ?thesis using Fun by auto
  next
    case (PCons i p')
    with Fun have i: "i < length ts" and p': "p' \<in> poss (ts ! i)" by auto
    from i have mem: "ts ! i \<in> set ts" by auto
    from Fun(3) have "q \<in> poss (ts ! i |_ p')" unfolding PCons using i by auto
    from Fun(1)[OF mem p' this]
    show ?thesis unfolding PCons using i by auto
  qed
qed


text {* A \emph{context} is a term containing exactly one \emph{hole}. *}
datatype ('f, 'v) ctxt =
  Hole ("\<box>")
| More 'f "('f, 'v) term list" "('f, 'v) ctxt" "('f, 'v) term list"


text {*
Replacing the hole in a context~@{text C} by a term ~@{text t} is called, application
of @{text C} to @{text t}.
*}
fun
  ctxt_apply :: "('f, 'v) ctxt \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term" ("_\<langle>_\<rangle>" [1000, 0] 1000)
where
  "\<box>\<langle>s\<rangle> = s" |
  "(More f ss1 C ss2)\<langle>s\<rangle> = Fun f (ss1 @ C\<langle>s\<rangle> # ss2)"

lemma ctxt_eq[simp]: "(C\<langle>s\<rangle> = C\<langle>t\<rangle>) = (s = t)" by (induct C) auto

fun
  ctxt_of_pos_term :: "pos \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) ctxt"
where
  "ctxt_of_pos_term \<epsilon> t = \<box>" |
  "ctxt_of_pos_term (i<#ps) (Fun f ts) =
    More f (take i ts) (ctxt_of_pos_term ps (ts!i)) (drop (Suc i) ts)"

lemma ctxt_supt_id:
  assumes "p \<in> poss t" shows "(ctxt_of_pos_term p t)\<langle>t|_p\<rangle> = t"
using assms by (induct t arbitrary: p) (auto simp: id_take_nth_drop[symmetric])

text {*
Let @{text s} and @{text t} be terms. The following three statements are equivalent:
\begin{enumerate}
 \item \label{A}@{text "s \<unrhd> t"}
 \item \label{B}@{text "\<exists>p\<in>poss s. s|_p = t"}
 \item \label{C}@{text "\<exists>C. s = C\<langle>t\<rangle>"}
\end{enumerate}
*}

fun
  ctxt_compose :: "('f, 'v) ctxt \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f, 'v) ctxt" (infixl "\<circ>c" 65)
where
  "\<box> \<circ>c D = D" |
  "(More f ss1 C ss2) \<circ>c D = More f ss1 (C \<circ>c D) ss2"

interpretation ctxt_monoid_mult: monoid_mult "\<box>" "op \<circ>c"
proof
  fix C D E show "C \<circ>c D \<circ>c E = C \<circ>c (D \<circ>c E)" by (induct C) simp_all
next
  fix C show "\<box> \<circ>c C = C" by simp
next
  fix C show "C \<circ>c \<box> = C" by (induct C) simp_all
qed

declare ctxt_monoid_mult.mult_assoc[simp]

instantiation ctxt :: (type, type) monoid_mult begin
  definition [simp]: "1 \<equiv> \<box>"
  definition [simp]: "op * = op \<circ>c"
  instance by (intro_classes) simp_all
end

lemma ctxt_compose_apply_distrib[simp]: "(C \<circ>c D)\<langle>t\<rangle> = C\<langle>D\<langle>t\<rangle>\<rangle>" by (induct C) simp_all

text {* The position of the hole in a context is uniquely determined. *}
fun
  hole_pos :: "('f, 'v) ctxt \<Rightarrow> pos"
where
  "hole_pos \<box> = \<epsilon>" |
  "hole_pos (More f ss D ts) = length ss <# hole_pos D"

lemma hole_pos_ctxt_of_pos_term: assumes "p \<in> poss t"
  shows "hole_pos (ctxt_of_pos_term p t) = p"
  using assms
proof (induct t arbitrary: p, simp)
  case (Fun f ts)
  show ?case 
  proof (cases p)
    case Empty
    thus ?thesis using Fun by auto
  next
    case (PCons i q)
    with Fun(2) have i: "i < length ts" and q: "q \<in> poss (ts ! i)" by auto
    hence "ts ! i \<in> set ts" by auto
    from Fun(1)[OF this q] PCons i show ?thesis by simp
  qed
qed

lemma hole_pos_id_ctxt: assumes "C\<langle>s\<rangle> = t"
  shows "ctxt_of_pos_term (hole_pos C) t = C"
using assms 
proof (induct C arbitrary: t, simp)
  case (More f bef C aft)
  thus ?case
  proof (cases t, simp)
    case (Fun g ts)
    with More have [simp]: "g = f" by simp
    from Fun More have bef: "take (length bef) ts = bef" by auto    
    from Fun More have aft: "drop (Suc (length bef)) ts = aft" by auto
    from Fun More have Cs: "C\<langle>s\<rangle> = ts ! length bef" by auto
    from Fun More show ?thesis by (simp add: bef aft More(1)[OF Cs])
  qed
qed

lemma supteq_imp_subt_at:
  assumes "s \<unrhd> t" shows "\<exists>p\<in>poss s. s|_p = t"
using assms proof (induct s t rule: supteq.induct)
  case (refl s)
  have "\<epsilon> \<in> poss s" by (induct s rule: term_induct) auto
  have "s|_\<epsilon> = s" by simp
  from `\<epsilon> \<in> poss s` and `s|_\<epsilon> = s` show ?case by best
next
  case (subt u ss t f) 
  then obtain p where "p \<in> poss u" and "u|_p = t" by best
  from `u \<in> set ss` obtain i
    where "i < length ss" and "u = ss!i" by (auto simp: in_set_conv_nth)
  from `i < length ss` and `p \<in> poss u`
    have "i<#p \<in> poss (Fun f ss)" unfolding `u = ss!i` by simp
  have "Fun f ss|_i<#p = t"
    unfolding subt_at.simps unfolding `u = ss!i`[symmetric] by (rule `u|_p = t`)
  with `i<#p \<in> poss (Fun f ss)` show ?case by best
qed

lemma subt_at_imp_ctxt:
  assumes "p \<in> poss s" shows "\<exists>C. C\<langle>s|_p\<rangle> = s"
using assms proof (induct p arbitrary: s)
  case (Empty s)
  have "\<box>\<langle>s|_\<epsilon>\<rangle> = s" by simp
  thus ?case by best
next
  case (PCons i p s)
  then obtain f ss where "s = Fun f ss" by (cases s) auto
  with `i<#p \<in> poss s` obtain u :: "('a,'b) term"
    where "u = ss!i" and "p \<in> poss u" and "i < length ss" by auto
  from PCons and `p\<in>poss u` obtain D where "D\<langle>u|_p\<rangle> = u" by auto
  let ?ss1 = "take i ss" and ?ss2 = "drop (Suc i) ss"
  let ?E = "More f ?ss1 D ?ss2"
  have "?ss1@D\<langle>u|_p\<rangle>#?ss2 = ss" (is "?ss = ss") unfolding `D\<langle>u|_p\<rangle> = u` unfolding `u = ss!i`
    unfolding id_take_nth_drop[OF `i < length ss`, symmetric] ..
  have "s|_i<#p = u|_p" unfolding `s = Fun f ss` using `u = ss!i` by simp
  have "?E\<langle>s|_i<#p\<rangle> = s"
    unfolding ctxt_apply.simps `s|_i<#p = u|_p` `?ss = ss` unfolding `s = Fun f ss` ..
  thus ?case by best
qed

lemma ctxt_imp_supteq[simp]: shows "C\<langle>t\<rangle> \<unrhd> t" by (induct C) auto

lemma subt_at_imp_supteq': assumes "p \<in> poss s" and "s|_p = t" shows "s \<unrhd> t"
proof -
 from `p \<in> poss s` obtain C where "C\<langle>s|_p\<rangle> = s" using subt_at_imp_ctxt by best
 thus ?thesis unfolding `s|_p = t` using ctxt_imp_supteq by auto
qed

lemma subt_at_imp_supteq: assumes "p \<in> poss s" shows "s \<unrhd> s|_p"
proof -
 have "s|_p = s|_p" by auto
 with assms show ?thesis by (rule subt_at_imp_supteq')
qed



(* Conversions between contexts and subterms. *)
lemma supteq_ctxtE[elim]: assumes "s \<unrhd> t" obtains C where "s = C\<langle>t\<rangle>"
using assms proof (induct arbitrary: thesis)
  case (refl s)
  have "s = \<box>\<langle>s\<rangle>" by simp
  from refl[OF this] show ?case .
next
  case (subt u ss t f)
  then obtain C where "u = C\<langle>t\<rangle>" by auto
  from split_list[OF `u \<in> set ss`] obtain ss1 and ss2 where "ss = ss1 @ u # ss2" by auto
  hence "Fun f ss = (More f ss1 C ss2)\<langle>t\<rangle>" using `u = C\<langle>t\<rangle>` by simp
  with subt show ?case by best
qed

lemma ctxt_supteq[intro]: assumes "s = C\<langle>t\<rangle>" shows "s \<unrhd> t"
proof (cases C)
  case Hole thus ?thesis using assms by auto
next
  case (More f ss1 D ss2)
  with assms have s: "s = Fun f (ss1 @ D\<langle>t\<rangle> # ss2)" (is "_ = Fun _ ?ss") by simp
  have "D\<langle>t\<rangle> \<in> set ?ss" by simp
  moreover have "D\<langle>t\<rangle> \<unrhd> t" by (induct D) auto
  ultimately show ?thesis unfolding s ..
qed

lemma supteq_ctxt_conv: "(s \<unrhd> t) = (\<exists>C. s = C\<langle>t\<rangle>)" by auto

(* Conversions between contexts and proper subterms. *)
lemma supt_ctxtE[elim]: assumes "s \<rhd> t" obtains C where "C \<noteq> \<box>" and "s = C\<langle>t\<rangle>"
using assms proof (induct arbitrary: thesis)
  case (arg s ss f)
  from split_list[OF `s \<in> set ss`] obtain ss1 and ss2 where ss: "ss = ss1 @ s # ss2" by auto
  let ?C = "More f ss1 \<box> ss2"
  have "?C \<noteq> \<box>" by simp
  moreover have "Fun f ss = ?C\<langle>s\<rangle>" by (simp add: ss)
  ultimately show ?case using arg by best
next
  case (subt s ss t f)
  then obtain C where "C \<noteq> \<box>" and "s = C\<langle>t\<rangle>" by auto
  from split_list[OF `s \<in> set ss`] obtain ss1 and ss2 where ss: "ss = ss1 @ s # ss2" by auto
  have "More f ss1 C ss2 \<noteq> \<box>" by simp
  moreover have "Fun f ss = (More f ss1 C ss2)\<langle>t\<rangle>" using `s = C\<langle>t\<rangle>` by (simp add: ss)
  ultimately show ?case using subt(4) by best
qed

(* More transitivity rules. *)
lemma supt_supteq_trans[trans]:
  assumes "s \<rhd> t" and "t \<unrhd> u" shows "s \<rhd> u"
proof (cases "t = u")
  assume "t = u" with assms show ?thesis by simp
next
  assume "t \<noteq> u"
  with assms have "t \<rhd> u" by auto
  with `s \<rhd> t` show ?thesis by (rule supt_trans)
qed

lemma supteq_supt_trans[trans]:
  assumes "s \<unrhd> t" and "t \<rhd> u" shows "s \<rhd> u"
proof (cases "s = t")
  assume "s = t" with assms show ?thesis by simp
next
  assume "s \<noteq> t"
  with assms have "s \<rhd> t" by auto
  thus ?thesis using `t \<rhd> u` by (rule supt_trans)
qed

lemma ctxt_supt[intro 2]: assumes "C \<noteq> \<box>" and "s = C\<langle>t\<rangle>" shows "s \<rhd> t"
proof (cases C)
  case Hole with assms show ?thesis by simp
next
  case (More f ss1 D ss2)
  with assms have s: "s = Fun f (ss1 @ D\<langle>t\<rangle> # ss2)" by simp
  have "D\<langle>t\<rangle> \<in> set (ss1 @ D\<langle>t\<rangle> # ss2)" by simp
  hence "s \<rhd> D\<langle>t\<rangle>" unfolding s ..
  also have "D\<langle>t\<rangle> \<unrhd> t" by (induct D) auto
  finally show "s \<rhd> t" .
qed

lemma supt_not_refl[elim!]: assumes "t \<rhd> t" shows False
proof -
  from assms have "t \<noteq> t" by auto
  thus False by simp
qed

lemma supt_ctxt_conv: "(s \<rhd> t) = (\<exists>C. C \<noteq> \<box> \<and> s = C\<langle>t\<rangle>)" (is "_ = ?rhs")
proof
  assume "s \<rhd> t"
  hence "s \<unrhd> t" and "s \<noteq> t" by auto
  from `s \<unrhd> t` obtain C where "s = C\<langle>t\<rangle>" by auto
  with `s \<noteq> t` have "C \<noteq> \<box>" by auto
  with `s = C\<langle>t\<rangle>` show "?rhs" by auto
next
  assume "?rhs" thus "s \<rhd> t" by auto
qed

text {*
By adding \emph{non-empty} to statements \ref{B} and \ref{C} a similar characterisation for
proper subterms is obtained:
\begin{enumerate}
 \item @{text "s \<rhd> t"}
 \item @{text "\<exists>i p. i#p\<in>poss s \<and> s|_i<#p = t"}
 \item @{text "\<exists>C. C \<noteq> \<box> \<and> s = C\<langle>t\<rangle>"}
\end{enumerate}
*}

lemma supt_imp_subt_at_nepos: assumes "s \<rhd> t" shows "\<exists>i p. i<#p\<in>poss s \<and> s|_i<#p = t"
proof -
  from assms have "s \<unrhd> t" and "s \<noteq> t" unfolding supt_supteq_conv by auto
  then obtain p where supteq: "p \<in> poss s" "s|_p = t" using supteq_imp_subt_at by best
  have "p \<noteq> \<epsilon>"
  proof
    assume "p = \<epsilon>" hence "s = t" using `s|_p = t` by simp
    thus False using `s \<noteq> t` by simp
  qed
  then obtain i q where "p = i<#q" by (cases p) simp
  with supteq show ?thesis by auto
qed


lemma arg_neq:
  assumes "i < length ss" and "ss!i = Fun f ss" shows "False"
proof -
  from `i < length ss` have "(ss!i) \<in> set ss" by auto
  with assms show False by simp
qed

lemma supteq_not_supt[elim]: assumes "s \<unrhd> t" and "\<not> (s \<rhd> t)" shows "s = t"
using assms by (induct) auto

lemma supteq_not_supt_conv[simp]: "op \<unrhd> - op \<rhd> = Id" by auto

lemma subt_at_nepos_neq: assumes "i<#p \<in> poss s" shows "s|_(i<#p) \<noteq> s"
proof (cases s)
  fix x assume "s = Var x"
  hence "i<#p \<notin> poss s" by simp
  with assms show ?thesis by simp
next
  fix f ss assume "s = Fun f ss" show ?thesis
  proof
    assume "s|_i<#p = s"
    from assms have "i < length ss" unfolding `s = Fun f ss` by auto
    hence "ss!i \<in> set ss" by simp
    hence "Fun f ss \<rhd> ss!i" by (rule supt.arg)
    hence "Fun f ss \<noteq> ss!i" unfolding supt_supteq_conv by simp
    from `s|_i<#p = s` and assms
      have "ss!i \<unrhd> Fun f ss" using subt_at_imp_supteq' unfolding `s = Fun f ss` by auto
    with supt_not_sym[OF `Fun f ss \<rhd> ss!i`] have "ss!i = Fun f ss" by auto
    with `i < length ss` show False by (rule arg_neq)
  qed
qed

lemma subt_at_nepos_imp_supt: assumes "i<#p \<in> poss s" shows "s \<rhd> s|_i<#p"
proof -
  from assms have "s \<unrhd> s|_i<#p" by (rule subt_at_imp_supteq)
  from assms have "s|_i<#p \<noteq> s" by (rule subt_at_nepos_neq)
  from `s \<unrhd> s|_i<#p` and `s|_i<#p \<noteq> s` show ?thesis by (auto simp: supt_supteq_conv)
qed

lemma subt_at_nepos_imp_nectxt:
  assumes "i<#p \<in> poss s" and "s|_i<#p = t" shows "\<exists>C. C \<noteq> \<box> \<and> C\<langle>t\<rangle> = s"
proof -
  from assms obtain C where "C\<langle>s|_i<#p\<rangle> = s" using subt_at_imp_ctxt by best
  from `i<#p \<in> poss s`
    have "t \<noteq> s" unfolding `s|_i<#p = t`[symmetric] using subt_at_nepos_neq by best
  from assms and `C\<langle>s|_i<#p\<rangle> = s` have "C\<langle>t\<rangle> = s" by simp
  have "C \<noteq> \<box>"
  proof
    assume "C = \<box>"
    with `C\<langle>t\<rangle> = s` have "t = s" by simp
    with `t \<noteq> s` show False by simp
  qed
  with `C\<langle>t\<rangle> = s` show ?thesis by auto
qed

lemma nectxt_imp_supt_ctxt: "C \<noteq> \<box> \<Longrightarrow> C\<langle>t\<rangle> \<rhd> t"
proof (induct C)
  case Hole thus ?case by auto
next
  case (More f ls C rs) thus ?case
  proof (cases "C = \<box>")
    case True
    hence a: "(More f ls C rs)\<langle>t\<rangle> = Fun f (ls @ t#rs)" by auto
    hence "t \<in> set (ls @ t#rs)" by auto
    hence "Fun f (ls@t#rs) \<rhd> t" by (rule supt.arg)
    with a show ?thesis by auto
  next
    case False
    from `C \<noteq> \<box>` and `C \<noteq> \<box> \<Longrightarrow> C\<langle>t\<rangle> \<rhd> t` have b: "C\<langle>t\<rangle> \<rhd> t" by simp
    have c: "(More f ls C rs)\<langle>t\<rangle> = Fun f (ls @ C\<langle>t\<rangle>#rs)" by auto
    have d: "C\<langle>t\<rangle> \<in> set (ls @ C\<langle>t\<rangle>#rs)" by auto
    have "Fun f (ls @ C\<langle>t\<rangle>#rs) \<rhd> t"
      using supt.arg[where s = "C\<langle>t\<rangle>" and ss = "ls@C\<langle>t\<rangle>#rs",OF d] b by (rule supt_trans)
    with c show ?thesis by auto
  qed
qed

lemma supt_var: "\<not> (Var x \<rhd> u)"
proof
  assume "Var x \<rhd> u"
  then obtain C where "C \<noteq> \<box>" and "Var x = C\<langle>u\<rangle>" ..
  thus False by (cases C) auto
qed

lemma supt_const: "\<not> (Fun f [] \<rhd> u)"
proof
  assume "Fun f [] \<rhd> u"
  then obtain C where "C \<noteq> \<box>" and "Fun f [] = C\<langle>u\<rangle>" ..
  thus False by (cases C) auto
qed

lemma supteq_var_imp_eq:
  "(Var x \<unrhd> t) = (t = Var x)" (is "_ = (_ = ?x)")
proof
  assume "t = Var x" thus "Var x \<unrhd> t" by auto
next
  assume st: "?x \<unrhd> t"
  from st obtain C where "?x = C\<langle>t\<rangle>" by best
  thus "t = ?x" by (cases C) auto
qed

lemma Var_supt[elim!]: assumes "Var x \<rhd> t" shows P
using assms supt_var[of x t] by simp

lemma Fun_supt[elim]: assumes "Fun f ts \<rhd> s" obtains t where "t \<in> set ts" and "t \<unrhd> s"
using assms by (cases) (auto simp: supt_supteq_conv)

lemma subterm_induct:
  assumes "\<And>t. \<forall>s\<lhd>t. P s \<Longrightarrow> P t"
  shows [case_names subterm]: "P t"
proof -
  have "\<forall>s\<unlhd>t. P s"
  proof (induct t)
    case (Var x)
    have "\<forall>s\<lhd>Var x. P s" by blast
    with assms show ?case by blast
  next
    case (Fun f ts)
    have "\<forall>s\<lhd>Fun f ts. P s"
    proof (intro allI impI)
      fix s assume "Fun f ts \<rhd> s"
      then obtain t where "t \<in> set ts" and "t \<unrhd> s" ..
      with Fun show "P s" by simp
    qed
    with assms show ?case by blast
  qed
  thus ?thesis by auto
qed



text {*A \emph{substitution} is a mapping~$\sigma$ from variables to terms. We call a
substitution that alters the type of variables a \emph{general substitution}, since it
does not have all properties that are expected of (standard) substitutions (e.g., there
is no empty substitution.*}
type_synonym ('f, 'v, 'w) gsubst = "'v \<Rightarrow> ('f, 'w) term"
type_synonym ('f, 'v) subst  = "('f, 'v, 'v) gsubst"

fun
  subst_apply_term :: "('f, 'v) term \<Rightarrow> ('f, 'v, 'w) gsubst \<Rightarrow> ('f, 'w) term"
  (infixl "\<cdot>" 60)
where
  "(Var x) \<cdot> \<sigma> = \<sigma>(x)"
| "(Fun f ss) \<cdot> \<sigma> = Fun f (map (\<lambda>t. t \<cdot> \<sigma>) ss)"

definition
  subst_compose :: "('f, 'u, 'v) gsubst \<Rightarrow> ('f, 'v, 'w) gsubst \<Rightarrow> ('f, 'u, 'w) gsubst"
  (infixl "\<circ>s" 65)
where
  "\<sigma> \<circ>s \<tau> = (\<lambda>x. (\<sigma> x) \<cdot> \<tau>)"

lemma subst_compose_apply_term_distrib[simp]: "t \<cdot> (\<sigma> \<circ>s \<tau>) = t \<cdot> \<sigma> \<cdot> \<tau>"
  by (induct t \<sigma> rule: subst_apply_term.induct) (simp_all add: subst_compose_def)

lemma subst_compose_assoc[simp]: "\<sigma> \<circ>s \<tau> \<circ>s \<mu> = \<sigma> \<circ>s (\<tau> \<circ>s \<mu>)"
proof (rule ext)
  fix x show "(\<sigma> \<circ>s \<tau> \<circ>s \<mu>) x = (\<sigma> \<circ>s (\<tau> \<circ>s \<mu>)) x"
  proof -
    have "(\<sigma> \<circ>s \<tau> \<circ>s \<mu>) x = \<sigma>(x) \<cdot> \<tau> \<cdot> \<mu>" by (simp add: subst_compose_def)
    also have "\<dots> = \<sigma>(x) \<cdot> (\<tau> \<circ>s \<mu>)" by simp
    finally show ?thesis by (simp add: subst_compose_def)
  qed
qed

lemma subst_apply_term_empty[simp]: "t \<cdot> Var = t"
proof (induct t)
  case (Var x) show ?case by simp
next
  case (Fun f ts)
  from map_ext[OF Fun[COMP impI]] show ?case by simp
qed

interpretation subst_monoid_mult: monoid_mult "Var" "op \<circ>s"
  by (unfold_locales) (simp, simp_all add: subst_compose_def)

lemma subt_at_subst: assumes "p \<in> poss s" shows "(s \<cdot> \<sigma>) |_ p = (s |_ p) \<cdot> \<sigma>"
  using assms
proof (induct s arbitrary: p)
  case (Var x)
  thus ?case by auto
next
  case (Fun f ts p)
  show ?case
  proof (cases p)
    case Empty
    thus ?thesis by auto
  next
    case (PCons i p')
    with Fun(2) have i: "i < length ts" and p': "p' \<in> poss (ts ! i)" by auto
    from i have ti: "ts ! i \<in> set ts" by auto
    show ?thesis using Fun(1)[OF ti p'] i PCons by auto
  qed
qed

definition subst_domain :: "('f, 'v)subst \<Rightarrow> 'v set" where
  "subst_domain \<sigma> = {x. \<sigma>(x) \<noteq> (Var x)}"

fun subst_range :: "('f, 'v) subst \<Rightarrow> ('f, 'v) term set" where
  "subst_range \<sigma> = \<sigma> ` (subst_domain \<sigma>)"

definition
  subst_restrict :: "('f, 'v) subst \<Rightarrow> 'v set \<Rightarrow> ('f, 'v) subst"
  (infix "|s" 60)
where
  "\<sigma> |s V \<equiv> (\<lambda>x. if x \<in> V then \<sigma>(x) else Var x)"

lemma subst_domain_Var[simp]: "subst_domain Var = {}" by (simp add: subst_domain_def)

lemma subst_restrict_Int[simp]: "(\<sigma> |s V )|s W = \<sigma> |s (V \<inter> W)"
  by (rule ext) (simp add: subst_restrict_def)

lemma subst_domain_Var_conv[iff]: "subst_domain \<sigma> = {} \<longleftrightarrow> \<sigma> = Var"
proof
  assume "subst_domain \<sigma> = {}"
  show "\<sigma> = Var"
  proof (rule ext)
  fix x show "\<sigma>(x) = Var x"
    proof (rule ccontr)
      assume "\<sigma>(x) \<noteq> Var x"
      hence "x \<in> subst_domain \<sigma>" by (simp add: subst_domain_def)
      with `subst_domain \<sigma> = {}` show False by simp
    qed
  qed
next
  assume "\<sigma> = Var" thus "subst_domain \<sigma> = {}" by simp
qed

fun
  subst_apply_ctxt :: "('f, 'v) ctxt \<Rightarrow> ('f, 'v, 'w) gsubst \<Rightarrow> ('f, 'w) ctxt"
  (infixl "\<cdot>c" 60)
where
  "\<box> \<cdot>c \<sigma> = \<box>"
| "(More f ss1 D ss2) \<cdot>c \<sigma> = More f (map (\<lambda>t. t \<cdot> \<sigma>) ss1) (D \<cdot>c \<sigma>) (map (\<lambda>t. t \<cdot> \<sigma>) ss2)"

lemma subst_apply_term_ctxt_apply_distrib[simp]: "C\<langle>t\<rangle> \<cdot> \<mu> = (C \<cdot>c \<mu>)\<langle>t \<cdot> \<mu>\<rangle>"
  by (induct C) auto

lemma subst_compose_ctxt_compose_distrib[simp]: "(C \<circ>c D) \<cdot>c \<sigma> = (C \<cdot>c \<sigma>) \<circ>c (D \<cdot>c \<sigma>)"
  by (induct C) auto

lemma ctxt_compose_subst_compose_distrib[simp]: "C \<cdot>c (\<sigma> \<circ>s \<tau>) = (C \<cdot>c \<sigma>) \<cdot>c \<tau>"
  by (induct C) auto

lemma subst_compose_Var[simp]: "\<sigma> \<circ>s Var = \<sigma>" by (simp add: subst_compose_def)

lemma Var_subst_compose[simp]: "Var \<circ>s \<sigma> = \<sigma>" by (simp add: subst_compose_def)

text {*We use the same logical constant as for the power operations on
functions and relations, in order to share their syntax.*}
overloading
  substpow \<equiv> "compow :: nat \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) subst"
begin

primrec substpow :: "nat \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) subst" where
  "substpow 0 \<sigma> = Var"
| "substpow (Suc n) \<sigma> = \<sigma> \<circ>s substpow n \<sigma>"

end

lemma subst_power_compose_distrib:
  "\<mu> ^^ (m + n) = (\<mu> ^^ m \<circ>s \<mu> ^^ n)" by (induct m) simp_all

lemma subst_power_Suc: "\<mu> ^^ (Suc i) = \<mu> ^^ i \<circ>s \<mu>"
proof -
  have "\<mu> ^^ (Suc i) = \<mu> ^^ (i + Suc 0)" by simp
  thus ?thesis unfolding  subst_power_compose_distrib by simp
qed

lemma subst_pow_mult: "((\<sigma> :: ('f,'v)subst)^^ n) ^^ m = \<sigma> ^^ (n * m)"
  by (induct m arbitrary: n, auto simp: subst_power_compose_distrib)


lemma subst_domain_pow: "subst_domain (\<sigma> ^^ n) \<subseteq> subst_domain \<sigma>"
  unfolding subst_domain_def
  by (induct n, auto simp: subst_compose_def)

lemma subt_at_PCons_distr[simp]:
 assumes "i<#p \<in> poss t" shows "t|_i<#p = (t|_<i>)|_p" using assms by (induct t) auto

lemma subt_at_append[simp]: "p \<in> poss t \<Longrightarrow> t |_ (p <#> q) = (t |_ p) |_ q"
proof (induct t arbitrary: p)
  case (Var x) thus ?case by auto
next
  case (Fun f ts)
  show ?case 
  proof (cases p)
    case Empty
    thus ?thesis by auto
  next
    case (PCons i p')
    with Fun(2) have i: "i < length ts" and p': "p' \<in> poss (ts ! i)" by auto
    from i have ti: "ts ! i \<in> set ts" by auto
    show ?thesis using Fun(1)[OF ti p'] unfolding PCons by auto
  qed
qed

lemma empty_pos_in_poss[simp]: "\<epsilon> \<in> poss t" by (induct t) auto

lemma poss_append_poss[simp]: "(p <#> q \<in> poss t) = (p \<in> poss t \<and> q \<in> poss (t |_ p))" (is "?l = ?r")
proof
  assume ?r
  with pos_append_poss[of p t q] show ?l by auto
next
  assume ?l
  thus ?r
  proof (induct p arbitrary: t)
    case Empty
    thus ?case by auto
  next
    case (PCons i p t)
    then obtain f ts where t: "t = Fun f ts" by (cases t, auto)
    note IH = PCons[unfolded t]
    from IH(2) have i: "i < length ts" and rec: "p <#> q \<in> poss (ts ! i)" by auto
    from IH(1)[OF rec] i show ?case unfolding t by auto
  qed
qed  

lemma poss_imp_subst_poss[simp]: "p \<in> poss t \<Longrightarrow> p \<in> poss(t\<cdot>\<sigma>)" by (induct t arbitrary: p) auto

lemma iterate_term: 
  assumes id: "t \<cdot> \<sigma> |_ p = t" and pos: "p \<in> poss (t \<cdot> \<sigma>)"
  shows "t \<cdot> \<sigma> ^^ n |_ (p^n) = t \<and> p ^ n \<in> poss (t \<cdot> \<sigma> ^^ n)"
proof (induct n)
  case 0 show ?case by simp
next
  case (Suc n)
  hence p: "p ^ n \<in> poss (t \<cdot> \<sigma> ^^ n)" by simp
  note p' = poss_imp_subst_poss[OF p]
  note p'' = subt_at_append[OF p']
  have idt: "t \<cdot> \<sigma> ^^ (Suc n) = t \<cdot> \<sigma>^^ n \<cdot> \<sigma>" unfolding subst_power_Suc by simp
  have "t \<cdot> \<sigma> ^^ (Suc n) |_ (p ^ Suc n)
    = t \<cdot> \<sigma> ^^ n \<cdot> \<sigma> |_ (p ^ n) <#> p" unfolding idt power_pos_Suc ..
  also have "... = ((t \<cdot> \<sigma> ^^ n |_ p ^ n) \<cdot> \<sigma>) |_ p" unfolding p'' subt_at_subst[OF p] ..
  also have "... = t \<cdot> \<sigma> |_ p" unfolding Suc[THEN conjunct1] ..
  also have "... = t" unfolding id ..
  finally have one: "t \<cdot> \<sigma> ^^ Suc n |_ (p ^ Suc n) = t" .
  show ?case
  proof (rule conjI[OF one])
    show "p ^ Suc n \<in> poss (t \<cdot> \<sigma> ^^ Suc n)"
      unfolding power_pos_Suc poss_append_poss idt
    proof (rule conjI[OF poss_imp_subst_poss[OF p]])
      have "t \<cdot> \<sigma> ^^ n \<cdot> \<sigma> |_ p ^ n = t \<cdot> \<sigma> ^^ n |_ p ^ n \<cdot> \<sigma>"
        by (rule subt_at_subst[OF p])
      also have "... = t \<cdot> \<sigma>" using Suc by simp
      finally show "p \<in> poss t \<cdot> \<sigma> ^^ n \<cdot> \<sigma> |_ p ^ n" using pos by auto
    qed
  qed
qed

lemma hole_pos_poss[simp]: "hole_pos C \<in> poss(C\<langle>t\<rangle>)" by (induct C) auto

lemma hole_pos_poss_conv: "((hole_pos C <#> q) \<in> poss(C\<langle>t\<rangle>)) = (q \<in> poss t)"
by (induct C) auto

lemma subt_at_hole_pos[simp]: "C\<langle>t\<rangle>|_(hole_pos C) = t" by (induct C) auto

lemma hole_pos_power_poss[simp]: "(hole_pos C)^(n::nat) \<in> poss((C^n)\<langle>t\<rangle>)"
  by (induct n) (auto simp: hole_pos_poss_conv one_ctxt_def one_pos_def times_ctxt_def times_pos_def)

lemma subst_subt_at_distr[simp]: assumes "p \<in> poss t" shows "t|_p \<cdot> \<sigma> = (t \<cdot> \<sigma>)|_p"
using assms by (induct t arbitrary: p) auto



lemma poss_imp_ctxt_subst_poss[simp]: assumes "p \<in> poss(C\<langle>t\<rangle>)" shows "p \<in> poss((C \<cdot>c \<sigma>)\<langle>t \<cdot> \<sigma>\<rangle>)"
proof -
 have "p \<in> poss(C\<langle>t\<rangle> \<cdot> \<sigma>)" by (rule poss_imp_subst_poss[OF assms])
 thus ?thesis by simp
qed

lemma poss_PCons_poss[simp]: "(i <# p \<in> poss t) = (i < length (args t) \<and> p \<in> poss (args t ! i))"
  by (cases t, auto)

lemma less_pos_imp_supt: assumes less: "p' < p" and p: "p \<in> poss t"
  shows "t |_ p \<lhd> t |_ p'"
proof -
  from less obtain p'' where p'': "p = p' <#> p''" unfolding less_pos_def less_eq_pos_def by auto
  with less have ne: "p'' \<noteq> \<epsilon>" by auto
  then obtain i q where ne: "p'' = i <# q" by (cases p'', auto)
  from p have p': "p' \<in> poss t" unfolding p'' by simp
  from p have "p'' \<in> poss (t |_ p')" unfolding p''  by simp
  from subt_at_nepos_imp_supt[OF this[unfolded ne]] have "t |_ p' \<rhd> t |_ p' |_ p''" unfolding ne by simp
  thus "t |_ p' \<rhd> t |_ p" unfolding p'' subt_at_append[OF p'] .
qed

lemma less_eq_pos_imp_supt_eq: assumes less_eq: "p' \<le> p" and p: "p \<in> poss t"
  shows "t |_ p \<unlhd> t |_ p'"
proof -
  from less_eq obtain p'' where p'': "p = p' <#> p''" unfolding less_eq_pos_def by auto
  from p have p': "p' \<in> poss t" unfolding p'' by simp
  from p have "p'' \<in> poss (t |_ p')" unfolding p'' by simp
  from subt_at_imp_supteq[OF this] have "t |_ p' \<unrhd> t |_ p' |_ p''" by simp
  thus "t |_ p' \<unrhd> t |_ p" unfolding p'' subt_at_append[OF p'] .
qed




lemma funas_term_poss_conv: "funas_term t = {(f, length ts) | p f ts. p \<in> poss t \<and> t|_p = Fun f ts}"
proof (induct t)
  case (Fun f ts)
  let ?f = "\<lambda> f ts. (f,length ts)"
  let ?fs = "\<lambda> t. {?f f ts | p f ts. p \<in> poss t \<and> t|_p = Fun f ts}"
  let ?l = "funas_term (Fun f ts)"
  let ?r = "?fs (Fun f ts)" 
  {
    fix gn
    have "(gn \<in> ?l) = (gn \<in> ?r)"
    proof (cases "gn = (f,length ts)")
      case True
      thus ?thesis by force
    next
      case False
      obtain g n where gn: "gn = (g,n)" by force
      have "(gn \<in> ?l) = (\<exists> t \<in> set ts. gn \<in> funas_term t)" using False by auto
      also have "... = (\<exists> i < length ts. gn \<in> funas_term (ts ! i))" unfolding set_conv_nth by auto
      also have "... = (\<exists> i < length ts. (g,n) \<in> ?fs (ts ! i))" using Fun[unfolded set_conv_nth] gn by blast
      also have "... = ((g,n) \<in> ?fs (Fun f ts))" (is "?l' = ?r'")
      proof 
        assume ?l' 
        then obtain i p ss where p: "p \<in> poss (ts ! i)" "ts ! i |_ p = Fun g ss" "n = length ss" "i < length ts" by auto
        show ?r'
          by (rule, rule exI[of _ "i <# p"], intro exI conjI, unfold p(3), rule refl, insert p(1) p(2) p(4), auto)
      next
        assume ?r'
        then obtain p ss where p: "p \<in> poss (Fun f ts)" "Fun f ts |_ p = Fun g ss" "n = length ss" by auto
        from p False gn obtain i p' where pp: "p = i <# p'" by (cases p, auto)
        show ?l'
          by (rule exI[of _ i], insert p pp, auto)
      qed
      finally show ?thesis unfolding gn .
    qed
  }
  thus ?case by blast
qed simp


definition
  subst :: "'v \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) subst" ("({_/ '/ _})") 
where
  "{x/t} = (\<lambda>y. if x = y then t else Var y)"

lemma Var_subst[simp]: "Var x \<cdot> {x/t} = t" by (simp add: subst_def)

inductive
  subst_instance :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" ("(_/ \<preceq> _)" [56, 56] 55)
where
  subst_instanceI[intro]: "s \<cdot> \<sigma> = t \<Longrightarrow> s \<preceq> t"

lemma subst_instance_trans[trans]:
  assumes "s \<preceq> t" and "t \<preceq> u" shows "s \<preceq> u"
proof -
  from `s \<preceq> t` obtain \<sigma> where "s\<cdot>\<sigma> = t" by (cases rule: subst_instance.cases) best
  from `t \<preceq> u` obtain \<tau> where "t\<cdot>\<tau> = u" by (cases rule: subst_instance.cases) best
  hence "(s\<cdot>\<sigma>)\<cdot>\<tau> = u" unfolding `s\<cdot>\<sigma> = t` .
  hence "s\<cdot>(\<sigma> \<circ>s \<tau>) = u" by simp
  thus ?thesis by (rule subst_instanceI)
qed
 
lemma subst_instance_refl: "s \<preceq> s"
  using subst_instanceI[where \<sigma> = "Var" and s = s and t = s] by simp

lemma subst_neutral: "subst_domain \<sigma> \<subseteq> V \<Longrightarrow> (Var x)\<cdot>(\<sigma> |s (V - {x})) = (Var x)"
  by (auto simp: subst_domain_def subst_restrict_def)

lemma subst_restrict_domain[simp]: "\<sigma> |s subst_domain \<sigma> = \<sigma>"
proof -
  have "\<sigma> |s subst_domain \<sigma> = (\<lambda>x. if x \<in> subst_domain \<sigma> then \<sigma>(x) else Var x)"
    by (simp add: subst_restrict_def)
  also have "\<dots> = \<sigma>" by (rule ext) (simp add: subst_domain_def)
  finally show ?thesis .
qed

lemma notin_subst_domain_imp_Var: assumes "x \<notin> subst_domain \<sigma>" shows "\<sigma> x = Var x"
  using assms by (auto simp: subst_domain_def)

lemma subst_domain_neutral[simp]:
  assumes "subst_domain \<sigma> \<subseteq> V" shows "(\<sigma> |s V) = \<sigma>"
proof -
  {
    fix x
    have "(if x \<in> V then \<sigma>(x) else Var x) = (if x \<in> subst_domain \<sigma> then \<sigma>(x) else Var x)"
    proof (cases "x \<in> subst_domain \<sigma>")
      case True
      hence x: "x \<in> V = True" using assms by auto
      show ?thesis unfolding x using True by simp
    next
      case False
      hence x: "x \<notin> subst_domain (\<sigma>)" .
      show ?thesis unfolding notin_subst_domain_imp_Var[OF x] if_cancel ..
    qed
  }
  hence "\<And>x.(if x \<in> V then \<sigma> x else Var x) = (if x \<in> subst_domain \<sigma> then \<sigma> x else Var x)" .
  hence "\<And>x.(\<lambda>x. if x \<in> V then \<sigma> x else Var x) x = (\<lambda>x. if x \<in> subst_domain \<sigma> then \<sigma> x else Var x) x" .
  hence "\<And>x. (\<lambda>x. if x \<in> V then \<sigma> x else Var x) x = \<sigma> x" by (auto simp: subst_domain_def)
  hence "(\<lambda>x. if x \<in> V then \<sigma> x else Var x) = \<sigma>" by (rule ext)
  hence "(\<sigma>) |s V = \<sigma>" by (simp add: subst_restrict_def)
  thus ?thesis .
qed

lemma subst_restrict_UNIV[simp]: "\<sigma> |s UNIV = \<sigma>" by (auto simp: subst_restrict_def)

lemma subst_restrict_empty[simp]: "\<sigma> |s {} = Var" by (simp add: subst_restrict_def)

lemma vars_term_subst: "vars_term (t \<cdot> \<sigma>) = \<Union> vars_term ` \<sigma> ` vars_term t"
  by (induct t, auto)

lemma vars_term_subst_pow:
  "vars_term (t \<cdot> \<sigma>^^n) \<subseteq> vars_term t \<union> (\<Union> vars_term ` subst_range \<sigma>)" (is "_ \<subseteq> ?R t")
  unfolding vars_term_subst
proof (induct n arbitrary: t)
  case 0 show ?case by auto
next
  case (Suc n t)
  show ?case
  proof
    fix x
    assume "x \<in> \<Union> vars_term ` (\<sigma> ^^ Suc n) ` vars_term t"
    then obtain y u where 1: "y \<in> vars_term t" "u = (\<sigma> ^^ Suc n) y" "x \<in> vars_term u"
      by auto
    from 1(2) have "u = \<sigma> y \<cdot> \<sigma> ^^ n" by (auto simp: subst_compose_def)
    from 1(3)[unfolded this, unfolded vars_term_subst]
    have "x \<in> \<Union> vars_term ` (\<sigma> ^^ n) ` vars_term (\<sigma> y)" .
    with Suc[of "\<sigma> y"] have x: "x \<in> ?R (\<sigma> y)" by auto 
    thus "x \<in> ?R t" 
    proof
      assume "x \<in> vars_term (\<sigma> y)"
      thus ?thesis using 1(1) by (cases "\<sigma> y = Var y", auto simp: subst_domain_def)
    qed auto
  qed
qed

lemma term_subst_eq: assumes "\<And> x. x \<in> vars_term t \<Longrightarrow> \<sigma> x = \<tau> x" 
  shows "t \<cdot> \<sigma> = t \<cdot> \<tau>"
  using assms by (induct t, auto)

lemma term_subst_eq_rev: "t \<cdot> \<sigma> = t \<cdot> \<tau> \<Longrightarrow> \<forall> x \<in> vars_term t. \<sigma> x = \<tau> x"
proof (induct t, simp)
  case (Fun f ts)
  thus ?case by auto
qed

lemma term_subst_eq_conv: "(t \<cdot> \<sigma> = t \<cdot> \<tau>) = (\<forall> x \<in> vars_term t. \<sigma> x = \<tau> x)"
  using term_subst_eq[of t \<sigma> \<tau>] term_subst_eq_rev[of t \<sigma> \<tau>] by auto

lemma coincidence_lemma:"t\<cdot>\<sigma> = t\<cdot>(\<sigma> |s (vars_term t))"
  unfolding term_subst_eq_conv subst_restrict_def by auto

lemma subst_domain_vars_term_subset:
  "subst_domain (\<sigma>  |s  vars_term t) \<subseteq> vars_term t"
  by (auto simp: subst_domain_def subst_restrict_def)

lemma subst_restrict_single_Var[simp]: assumes "x \<notin> subst_domain \<sigma>" shows "\<sigma> |s {x} = Var"
proof -
  have A: "\<And>x. x \<notin> subst_domain \<sigma> \<Longrightarrow> \<sigma> x = Var x" by (simp add: subst_domain_def)
  have "\<sigma> |s {x} = (\<lambda>y. if y \<in> {x} then \<sigma> y else Var y)" by (simp add: subst_restrict_def)
  also have "\<dots> = (\<lambda>y. if y = x then \<sigma> y else Var y)" by simp
  also have "\<dots> = (\<lambda>y. if y = x then \<sigma> x else Var y)" by (simp cong: if_cong)
  also have "\<dots> = (\<lambda>y. if y = x then Var x else Var y)" unfolding A[OF assms] by simp
  also have "\<dots> = (\<lambda>y. if y = x then Var y else Var y)" by (simp cong: if_cong)
  also have "\<dots> = (\<lambda>y. Var y)" by simp
  finally show ?thesis by simp
qed

lemma subst_restrict_single_Var':
  assumes "x \<notin> subst_domain \<sigma>" and "\<sigma> |s V = Var" shows "\<sigma> |s ({x} \<union> V) = Var"
proof -
  have "(\<lambda>y. if y \<in> V then \<sigma> y else Var y) = (\<lambda>y. Var y)"
   using `\<sigma> |s V = Var`  by (simp add: subst_restrict_def)
  hence "(\<lambda>y. if y \<in> V then \<sigma> y else Var y) = (\<lambda>y. Var y)" by simp
  hence A: "\<And>y. (if y \<in> V then \<sigma> y else Var y) = Var y" by (rule fun_cong)
  {
    fix y
    have "(if y \<in> {x} \<union> V then \<sigma> y else Var y) = Var y" 
    proof (cases "y = x")
      assume "y = x" thus ?thesis using `x \<notin> subst_domain \<sigma>` by (auto simp: subst_domain_def)
    next
      assume "y \<noteq> x" thus ?thesis using A by simp
    qed
  }
  hence "\<And>y. (if y \<in> {x} \<union> V then \<sigma> y else Var y) = Var y" by simp
  thus ?thesis by (auto simp: subst_restrict_def)
qed

lemma subst_restrict_empty_set:"finite V \<Longrightarrow> V \<inter> subst_domain \<sigma> = {} \<Longrightarrow> \<sigma> |s V = Var"
proof(induct rule: finite.induct)
  case emptyI thus ?case by simp
next
  case (insertI V x)
  hence "V \<inter> subst_domain \<sigma> = {}" by simp
  with insertI have "\<sigma> |s V = Var" by simp
  thus ?case using insertI subst_restrict_single_Var'[where \<sigma> = \<sigma> and x = x and V = V] by simp
qed

lemma subst_restrict_Var: "x \<noteq> y \<Longrightarrow> (Var y)\<cdot>(\<sigma> |s (UNIV-{x})) = Var y\<cdot>\<sigma>"
  by (auto simp: subst_restrict_def)

lemma vars_subst_ex:
  assumes A: "vars_term s \<subseteq> vars_term t" and B: "t \<cdot> \<sigma> = t \<cdot> \<tau>" shows "s \<cdot> \<sigma> = s \<cdot> \<tau>"
using A proof (induct s)
 case (Var x) thus ?case using B by (induct t) auto
next
 case (Fun f ss) thus ?case by auto
qed

lemma supteq_stable:
  assumes "s \<unrhd> t" shows "s \<cdot> \<sigma> \<unrhd> t \<cdot> \<sigma>"
using assms
proof (induct rule: supteq.induct)
  case (refl s) show ?case by (rule refl)
next
  case (subt u ss t f)
  from `u \<in> set ss` have "u\<cdot>\<sigma> \<in> set (map (\<lambda>t. t\<cdot>\<sigma>) ss)" "u\<cdot>\<sigma> \<unrhd> u\<cdot>\<sigma>" by auto
  hence "(Fun f ss)\<cdot>\<sigma> \<unrhd> u\<cdot>\<sigma>" unfolding subst_apply_term.simps by blast
  from this and `u\<cdot>\<sigma> \<unrhd> t\<cdot>\<sigma>` show ?case by (rule supteq_trans)
qed

lemma supt_stable: assumes "s \<rhd> t" shows "s\<cdot>\<sigma> \<rhd> t\<cdot>\<sigma>"
using assms
proof (induct rule: supt.induct)
 case (arg s ss f) 
 hence "s\<cdot>\<sigma> \<in> set (map (\<lambda>t. t\<cdot>\<sigma>) ss)" by simp
thus ?case unfolding subst_apply_term.simps by (rule supt.arg)
next
 case (subt u ss t f)
 from `u \<in> set ss` have "u\<cdot>\<sigma> \<in> set (map (\<lambda>t. t\<cdot>\<sigma>) ss)" by simp
 hence "(Fun f ss)\<cdot>\<sigma> \<rhd> u\<cdot>\<sigma>" unfolding subst_apply_term.simps by (rule supt.arg)
 from this and `u\<cdot>\<sigma> \<rhd> t\<cdot>\<sigma>` show ?case by (rule supt_trans)
qed

fun
  funas_ctxt :: "('f, 'v) ctxt \<Rightarrow> 'f sig"
where
  "funas_ctxt Hole = {}" |
  "funas_ctxt (More f ss1 D ss2) = {(f, Suc (length (ss1 @ ss2)))}
     \<union> funas_ctxt D \<union> \<Union>(set (map funas_term (ss1 @ ss2)))"

lemma funas_term_ctxt_apply[simp]: "funas_term (C\<langle>t\<rangle>) = funas_ctxt C \<union> funas_term t"
proof (induct t)
  case (Var x) show ?case by (induct C) auto
next
  case (Fun f ts) show ?case by (induct C arbitrary: f ts) auto
qed

lemma funas_term_subst:
  "funas_term (t\<cdot>\<sigma>) = funas_term t \<union> (\<Union> funas_term ` \<sigma> ` vars_term t)"
by (induct t) auto

lemma funas_term_subst_pow:
  "funas_term (t \<cdot> \<sigma>^^n) \<subseteq> funas_term t \<union> (\<Union> funas_term ` subst_range \<sigma>)"
proof -
  {
    fix Xs
    have "\<Union> funas_term ` (\<sigma> ^^ n) ` Xs  \<subseteq> \<Union> funas_term ` subst_range \<sigma>"
    proof (induct n arbitrary: Xs)
      case 0 show ?case by auto
    next
      case (Suc n Xs)
      show ?case (is "\<Union> ?L \<subseteq> ?R")
      proof (rule subsetI)
        fix f
        assume "f \<in> \<Union> ?L"
        then obtain x where "f \<in> funas_term ((\<sigma> ^^ Suc n) x)" by auto
        hence "f \<in> funas_term (\<sigma> x \<cdot> \<sigma> ^^ n)" by (auto simp: subst_compose_def)
        from this[unfolded funas_term_subst]
        show "f \<in> ?R" using Suc[of "vars_term (\<sigma> x)"]
          unfolding subst_range.simps subst_domain_def by (cases "\<sigma> x = Var x", auto)
      qed
    qed
  }
  thus ?thesis unfolding funas_term_subst by auto
qed

lemma get_var_or_const: "\<exists> C t. s = C\<langle>t\<rangle> \<and> args t = []"
proof (induct s)
  case (Var y)
  show ?case by (rule exI[of _ Hole], auto)
next
  case (Fun f ts)
  show ?case
  proof (cases ts)
    case Nil
    show ?thesis unfolding Nil 
      by (rule exI[of _ Hole], auto)
  next
    case (Cons s ss)
    hence "s \<in> set ts" by auto
    from Fun[OF this] obtain C where C: "\<exists> t. s = C\<langle>t\<rangle> \<and> args t = []" by auto
    show ?thesis unfolding Cons
      by (rule exI[of _ "More f [] C ss"], insert C, auto) 
  qed
qed


lemma supteq_Var_id[simp]: assumes "Var x \<unrhd> s" shows "s = Var x"
using assms proof (induct s rule: term_induct)
  case (Var y) thus ?case by (cases rule: supteq.cases) simp_all
next
  case (Fun f ts)
  from `Var x \<unrhd> Fun f ts` have False by (cases rule: supteq.cases) 
  thus ?case by simp
qed


lemma arg_not_term[simp]: assumes "t \<in> set ts" shows "Fun f ts \<noteq> t"
proof (rule ccontr)
  assume "\<not> Fun f ts \<noteq>  t"
  hence "size(Fun f ts) = size t" by simp
  moreover from assms have "size t < list_size size ts" by (induct ts) auto
  ultimately show False by simp
qed

lemma arg_subteq[simp]: assumes "t \<in> set ts" shows "Fun f ts \<unrhd> t"
using assms by auto

lemma supt_imp_args:
  assumes "\<forall>t. s \<rhd> t \<longrightarrow> P t" shows "\<forall>t\<in>set(args s). P t"
using assms by (cases s) (simp_all add: supt_supteq_conv)

lemma ctxt_apply_eq_False[simp]:
  assumes "t = (More f ss1 D ss2)\<langle>t\<rangle>" (is "_ = ?C\<langle>_\<rangle>") shows "False"
proof -
  have "?C \<noteq> \<box>" by auto
  from this and  assms have "t \<rhd> t" by (rule ctxt_supt)
  thus False by auto
qed

lemma supteq_imp_funas_subset[simp]:
  "s \<unrhd> t \<Longrightarrow> funas_term t \<subseteq> funas_term s"
by (induct rule: supteq.induct) auto

lemma funas_ctxt_compose[simp]:
  "funas_ctxt (C \<circ>c D) = funas_ctxt C \<union> funas_ctxt D"
  by (induct C, auto)

lemma set_set_cons:
  assumes "P x" and set:"\<And>y. y \<in> set xs \<Longrightarrow> P y" shows "y \<in> set(x#xs) \<Longrightarrow> P y"
using assms by auto



lemma ctxt_power_compose_distr:
  "C ^ (m + n) = C ^ m \<circ>c C ^ n"
by (induct m) (simp_all add: one_ctxt_def times_ctxt_def)

fun
  vars_ctxt :: "('f, 'v) ctxt \<Rightarrow> 'v set"
where
  "vars_ctxt \<box> = {}" |
  "vars_ctxt (More f ss1 D ss2) = vars_ctxt D \<union> \<Union>set (map vars_term (ss1 @ ss2))"

lemma subst_apply_id': assumes "vars_term t \<inter> V = {}" shows "t\<cdot>(\<sigma> |s V) = t"
using assms proof (induct t)
  case (Var x) thus ?case by (simp add: subst_restrict_def)
next
  case (Fun f ts)
  hence IH: "\<forall>s\<in>set ts. s\<cdot>(\<sigma> |s V) = s" by auto
  with map_idI[of ts "\<lambda>t. t\<cdot>(\<sigma> |s V)"] show ?case by simp
qed

lemma subst_apply_ctxt_id: assumes "vars_ctxt C \<inter> V = {}" shows "C \<cdot>c (\<sigma> |s V) = C"
using assms proof (induct C)
  case Hole show ?case by simp
next
  case (More f ss1 D ss2)
  hence IH: "D \<cdot>c (\<sigma> |s V) = D" by auto
  from More have "\<forall>s\<in>set(ss1@ss2). vars_term s \<inter> V = {}" by auto
  with subst_apply_id' have args: "\<forall>s\<in>set(ss1@ss2). s\<cdot>(\<sigma> |s V) = s" by best
  from args have "\<forall>s\<in>set ss1. s\<cdot>(\<sigma> |s V) = s" by simp
  with map_idI[of ss1 "\<lambda>t. t\<cdot>(\<sigma> |s V)"] have ss1: "map (\<lambda>s. s\<cdot>(\<sigma> |s V)) ss1 = ss1" by best
  from args have "\<forall>s\<in>set ss2. s\<cdot>(\<sigma> |s V) = s" by simp
  with map_idI[of ss2 "\<lambda>t. t\<cdot>(\<sigma> |s V)"] have ss2: "map (\<lambda>s. s\<cdot>(\<sigma> |s V)) ss2 = ss2" by best
  show ?case by (simp add: ss1 ss2 IH)
qed

lemma vars_term_Var_id: "vars_term o Var = (\<lambda>x. {x})" by (rule ext) simp

lemma ctxt_exhaust_rev[case_names Hole More]:
  assumes "C = \<box> \<Longrightarrow> P" and "\<And>D f ss1 ss2. C = D \<circ>c (More f ss1 \<box> ss2) \<Longrightarrow> P" shows "P"
proof (cases C)
  case Hole with assms show ?thesis by simp
next
  case (More g ts1 E ts2)
  hence "\<exists>D f ss1 ss2. C = D \<circ>c (More f ss1 \<box> ss2)"
  proof (induct E arbitrary: C g ts1 ts2)
    case Hole hence "C = \<box> \<circ>c (More g ts1 \<box> ts2)" by simp
    thus ?case by best
  next
    case (More h us1 F us2)
    from More(1)[of "More h us1 F us2"]
     obtain G i vs1 vs2 where IH: "More h us1 F us2 = G \<circ>c More i vs1 \<box> vs2" by force
    from More have "C = (More g ts1 \<box> ts2 \<circ>c G) \<circ>c More i vs1 \<box> vs2" unfolding IH by simp
    thus ?case by best
  qed
  thus ?thesis using assms by auto
qed

lemma funas_ctxt_subst[simp]: "funas_ctxt (C \<cdot>c \<sigma>) = funas_ctxt C \<union> \<Union> funas_term ` \<sigma> ` vars_ctxt C"
proof (induct C)
  case (More f bef C aft)
  let ?n = "Suc (length bef + length aft)"
  let ?\<sigma> = "map (\<lambda> t. t \<cdot> \<sigma>)"
  let ?C = "More f bef C aft"
  obtain ba where ba: "ba = bef @ aft" by auto
  let ?fba = "\<Union> funas_term ` (set (?\<sigma> (ba)))" 
  let ?fba' = "\<Union> {funas_term (ba ! i \<cdot> \<sigma>) | i. i < length ba}"
  let ?fba'' = "\<Union> funas_term ` (set ba) \<union> \<Union> funas_term ` \<sigma> ` (\<Union> vars_term ` (set ba))"
  let ?fC = "funas_ctxt (C \<cdot>c \<sigma>)"
  let ?fC' = "funas_ctxt C \<union> \<Union> funas_term ` \<sigma> ` vars_ctxt C"
  have "?fba = ?fba'" unfolding set_conv_nth  by auto
  also have "... = ?fba''" unfolding funas_term_subst set_conv_nth by auto
  finally have fba: "?fba = ?fba''" .
  have "funas_ctxt (?C \<cdot>c \<sigma>) = {(f,?n)} \<union> ?fC \<union> ?fba" by (simp add: ba)
  also have "... = {(f,?n)} \<union> ?fC' \<union> ?fba''" unfolding More fba ..
  finally have id: "funas_ctxt (?C \<cdot>c \<sigma>) = {(f,?n)} \<union> ?fC' \<union> ?fba''" .
  show ?case unfolding id by (auto simp: ba)
qed simp


fun
  subst_extend :: "('f, 'v,'w) gsubst \<Rightarrow> ('v \<times> ('f, 'w) term) list \<Rightarrow> ('f, 'v,'w) gsubst"
where
  "subst_extend \<sigma> vts = (\<lambda>x.
    (case lookup x vts of
      Some t \<Rightarrow> t
    | None   \<Rightarrow> \<sigma>(x)))"

lemma subst_extend_id:
  assumes "V \<inter> set vs = {}" and "vars_term t \<subseteq> V"
  shows "t \<cdot> subst_extend \<sigma> (zip vs ts) = t \<cdot> \<sigma>"
using assms
proof (induct t)
 case (Var x) thus ?case
   using lookup_sound [of x "zip vs ts"]
   using set_zip_leftD [of x _ vs ts]
   using IntI [of x V "set vs"]
   using emptyE
   by (case_tac "lookup x (zip vs ts)") auto
next
 case (Fun f ts) thus ?case by auto
qed

lemma funas_term_args: 
  shows "\<Union> funas_term ` set (args t) \<subseteq> funas_term t" by (cases t, auto)


lemma subst_extend_absorb:
  assumes "distinct vs" and "length vs = length ss"
  shows "map (\<lambda>t. t\<cdot>(subst_extend \<sigma> (zip vs ss))) (map Var vs) = ss" (is "?ss = _")
proof -
  let ?\<sigma> = "subst_extend (\<sigma>) (zip vs ss)"
  from assms have "length vs \<le> length ss" by simp
  from assms have "length ?ss = length ss" by simp
  moreover have "\<forall>i<length ?ss. ?ss!i = ss!i"
  proof (intro allI impI)
    fix i assume "i < length ?ss"
    hence i: "i < length(map Var vs)" by simp
    hence len: "i < length vs" by simp
    have "?ss!i = (map Var vs!i)\<cdot>?\<sigma>" unfolding nth_map[OF i,of "\<lambda>t. t\<cdot>?\<sigma>"] by simp
    also have "\<dots> = Var(vs!i)\<cdot>?\<sigma>" unfolding nth_map[OF len] by simp
    also have "\<dots> = (case lookup (vs!i) (zip vs ss) of None \<Rightarrow> \<sigma>(vs!i) | Some t \<Rightarrow> t)" by simp
    also have "\<dots> = ss!i" unfolding lookup_Some[OF `distinct vs` `length vs \<le> length ss` len] by simp
    finally show "?ss!i = ss!i" by simp
  qed
  ultimately show ?thesis by (rule nth_equalityI)
qed

lemma vars_term_finite[simp]: "finite(vars_term t)" by (induct t) simp_all

fun linear_term :: "('f, 'v) term \<Rightarrow> bool" where
  [code del]: "linear_term (Var _) = True" (*why "code del"?*)
| "linear_term (Fun _ ts) = (is_partition (map vars_term ts) \<and> (\<forall>t\<in>set ts. linear_term t))"

fun map_term :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('g, 'v) term" where
  "map_term fun (Var x) = Var x"
| "map_term fun (Fun f ts) = Fun (fun f) (map (map_term fun) ts)"

fun map_subst :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('g, 'v) subst" where
  "map_subst fg \<sigma> = (\<lambda>x. map_term fg (\<sigma> x))"

fun map_ctxt :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('g, 'v) ctxt" where
  "map_ctxt fg \<box> = \<box>"
| "map_ctxt fg (More f ss C ts) = More (fg f) (map (map_term fg) ss) (map_ctxt fg C) (map (map_term fg) ts)"

lemma map_comp:
  "Term.map v f (Term.map w g t) = Term.map (v \<circ> w) (f \<circ> g) t"
  by (induct t) (simp_all add: Term.map_def)

lemma map_term_comp: "map_term fg (map_term gh t) = map_term (fg \<circ> gh) t"
  by (induct t) simp_all

lemma map_subst_distrib[simp]: "map_term fg (t \<cdot> \<sigma>) = map_term fg t \<cdot> map_subst fg \<sigma>"
  by (induct t) (auto)

lemma fold_ident[simp]: "Term.fold Var Fun t = t"
  by (induct t) (auto simp: map_ext[of _ "Term.fold Var Fun" id])

lemma map_term_ident[simp]: "map_term id t = t"
  by (induct t) (simp_all add: map_idI)

lemma map_term_power:
  fixes f :: "'f \<Rightarrow> 'f" 
  shows "((map_term f) ^^ n) = map_term (f ^^ n)"
proof (rule sym, intro ext)
  fix t
  show "map_term (f ^^ n) t = (map_term f ^^ n) t"
  proof (induct n)
    case 0
    show ?case
      by (simp, unfold id_def[symmetric], simp)
  next
    case (Suc n)
    show ?case by (simp add: Suc[symmetric] map_term_comp o_def)
  qed
qed

lemma map_ctxt_distrib[simp]: "map_term fg (C\<langle>t\<rangle>) = (map_ctxt fg C)\<langle>map_term fg t\<rangle>" 
  by  (induct C) (auto)

fun map_term_wa :: "('f \<Rightarrow> nat \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('g, 'v) term" where
  "map_term_wa fg (Var x) = (Var x)"
| "map_term_wa fg (Fun f ts) = Fun (fg f (length ts)) (map (map_term_wa fg) ts)"

lemma map_term_map_term_wa: "map_term fg = map_term_wa (\<lambda> f n. fg f)"
proof (intro ext)
  fix t
  show "map_term fg t = map_term_wa (\<lambda> f n. fg f) t"
    by (induct t, auto)
qed

fun map_ctxt_wa :: "('f \<Rightarrow> nat \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('g, 'v) ctxt" where
  "map_ctxt_wa fg \<box> = \<box>"
| "map_ctxt_wa fg (More f bef C aft) = More (fg f (Suc (length bef + length aft))) (map (map_term_wa fg) bef) (map_ctxt_wa fg C) (map (map_term_wa fg) aft)"

abbreviation map_subst_wa :: "('f \<Rightarrow> nat \<Rightarrow> 'g) \<Rightarrow> ('f,'v)subst \<Rightarrow> ('g,'v)subst"
where "map_subst_wa fg \<sigma> \<equiv> (\<lambda>x. map_term_wa fg (\<sigma> x))"

lemma map_term_wa_subst[simp]: "map_term_wa fg (t \<cdot> \<sigma>) = map_term_wa fg t \<cdot> map_subst_wa fg \<sigma>"
  by (induct t, auto)

lemma map_term_wa_ctxt[simp]: "map_term_wa fg (C\<langle>t\<rangle>) = (map_ctxt_wa fg C) \<langle>map_term_wa fg t\<rangle>"
  by (induct C, auto)


lemma map_term_wa_id: assumes id: "\<And> f n. fg f n = f"
  shows "map_term_wa fg t = t"
proof (induct t, simp)
  case (Fun f ss)
  thus ?case by (simp add: id, induct ss, auto)
qed

fun weak_match :: "('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where
  "weak_match _ (Var _) = True"
| "weak_match (Var _) (Fun _ _) = False"
| "weak_match (Fun f ts) (Fun g ss) = (f = g \<and> length ts = length ss \<and> (\<forall> i < length ts. weak_match (ts ! i) (ss ! i)))"

lemma weak_match_refl: "weak_match t t"
  by (induct t) auto

lemma weak_match_match: "weak_match (t \<cdot> \<sigma>) t"
  by (induct t) auto

lemma notin_subst_restrict[simp]: assumes "x \<notin> V" shows "(\<sigma>  |s  V) x = Var x"
  using assms by (auto simp: subst_restrict_def)

lemma in_subst_restrict[simp]: assumes "x \<in> V" shows "(\<sigma>  |s  V) x = \<sigma> x"
  using assms by (auto simp: subst_restrict_def)

lemma subst_merge: 
  assumes part: "is_partition (map vars_term ts)"
  shows "\<exists>\<sigma>. \<forall>i<length ts. \<forall>x\<in>vars_term (ts ! i). \<sigma> x = \<tau> i x"
proof -
  let ?\<tau> = "map \<tau> [0 ..< length ts]"
  let ?\<sigma> = "fun_merge ?\<tau> (map vars_term ts)"
  show ?thesis
    by (rule exI[of _ ?\<sigma>], intro allI impI ballI, 
      insert fun_merge_part[OF part, of _ _ ?\<tau>], auto)
qed  

lemma linear_weak_match:
  assumes "linear_term l" and "weak_match t s" and "s = l \<cdot> \<sigma>"
  shows "\<exists>\<tau>. t = l\<cdot>\<tau> \<and> (\<forall> x \<in> vars_term l. weak_match (Var x \<cdot> \<tau>) (Var x \<cdot> \<sigma>))"
using assms proof (induct l arbitrary: s t)
  case (Var x)
  show ?case
  proof (rule exI[of _ "(\<lambda> y. t)"], intro conjI, simp)
    from Var show "\<forall>x\<in>vars_term (Var x). weak_match (Var x \<cdot> (\<lambda>y. t)) (Var x \<cdot> \<sigma>)"
      by force
  qed
next
  case (Fun f ls)
  let ?n = "length ls"
  from Fun(4) obtain ss where s: "s = Fun f ss" and lss: "length ss = ?n"  by (cases s, auto)
  with Fun(4) have match: "\<And> i. i < ?n \<Longrightarrow> ss ! i = (ls ! i) \<cdot> \<sigma>" by auto
  from Fun(3) s lss obtain ts where t: "t = Fun f ts"
    and lts: "length ts = ?n" by (cases t, auto)
  with Fun(3) s have weak_match: "\<And> i. i < ?n \<Longrightarrow> weak_match (ts ! i) (ss ! i)" by auto
  from Fun(2) have linear: "\<And> i. i < ?n \<Longrightarrow> linear_term (ls ! i)" by simp
  let ?cond = "\<lambda> \<tau> i. ts ! i = ls ! i \<cdot> \<tau> \<and> (\<forall>x\<in>vars_term (ls ! i). weak_match (Var x \<cdot> \<tau>) (Var x \<cdot> \<sigma>))"
  {
    fix i
    assume i: "i < ?n"
    hence "ls ! i \<in> set ls" by simp
    from Fun(1)[OF this linear[OF i] weak_match[OF i] match[OF i]]
    have "\<exists> \<tau>. ?cond \<tau> i" .
  }
  hence "\<forall> i. \<exists> \<tau>. (i < ?n \<longrightarrow> ?cond \<tau> i)" by auto
  from choice[OF this] obtain subs where subs: "\<And> i. i < ?n \<Longrightarrow> ?cond (subs i) i" by auto  
  from Fun(2) have distinct: "is_partition(map vars_term ls)" by simp
  from subst_merge[OF this, of subs]
  obtain \<tau> where \<tau>: "\<And> i x . i < length ls \<Longrightarrow> x \<in> vars_term (ls ! i) \<Longrightarrow> \<tau> x = subs i x" by auto
  show ?case
  proof (rule exI[of _ \<tau>], simp add: t, intro ballI conjI)
    fix li x
    assume li: "li \<in> set ls" and x: "x \<in> vars_term li"
    from li obtain i where i: "i < ?n" and li: "li = ls ! i" unfolding set_conv_nth by auto
    with x have x: "x \<in> vars_term (ls ! i)" by simp
    from subs[OF i, THEN conjunct2, THEN bspec, OF x] show "weak_match (\<tau> x) (\<sigma> x)" unfolding \<tau>[OF i x[unfolded li]]
      by simp
  next
    show "ts = map (\<lambda> t. t \<cdot> \<tau>) ls" 
    proof (rule nth_equalityI, simp add: lts, intro allI impI)
      fix i
      assume "i < length ts"
      with lts have i: "i < ?n" by simp
      have "ts ! i = ls ! i \<cdot> subs i"
        by (rule subs[THEN conjunct1, OF i])
      also have "... = ls ! i \<cdot> \<tau>" unfolding term_subst_eq_conv using \<tau>[OF i] by auto
      finally show "ts ! i = map (\<lambda> t. t \<cdot> \<tau>) ls ! i"
        by (simp add: nth_map[OF i])
    qed
  qed
qed

lemma vars_map_term[simp]: "vars_term \<circ> map_term f = vars_term"
proof
  fix t 
  show "(vars_term \<circ> map_term f) t = vars_term t"
    by (induct t) (auto)
qed

lemma vars_map_term2[simp]: "vars_term (map_term f t) = vars_term t"
  using fun_cong[OF vars_map_term, of f t] by (simp del: vars_map_term)

lemma map_term_subst_split: assumes "map_term fg t = s \<cdot> \<sigma>"
  and  "linear_term s"
  shows "\<exists> u \<tau>. t = u \<cdot> \<tau> \<and> map_term fg u = s \<and> (\<forall> x \<in> vars_term s. map_term fg (\<tau> x) = (\<sigma> x))"
  using assms
proof (induct s arbitrary: t)
  case (Var x t)
  show ?case 
  proof (intro exI conjI)
    show "t = Var x \<cdot> (\<lambda> _. t)" by simp 
  qed (insert Var, auto) 
next
  case (Fun g ss t)
  from Fun(2) obtain f ts where t: "t = Fun f ts" by (cases t, auto)
  note Fun = Fun[unfolded t, simplified]
  let ?n = "length ss"
  from Fun have rec: "map (map_term fg) ts = map (\<lambda> t. t \<cdot> \<sigma>) ss"
    and g: "fg f = g"
    and lin: "\<And> s. s \<in> set ss \<Longrightarrow> linear_term s"
    and part: "is_partition (map vars_term ss)" by auto
  from arg_cong[OF rec, of length] have len: "length ts = ?n" by simp  
  from map_nth_conv[OF rec] have rec: "\<And> i. i < ?n \<Longrightarrow>  map_term fg (ts ! i) = ss ! i \<cdot> \<sigma>" unfolding len by auto
  let ?p = "\<lambda> i u \<tau>. ts ! i = u \<cdot> \<tau> \<and> map_term fg u = ss ! i \<and> (\<forall> x \<in> vars_term (ss ! i). map_term fg (\<tau> x) = \<sigma> x)"
  {
    fix i
    assume i: "i < ?n"
    hence "ss ! i \<in> set ss" by auto
    from Fun(1)[OF this rec[OF i] lin[OF this]]
    have "\<exists> u \<tau>. ?p i u \<tau>" .
  }
  hence "\<forall> i. \<exists> u \<tau>. i < ?n \<longrightarrow> ?p i u \<tau>" by blast
  from choice[OF this] obtain us where "\<forall> i. \<exists> \<tau>. i < ?n \<longrightarrow> ?p i (us i) \<tau>"  ..
  from choice[OF this] obtain \<tau>s where p: "\<And> i. i < ?n \<Longrightarrow> ?p i (us i) (\<tau>s i)" by blast
  from subst_merge[OF part, of \<tau>s] obtain \<tau> where \<tau>: "\<And> i x. i < ?n \<Longrightarrow> x \<in> vars_term (ss ! i) \<Longrightarrow> \<tau> x = \<tau>s i x" by blast
  {
    fix i
    assume i: "i < ?n"
    from p[OF i] have "map_term fg (us i) = ss ! i" by auto
    from arg_cong[OF this, of vars_term] vars_map_term[of fg] 
    have "vars_term (us i) = vars_term (ss ! i)" by auto
  } note vars = this
  let ?us = "map us [0 ..< ?n]"
  show ?case
  proof (intro exI conjI ballI)
    have ss: "ts = map (\<lambda> t. t \<cdot> \<tau>) ?us"
      unfolding list_eq_iff_nth_eq
      unfolding len
    proof (intro conjI allI impI)
      fix i
      assume i: "i < ?n"
      have us: "?us ! i = us i" using nth_map_upt[of i ?n 0] i by auto
      have "(map (\<lambda> t. t \<cdot> \<tau>) ?us) ! i = us i \<cdot> \<tau>" 
        unfolding us[symmetric]
        using nth_map[of i ?us "\<lambda> t. t \<cdot> \<tau>"]  i by force
      also have "... = us i \<cdot> \<tau>s i" 
        by (rule term_subst_eq, rule \<tau>[OF i], insert vars[OF i], auto )
      also have "... = ts ! i" using p[OF i] by simp
      finally 
      show "ts ! i = map (\<lambda> t. t \<cdot> \<tau>) ?us ! i" ..
    qed auto
    show "t = Fun f ?us  \<cdot> \<tau>"
      unfolding t
      unfolding ss by auto
  next
    show "map_term fg (Fun f ?us) = Fun g ss"
      using p g by (auto simp: list_eq_iff_nth_eq[of _ ss])
  next
    fix x
    assume x: "x \<in> vars_term (Fun g ss)"
    then obtain s where s: "s \<in> set ss" and x: "x \<in> vars_term s" by auto    
    from s[unfolded set_conv_nth] obtain i where s: "s = ss ! i" and i: "i < ?n" by auto
    note x = x[unfolded s]
    from p[OF i] vars[OF i] x \<tau>[OF i x]
    show "map_term fg (\<tau> x) = \<sigma> x" by auto
  qed
qed

lemma map_term_wa_ctxt_split: assumes "map_term_wa fg s = lC\<langle>lt\<rangle>"
  shows "\<exists> C t. s = C\<langle>t\<rangle> \<and> map_term_wa fg t = lt \<and> map_ctxt_wa fg C = lC"
  using assms
proof (induct lC arbitrary: s)
  case Hole
  show ?case
    by (rule exI[of _ Hole], insert Hole, auto)
next
  case (More lf lbef lC laft s)
  from More(2) obtain fs ss where s: "s = Fun fs ss" by (cases s, auto)
  note More = More[unfolded s, simplified]
  let ?lb = "length lbef"
  let ?la = "length laft"
  let ?n = "Suc (?lb + ?la)"  
  let ?m = "map_term_wa fg"
  from More(2) have rec: "map ?m ss = lbef @ lC\<langle>lt\<rangle> # laft"
    and lf: "lf = fg fs (length ss)" by blast+
  from arg_cong[OF rec, of length] have len: "length ss = ?n" by auto
  hence lb: "?lb < length ss" by auto
  note ss = id_take_nth_drop[OF this]
  from rec ss have "map ?m (take ?lb ss @ ss ! ?lb # drop (Suc ?lb) ss) = lbef @ lC\<langle>lt\<rangle> # laft" by auto
  hence id: "take ?lb (map ?m ss) @ ?m (ss ! ?lb) # drop (Suc ?lb) (map ?m ss) = lbef @ lC\<langle>lt\<rangle> # laft"
    (is "?l1 @ ?l2 # ?l3 = ?r1 @ ?r2 # ?r3")
    unfolding take_map drop_map 
    by auto
  from len have len2: "\<And> P. length ?l1 = length ?r1 \<or> P"
    unfolding length_take by auto
  from id[unfolded List.append_eq_append_conv[OF len2]]
  have id: "?l1 = ?r1" "?l2 = ?r2" "?l3 = ?r3" by auto
  from More(1)[OF id(2)] obtain C t where sb: "ss ! ?lb = C\<langle>t\<rangle>" and map: "map_term_wa fg t = lt" and ma: "map_ctxt_wa fg C = lC" by auto  
  let ?C = "More fs (take ?lb ss) C (drop (Suc ?lb) ss)"
  have s: "s = ?C\<langle>t\<rangle>"
    unfolding s using ss[unfolded sb] by simp
  have len3: "Suc (length (take ?lb ss) + length (drop (Suc ?lb) ss)) = length ss"
    unfolding length_take length_drop len by auto
  show ?case
  proof (intro exI conjI, rule s, rule map)
    show "map_ctxt_wa fg ?C = More lf lbef lC laft"
      unfolding map_ctxt_wa.simps 
      unfolding len3
      using id ma lf
      unfolding take_map drop_map
      by auto
  qed
qed



lemma subst_extend_flat_ctxt:
  assumes dist: "distinct vs"
    and len1: "length(take i (map Var vs)) = length ss1"
    and len2: "length(drop (Suc i) (map Var vs)) = length ss2"
    and i: "i < length vs"
  shows "More f (take i (map Var vs)) \<box> (drop (Suc i) (map Var vs)) \<cdot>c subst_extend \<sigma> (zip (take i vs@drop (Suc i) vs) (ss1@ss2)) = More f ss1 \<box> ss2"
proof -
  let ?V = "map Var vs"
  let ?vs1 = "take i vs" and ?vs2 = "drop (Suc i) vs"
  let ?ss1 = "take i ?V" and ?ss2 = "drop (Suc i) ?V"
  let ?\<sigma> = "subst_extend \<sigma> (zip (?vs1@?vs2) (ss1@ss2))"
  from len1 and len2 have len: "length(?vs1@?vs2) = length(ss1@ss2)" using i by simp
  from distinct_take_drop[OF dist i] have "distinct(?vs1@?vs2)" .
  from subst_extend_absorb[OF this len,of "\<sigma>"]
    have map: "map (\<lambda>t. t\<cdot>?\<sigma>) (?ss1@?ss2) = ss1@ss2" unfolding take_map drop_map map_append .
  from len1 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss1 = ss1" by auto
  moreover from len2 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss2 = ss2" by auto
  ultimately show ?thesis by simp
qed


lemma subst_extend_flat_ctxt'':
  assumes dist: "distinct vs"
    and len1: "length(take i (map Var vs)) = length ss1"
    and len2: "length(drop i (map Var vs)) = length ss2"
    and i: "i < length vs"
  shows "More f (take i (map Var vs)) \<box> (drop i (map Var vs)) \<cdot>c subst_extend \<sigma> (zip (take i vs@drop i vs) (ss1@ss2)) = More f ss1 \<box> ss2"
proof -
  let ?V = "map Var vs"
  let ?vs1 = "take i vs" and ?vs2 = "drop i vs"
  let ?ss1 = "take i ?V" and ?ss2 = "drop i ?V"
  let ?\<sigma> = "subst_extend \<sigma> (zip (?vs1@?vs2) (ss1@ss2))"
  from len1 and len2 have len: "length(?vs1@?vs2) = length(ss1@ss2)" using i by simp
  have "distinct(?vs1@?vs2)" using dist unfolding append_take_drop_id by simp
  from subst_extend_absorb[OF this len,of "\<sigma>"]
    have map: "map (\<lambda>t. t\<cdot>?\<sigma>) (?ss1@?ss2) = ss1@ss2" unfolding take_map drop_map map_append .
  from len1 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss1 = ss1" unfolding map_append by auto
  moreover from len2 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss2 = ss2" unfolding map_append by auto
  ultimately show ?thesis by simp
qed

lemma distinct_map_Var:
  assumes "distinct xs" shows "distinct (map Var xs)"
  using assms by (induct xs) auto

fun the_Var :: "('f, 'v) term \<Rightarrow> 'v" where
  "the_Var (Var x) = x"

lemma Var_the_Var_ident[simp]:
  "is_Var t \<Longrightarrow> Var (the_Var t) = t"
  by (cases t) simp_all

lemma terms_to_vars:
  assumes "\<forall>t\<in>set ts. is_Var t"
  shows "\<Union>set (map vars_term ts) = set (map the_Var ts)"
  using assms by (induct ts) auto

lemma Var_the_Var_id:
  assumes "\<forall>t\<in>set ts. is_Var t"
  shows "map Var (map the_Var ts) = ts"
  using assms by (induct ts) auto

lemma distinct_the_vars:
  assumes "\<forall>t\<in>set ts. is_Var t"
    and "distinct ts"
  shows "distinct (map the_Var ts)"
  using assms by (induct ts) auto

lemma weak_match_map_term: "weak_match t s \<Longrightarrow> weak_match (map_term g t) (map_term g s)"
proof (induct s arbitrary: t)
  case (Fun f ss t)
  from Fun(2) obtain ts where t: "t = Fun f ts" by (cases t, auto)
  from Fun(1)[unfolded set_conv_nth] Fun(2)[unfolded t]
  show ?case unfolding t by force
qed simp


lemma linear_map_term[simp]: "linear_term (map_term f t) = linear_term t"
  by (induct t) simp_all

fun
  map_vars :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'w) term"
where
  "map_vars m (Var x) = (Var (m x))" |
  "map_vars m (Fun f ts) = Fun f (map (map_vars m) ts)"

lemma vars_term_map_vars: "vars_term (map_vars f t) = f ` vars_term t"
  by (induct t, auto)

lemma map_vars_compose:
  "map_vars m1 (map_vars m2 t) = map_vars (m1 o m2) t" 
by (induct t, auto)

lemma map_vars_id[simp]: "map_vars id t = t"
by (induct t) (auto intro: map_idI)

lemma map_vars_id'[simp]: "map_vars (\<lambda>x. x) t = t"
by (induct t) (auto intro: map_idI)

lemma var_type_conversion: 
  assumes inf: "infinite (UNIV :: 'v set)"
  and fin: "finite (T :: ('f,'w)term set)"
  shows "\<exists> (\<sigma> :: ('f,'w,'v)gsubst) \<tau>. \<forall> t \<in> T. t = t \<cdot> \<sigma> \<cdot> \<tau>"
proof -
  let ?m = "map_vars f :: ('f,'v)term \<Rightarrow> ('f,'w)term"
  obtain V where V: "V = \<Union> vars_term ` T" by auto
  have fin: "finite V" unfolding V
    by (rule, rule, rule fin,
      insert vars_term_finite, auto)
  from finite_imp_inj_to_nat_seg[OF fin] obtain to_nat :: "'w \<Rightarrow> nat" and n :: nat
    where to_nat: "to_nat ` V = {i. i < n}" "inj_on to_nat V" by blast+
  from infinite_countable_subset[OF inf] obtain of_nat :: "nat \<Rightarrow> 'v" where
    of_nat: "range of_nat \<subseteq> UNIV" "inj of_nat"  by auto
  let ?conv = "\<lambda> v. of_nat (to_nat v)"
  have inj: "inj_on ?conv V" using of_nat(2) to_nat(2) unfolding inj_on_def by auto
  let ?rev = "the_inv_into V ?conv"
  note rev = the_inv_into_f_eq[OF inj]
  obtain \<sigma> where \<sigma>: "\<sigma> = (\<lambda> v. Var (?conv v) :: ('f,'v)term)" by simp
  obtain \<tau> where \<tau>: "\<tau> = (\<lambda> v. Var (?rev v) :: ('f,'w)term)" by simp
  show ?thesis
  proof (rule exI[of _ \<sigma>], rule exI[of _ \<tau>], intro ballI)
    fix t
    assume t: "t \<in> T"
    have "t \<cdot> \<sigma> \<cdot> \<tau> = t \<cdot> (\<sigma> \<circ>s \<tau>)" by simp 
    also have "... = t \<cdot> Var"
    proof (rule term_subst_eq)
      fix x
      assume "x \<in> vars_term t"
      with t have x: "x \<in> V" unfolding V by auto
      show "(\<sigma> \<circ>s \<tau>) x = Var x" unfolding \<sigma> \<tau> subst_compose_def
        subst_apply_term.simps term.simps
        by (rule rev[OF refl x])
    qed
    finally show "t = t \<cdot> \<sigma> \<cdot> \<tau>" by simp
  qed
qed

fun
  merge_substs :: "('f, 'u, 'v) gsubst \<Rightarrow> ('f, 'w, 'v) gsubst \<Rightarrow> ('f, 'u + 'w, 'v) gsubst"
where
  "merge_substs \<sigma> \<tau> = (\<lambda>x.
    (case x of
      Inl y \<Rightarrow> \<sigma> y
    | Inr y \<Rightarrow> \<tau> y))"

lemma merge_substs_left:
  "map_vars Inl s \<cdot> (merge_substs \<sigma> \<delta>) = s \<cdot> \<sigma>"
  by (induct s) auto

lemma merge_substs_right:
  "map_vars Inr s \<cdot> (merge_substs \<sigma> \<delta>) = s \<cdot> \<delta>"
  by (induct s) auto


fun map_vars_subst_ran :: "('w \<Rightarrow> 'u) \<Rightarrow> ('f, 'v,'w) gsubst \<Rightarrow> ('f, 'v,'u) gsubst" where
  "map_vars_subst_ran m \<sigma> = (\<lambda>v. map_vars m (\<sigma> v))"

fun map_subst_domain :: "('v \<Rightarrow> 'u) \<Rightarrow> ('f, 'u,'w) gsubst \<Rightarrow> ('f, 'v,'w) gsubst" where
  "map_subst_domain m \<sigma> = (\<lambda>v. \<sigma> (m v))"

lemma map_vars_subst_ran: 
  shows "map_vars m (t \<cdot> \<sigma>) = t \<cdot> map_vars_subst_ran m \<sigma>"
  by (induct t) (auto)

lemma map_subst_domain: 
  shows "map_vars m t \<cdot> \<sigma> = t \<cdot> map_subst_domain m \<sigma>"
  by (induct t) (auto)

lemma size_subst: "size t \<le> size (t \<cdot> \<sigma>)"
proof (induct t)
  case (Var x)
  thus ?case by auto
next
  case (Fun f ss)
  thus ?case 
    by (simp add: o_def, induct ss, force+)
qed

lemma Fun_Nil_supt[elim!]: "Fun f [] \<rhd> t \<Longrightarrow> P" by auto

lemma map_vars_vars_term: assumes "\<And> x. x \<in> vars_term t \<Longrightarrow> f x = g x"
  shows "map_vars f t = map_vars g t"
  using assms
proof (induct t, simp)
  case (Fun h ts)
  {
    fix t
    assume t: "t \<in> set ts"
    with Fun(2) have "\<And> x. x \<in> vars_term t \<Longrightarrow> f x = g x" 
      by auto 
    from Fun(1)[OF t this] have "map_vars f t = map_vars g t" by simp
  }
  thus ?case by auto
qed


lemma map_term_ctxt_decomp: assumes "map_term fg t = C\<langle>s\<rangle>"
  shows "\<exists> D u. C = map_ctxt fg D \<and> s = map_term fg u \<and> t = D\<langle>u\<rangle>"
  using assms
proof (induct C arbitrary: t)
  case Hole
  show ?case
    by (rule exI[of _ Hole], rule exI[of _ t], insert Hole, auto)
next
  case (More g bef C aft)
  from More(2) obtain f ts where t: "t = Fun f ts" by (cases t, auto)
  from More(2)[unfolded t] have f: "fg f = g" and ts: "map (map_term fg) ts = bef @ C\<langle>s\<rangle> # aft" (is "?ts = ?bca") by auto
  from ts have "length ?ts = length ?bca" by auto
  hence len: "length ts = length ?bca" by auto
  note id = ts[unfolded map_nth_eq_conv[OF len], THEN spec, THEN mp]
  let ?i = "length bef"
  from len have i: "?i < length ts" by auto
  from id[of ?i] have "map_term fg (ts ! ?i) = C\<langle>s\<rangle>" by auto
  from More(1)[OF this] obtain D u where D: "C = map_ctxt fg D" and
    u: "s = map_term fg u" and id: "ts ! ?i = D\<langle>u\<rangle>" by auto
  from ts have "take ?i ?ts = take ?i ?bca" by simp
  also have "... = bef" by simp
  finally have bef: "map (map_term fg) (take ?i ts) = bef" by (simp add: take_map)
  from ts have "drop (Suc ?i) ?ts = drop (Suc ?i) ?bca" by simp
  also have "... = aft" by simp
  finally have aft: "map (map_term fg) (drop (Suc ?i) ts) = aft" by (simp add:drop_map)
  let ?bda = "take ?i ts @ D\<langle>u\<rangle> # drop (Suc ?i) ts"
  show ?case
  proof (rule exI[of _ "More f (take ?i ts) D (drop (Suc ?i) ts)"],
      rule exI[of _ u], simp add: u f D bef aft t)
    have "ts = take ?i ts @ ts ! ?i # drop (Suc ?i) ts"        
      by (rule id_take_nth_drop[OF i])
    also have "... = ?bda" by (simp add: id)
    finally show "ts = ?bda" .
  qed
qed


lemma ground_subst[simp]: "ground (t \<cdot> \<sigma>) = (\<forall>x \<in> vars_term t. ground (\<sigma> x))"
  by (induct t, auto)

fun ground_ctxt :: "('f,'v)ctxt \<Rightarrow> bool" where 
  "ground_ctxt Hole = True"
| "ground_ctxt (More f ss1 C ss2) =
    ((\<forall>s\<in>set ss1. ground s) \<and> (\<forall>s\<in>set ss2. ground s) \<and> ground_ctxt C)"

lemma ground_ctxt_apply[simp]: "ground (C\<langle>t\<rangle>) = (ground_ctxt C \<and> ground t)"
  by (induct C, auto)


lemma funas_term_map_vars[simp]: "funas_term (map_vars \<tau> t) = funas_term t"
  by (induct t, auto)

fun map_term_enum :: "('f \<Rightarrow> 'g list) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('g, 'v) term list" where
  "map_term_enum fgs (Var x) = [Var x]"
| "map_term_enum fgs (Fun f ts) = (
    let lts = map (map_term_enum fgs) ts;
      ss = concat_lists lts;
      gs = fgs f
    in concat (map (\<lambda>g. map (Fun g) ss) gs))"

lemma map_term_enum: assumes gf: "\<And> f g. g \<in> set (fgs f) \<Longrightarrow> gf g = f"
  shows "set (map_term_enum fgs t) = {u. map_term gf u = t \<and> (\<forall> g n. (g,n) \<in> funas_term u \<longrightarrow> g \<in> set (fgs (gf g)))}"
proof (induct t)
  case (Var x)
  show ?case (is "_ = ?R")
  proof -
    {
      fix t 
      assume "t \<in> ?R"
      hence "t = Var x" by (cases t, auto)
    } 
    thus ?thesis by auto
  qed
next
  case (Fun f ts)
  show ?case (is "?L = ?R")
  proof -
    {
      fix i
      assume "i < length ts"
      hence "ts ! i \<in> set ts" by auto
      note Fun[OF this]
    } note ind = this
    let ?cf = "\<lambda> u. (\<forall> g n. (g,n) \<in> funas_term u \<longrightarrow> g \<in> set (fgs (gf g)))"
    have id: "?L = {Fun g ss | g ss. g \<in> set (fgs f) \<and> length ss = length ts \<and> (\<forall>i<length ts. ss ! i \<in> set (map_term_enum fgs (ts ! i)))}" (is "_ = ?M1") by auto    
    have "?R = {Fun g ss | g ss. map_term gf (Fun g ss) = Fun f ts \<and> ?cf (Fun g ss)}" (is "_ = ?M2")
    proof -
      {
        fix u
        assume u: "u \<in> ?R"
        then obtain g ss where "u = Fun g ss" by (cases u, auto)
        with u have "u \<in> ?M2" by auto
      }
      hence "?R \<subseteq> ?M2" by auto
      moreover have "?M2 \<subseteq> ?R" by blast
      finally show ?thesis by auto
    qed
    also have "... = ?M1"
    proof -
      {
        fix u
        assume "u \<in> ?M1"
        then obtain g ss where u: "u = Fun g ss" and g: "g \<in> set (fgs f)" and 
          len: "length ss = length ts" and rec: "\<And> i. i < length ts \<Longrightarrow> ss ! i \<in> set (map_term_enum fgs (ts ! i))" by auto
        from gf[OF g] have gf: "gf g = f" by simp
        {
          fix i
          assume i: "i < length ts"
          from ind[OF i] rec[OF i]
          have "map_term gf (ss ! i) = ts ! i" by auto
        } note ssts = this
        have "map (map_term gf) ss = ts"
          by (unfold map_nth_eq_conv[OF len], insert ssts, auto)
        with gf
        have mt: "map_term gf (Fun g ss) = Fun f ts" by auto        
        have "u \<in> ?M2"
        proof (rule, rule, rule, rule, rule u, rule, rule mt, intro allI impI)
          fix h n
          assume h: "(h,n) \<in> funas_term (Fun g ss)"          
          show "h \<in> set (fgs (gf h))" 
          proof (cases "(h,n) = (g,length ss)")
            case True
            hence "h = g" by auto
            with g gf show ?thesis by auto
          next
            case False
            with h obtain s where s: "s \<in> set ss" and h: "(h,n) \<in> funas_term s" by auto
            from s[unfolded set_conv_nth] obtain i where i: "i < length ss" and si: "s = ss ! i" by force
            from i len have i': "i < length ts" by auto
            from ind[OF i'] rec[OF i'] h[unfolded si] show ?thesis by auto
          qed
        qed      
      }
      hence m1m2: "?M1 \<subseteq> ?M2" by blast
      {
        fix u
        assume u: "u \<in> ?M2"
        then obtain g ss where u: "u = Fun g ss"
          and map: "map_term gf (Fun g ss) = Fun f ts"
          and c: "?cf (Fun g ss)"
          by blast
        from map have len: "length ss = length ts" by auto
        from map have g: "gf g = f" by auto
        from map have map: "map (map_term gf) ss = ts" by auto
        from c have g2: "g \<in> set (fgs f)" using g by auto
        have "u \<in> ?M1"
        proof (intro CollectI exI conjI allI impI, rule u, rule g2, rule len)
          fix i 
          assume i: "i < length ts"
          with map[unfolded map_nth_eq_conv[OF len]]
          have id: "map_term gf (ss ! i) = ts ! i" by auto          
          from i len have si: "ss ! i \<in> set ss" by auto
          show "ss ! i \<in> set (map_term_enum fgs (ts ! i))"
            unfolding ind[OF i] 
          proof (intro CollectI conjI impI allI, rule id)
            fix g n
            assume "(g,n) \<in> funas_term (ss ! i)"
            with c si
            show "g \<in> set (fgs (gf g))" by auto
          qed
        qed
      }
      hence m2m1: "?M2 \<subseteq> ?M1" by blast
      show "?M2 = ?M1"
        by (rule, rule m2m1, rule m1m2)
    qed
    finally show ?case unfolding id by simp
  qed
qed

declare map_term_enum.simps[simp del]          

lemma funs_term_funas_term: "funs_term t = fst ` (funas_term t)"
  by (induct t) auto

lemma funas_term_map_term:
  "funas_term (map_term fg t) = (\<lambda> (f,n). (fg f,n)) ` (funas_term t)"
  by (induct t) auto

lemma supt_imp_arg_or_supt_of_arg:
  assumes "Fun f ss \<rhd> t"
  shows "t \<in> set ss \<or> (\<exists>s \<in> set ss. s \<rhd> t)"
  using assms by (rule supt.cases) auto

lemma vars_term_supteq:
 shows "x \<in> vars_term (t::('f, 'v) term) \<Longrightarrow> t \<unrhd> (Var x)"
   and "x \<in> vars_term ((Fun f ss)::('f, 'v) term) \<Longrightarrow> Fun f ss \<unrhd> (Var x)"
proof (induct t and ss rule: term.inducts)
 case (Var x) thus ?case using assms by auto
next
 case (Fun g ts)
 from `x \<in> vars_term(Fun g ts)` have "x \<in> vars_term(Fun f ts)" by simp
 with Fun have "Fun f ts \<unrhd> Var x" by simp
 have "Fun f ts \<noteq> Var x" by auto
 with `Fun f ts \<unrhd> Var x` have "Fun f ts \<rhd> Var x" by cases auto
 hence "Fun g ts \<rhd> Var x" by (cases rule: supt.cases) (auto intro: supt.intros)
 thus ?case by auto
next
 case Nil_term thus ?case by auto
next
 case (Cons_term u us)
 hence "x \<in> vars_term u \<or>  x \<in> \<Union>(set (map vars_term us))" by auto
 thus ?case
 proof
  assume "x \<in> vars_term u" with Cons_term show ?case by auto
 next
  assume "x \<in> \<Union>(set (map vars_term us))" 
  hence "x \<in> vars_term(Fun f us)" by auto
  with Cons_term have "Fun f us \<unrhd> Var x" by auto
  have "Fun f us \<noteq> Var x" by auto
  with `Fun f us \<unrhd> Var x` have "Fun f us \<rhd> Var x" by cases auto
  hence "Var x \<in> set us \<or> (\<exists>u \<in> set us. u \<rhd> Var x)" by cases auto
  thus ?case
  proof
   assume "Var x \<in> set us" hence "Var x \<in> set (u#us)" by simp
   thus ?thesis by auto
  next
   assume "\<exists>u \<in> set us. u \<rhd> Var x"
   then obtain v where "v \<in> set us" and "v \<rhd> Var x" by auto
   hence "v \<in> set (u#us)" by auto
   from `v \<rhd> Var x` have "v \<unrhd> Var x" by auto
   with `v \<in> set (u#us)` show ?thesis by auto
  qed
 qed
qed

lemma supt_Fun_imp_arg_supteq:
  assumes "Fun f ss \<rhd> t" shows "\<exists>s \<in> set ss. s \<unrhd> t"
  using assms by (cases rule: supt.cases) auto

lemma subt_iff_eq_or_subt_of_arg: 
  assumes "s = Fun f ss"
  shows "{t. s \<unrhd> t} = ((\<Union>u \<in> set ss. {t. u \<unrhd> t})\<union>{s})"
using assms proof (induct s)
  case (Var x) thus ?case by auto
next
  case (Fun g ts)
  hence "g = f" and "ts = ss" by auto
  show ?case
  proof
    show "{a. Fun g ts \<unrhd> a} \<subseteq> (\<Union>u\<in>set ss. {a. u \<unrhd> a}) \<union> {Fun g ts}"
    proof
      fix x
      assume "x \<in> {a. Fun g ts \<unrhd> a}"
      hence "Fun g ts \<unrhd> x" by simp
      hence "Fun g ts \<rhd> x \<or> Fun g ts = x" by auto
      thus "x \<in> (\<Union>u\<in>set ss. {a. u \<unrhd>a}) \<union> {Fun g ts}"
      proof
        assume "Fun g ts \<rhd> x"
        then obtain u where "u \<in> set ts" and "u \<unrhd> x" using supt_Fun_imp_arg_supteq by best
        hence "x \<in> {a. u \<unrhd> a}" by simp
        with `u \<in> set ts` have "x \<in> (\<Union>u\<in>set ts. {a. u \<unrhd> a})" by auto
        thus ?thesis unfolding `ts = ss` by simp
      next
        assume "Fun g ts = x" thus ?thesis by simp
      qed
    qed
  next
    show "(\<Union>u\<in>set ss. {a. u \<unrhd> a}) \<union> {Fun g ts} \<subseteq> {a. Fun g ts \<unrhd> a}"
    proof
      fix x
      assume "x \<in> (\<Union>u\<in>set ss. {a. u \<unrhd> a}) \<union> {Fun g ts}"
      hence "x \<in> (\<Union>u\<in>set ss. {a. u \<unrhd> a}) \<or> x = Fun g ts" by auto
      thus "x \<in> {a. Fun g ts \<unrhd> a}"
      proof
        assume "x \<in> (\<Union>u\<in>set ss. {a. u \<unrhd> a})"
        then obtain u where "u \<in> set ss" and "u \<unrhd> x" by auto
        thus ?thesis unfolding `ts = ss` by auto
      next
        assume "x = Fun g ts" thus ?thesis by auto
      qed
    qed
  qed
qed

text {* The set of subterms of a term is finite. *}
lemma finite_subterms: "finite {s. t \<unrhd> s}"
proof (induct t)
  case (Var x)
  hence "\<And>s. (Var x \<unrhd> s) = (Var x = s)" using supteq.cases by best
  thus ?case unfolding `!!s. (Var x \<unrhd> s) = (Var x = s)` by simp
next
  case (Fun f ss)
  have "Fun f ss = Fun f ss" by simp
  from Fun show ?case
  unfolding subt_iff_eq_or_subt_of_arg[OF `Fun f ss = Fun f ss`] by auto
qed

lemma Fun_supteq: "Fun f ts \<unrhd> u \<longleftrightarrow> Fun f ts = u \<or> (\<exists>t\<in>set ts. t \<unrhd> u)"
  using subt_iff_eq_or_subt_of_arg[of "Fun f ts" f ts] by auto

lemma subst_ctxt_distr: "s = C\<langle>t\<rangle>\<cdot>\<sigma> \<Longrightarrow> \<exists>D. s = D\<langle>t\<cdot>\<sigma>\<rangle>"
  using subst_apply_term_ctxt_apply_distrib by auto

lemma ctxt_of_pos_term_subst: "p \<in> poss t \<Longrightarrow> ctxt_of_pos_term p (t \<cdot> \<sigma>) = ctxt_of_pos_term p t \<cdot>c \<sigma>" 
proof (induct p arbitrary: t)
  case Empty
  show ?case by simp
next
  case (PCons i p t)
  then obtain f ts where t: "t = Fun f ts" and i: "i < length ts" and p: "p \<in> poss (ts ! i)" by (cases t, auto)
  note id = id_take_nth_drop[OF i, symmetric]
  with t have t: "t = Fun f (take i ts @ ts ! i # drop (Suc i) ts)" by auto
  from i have i': "min (length ts) i = i" by simp
  show ?case unfolding t using PCons(1)[OF p, symmetric] i' 
    by (simp add: id, insert i, auto simp: take_map drop_map)
qed

lemma subt_at_ctxt_of_pos_term: assumes t: "t = (ctxt_of_pos_term p t)\<langle>u\<rangle>" and p: "p \<in> poss t"
  shows "u = t|_p"
proof -
  let ?C = "ctxt_of_pos_term p t"
  from t ctxt_supt_id[OF p] have "?C\<langle>u\<rangle> = ?C\<langle>t |_ p\<rangle>" by simp
  thus ?thesis unfolding ctxt_eq .
qed

lemma subst_ext:
  assumes "\<forall>x\<in>V. \<sigma> x = \<tau> x" shows "\<sigma> |s V = \<tau> |s V"
proof
  fix x show "(\<sigma> |s V) x = (\<tau> |s V) x" using assms
  unfolding subst_restrict_def by (cases "x \<in> V") auto
qed

fun
  map_vars_ctxt :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f, 'v)ctxt \<Rightarrow> ('f, 'w)ctxt"
where
  "map_vars_ctxt m Hole = Hole" |
  "map_vars_ctxt m (More f bef c aft) = More f (map (map_vars m) bef) (map_vars_ctxt m c) (map (map_vars m) aft)"

lemma map_vars_ctxt_commute:
  "map_vars m (c\<langle>t\<rangle>) = (map_vars_ctxt m c)\<langle>map_vars m t\<rangle>"
  by (induct c) auto

lemma map_vars_inj_compose:
  assumes inj: "\<And> x. n (m x) = x"
  shows "map_vars n (map_vars m t) = t"
  unfolding map_vars_compose o_def inj by auto

lemma inj_map_vars_the_inv:
  assumes "inj f"
  shows "map_vars (the_inv f) (map_vars f t) = t"
  unfolding map_vars_compose o_def the_inv_f_f[OF assms]
  by simp

lemma map_vars_ctxt_subst: "map_vars_ctxt m (C \<cdot>c \<sigma>) = C \<cdot>c map_vars_subst_ran m \<sigma>"
  by (induct C, auto simp: map_vars_subst_ran)

lemma poss_map_vars[simp]: "poss (map_vars f t) = poss t"
  by (induct t, auto)

lemma map_vars_subt_at[simp]: 
  "p \<in> poss t \<Longrightarrow> map_vars f (t |_ p) = (map_vars f t) |_ p"
proof (induct p arbitrary: t)
  case Empty show ?case by auto
next
  case (PCons i p t)
  from PCons(2) obtain g ts where t: "t = Fun g ts" by (cases t, auto)
  from PCons show ?case unfolding t by auto
qed

lemma hole_pos_subst[simp]: "hole_pos (C \<cdot>c \<sigma>) = hole_pos C"
  by (induct C, auto)

lemma hole_pos_ctxt_compose[simp]: "hole_pos (C \<circ>c D) = hole_pos C <#> hole_pos D"
  by (induct C, auto)

lemma subst_left_right: "t \<cdot> \<mu>^^n \<cdot> \<mu> = t \<cdot> \<mu> \<cdot> \<mu>^^n"
proof -
  have "t \<cdot> \<mu> ^^ n \<cdot> \<mu> = t \<cdot> (\<mu> ^^ n \<circ>s \<mu>)" by simp
  also have "... = t \<cdot> (\<mu> \<circ>s \<mu> ^^ n)"
    using subst_power_compose_distrib[of n "Suc 0" \<mu>] by auto
  finally show ?thesis by simp
qed

lemma subst_right_left: "t \<cdot> \<mu> \<cdot> \<mu>^^n = t \<cdot> \<mu>^^n \<cdot> \<mu>" unfolding subst_left_right ..

lemma subt_at_id_imp_eps: assumes p: "p \<in> poss t" and id: "t |_ p = t"
  shows "p = \<epsilon>"
proof (cases p)
  case (PCons i q)
  from subt_at_nepos_imp_supt[OF p[unfolded PCons], unfolded PCons[symmetric]
    , unfolded id] have False by simp
  thus ?thesis by auto
qed simp

lemma pos_into_subst:
  assumes t: "t \<cdot> \<sigma> = s" and p: "p \<in> poss s" and nt: "\<not> (p \<in> poss t \<and> is_Fun (t |_ p))"
  shows "\<exists>q q' x. p = q <#> q' \<and> q \<in> poss t \<and> t |_ q = Var x"
  using p nt unfolding t[symmetric]
proof (induct t arbitrary: p)
  case (Var x)
  show ?case
    by (rule exI[of _ \<epsilon>], rule exI[of _ p], rule exI[of _ x], insert Var, auto)
next
  case (Fun f ts)
  from Fun(3) obtain i q where p: "p = i <# q" by (cases p, auto)
  note Fun = Fun[unfolded p]
  from Fun(2) have i: "i < length ts" and q: "q \<in> poss (ts ! i \<cdot> \<sigma>)" by auto
  hence mem: "ts ! i \<in> set ts" by auto
  from Fun(3) i have "\<not> (q \<in> poss (ts ! i) \<and> is_Fun (ts ! i |_ q))" by auto
  from Fun(1)[OF mem q this]
  obtain r r' x where q: "q = r <#> r' \<and> r \<in> poss (ts ! i) \<and> ts ! i |_ r = Var x" by blast  
  show ?case 
    by (rule exI[of _ "i <# r"], rule exI[of _ r'], rule exI[of _ x],
    unfold p, insert i q, auto)
qed

abbreviation (input) replace_at where "replace_at t p s \<equiv> (ctxt_of_pos_term p t)\<langle>s\<rangle>"

lemma context_of_pos_term_append: assumes "p \<in> poss t" shows "ctxt_of_pos_term (p <#> q) t = ctxt_of_pos_term p t \<circ>c ctxt_of_pos_term q (t |_ p)"
  using assms
proof (induct p arbitrary: t)
  case Empty show ?case by simp
next
  case (PCons i p t)
  from PCons(2) obtain f ts where t: "t = Fun f ts" and i: "i < length ts" and p: "p \<in> poss (ts ! i)" by (cases t, auto)
  from PCons(1)[OF p]
  show ?case unfolding t using i by auto
qed

lemma parallel_replace_at: assumes parallel: "p1 \<bottom> p2"
  and p1: "p1 \<in> poss t"
  and p2: "p2 \<in> poss t"
  shows "replace_at (replace_at t p1 s1) p2 s2 = replace_at (replace_at t p2 s2) p1 s1"
proof -
  from parallel_pos_prefix[OF parallel]
  obtain p i j q1 q2 where p1_id: "p1 = p <#> i <# q1" and p2_id: "p2 = p <#> j <# q2"
    and ij: "i \<noteq> j" by blast
  from p1 p2 show ?thesis unfolding p1_id p2_id
  proof (induct p arbitrary: t)
    case (PCons k p)
    from PCons(3) obtain f ts where t: "t = Fun f ts" and k: "k < length ts" by (cases t, auto)
    note PCons = PCons[unfolded t]
    let ?p1 = "p <#> i <# q1" let ?p2 = "p <#> j <# q2"
    from PCons(2) PCons(3) have "?p1 \<in> poss (ts ! k)" "?p2 \<in> poss (ts ! k)" by auto
    from PCons(1)[OF this] have rec: "replace_at (replace_at (ts ! k) ?p1 s1) ?p2 s2 = replace_at (replace_at (ts ! k) ?p2 s2) ?p1 s1" .
    from k have min: "min (length ts) k = k" by simp
    show ?case unfolding t using rec min k
      by (simp add: nth_append)
  next
    case Empty
    from Empty(2) obtain f ts where t: "t = Fun f ts" and j: "j < length ts" by (cases t, auto)
    note Empty = Empty[unfolded t]
    from Empty have i: "i < length ts" by auto
    let ?p1 = "i <# q1"
    let ?p2 = "j <# q2"
    let ?s1 = "replace_at (ts ! i) q1 s1"
    let ?s2 = "replace_at (ts ! j) q2 s2"
    let ?ts1 = "ts[i := ?s1]" 
    let ?ts2 = "ts[j := ?s2]" 
    from j have j': "j < length ?ts1" by simp
    from i have i': "i < length ?ts2" by simp
    have "replace_at (replace_at t ?p1 s1) ?p2 s2 = replace_at (Fun f ?ts1) ?p2 s2" unfolding t upd_conv_take_nth_drop[OF i] by simp
    also have "... = Fun f (?ts1[j := ?s2])" 
      unfolding upd_conv_take_nth_drop[OF j'] using ij by simp
    also have "... = Fun f (?ts2[i := ?s1])" using list_update_swap[OF ij]
      by simp
    also have "... = replace_at (Fun f ?ts2) ?p1 s1" 
      unfolding upd_conv_take_nth_drop[OF i'] using ij by simp
    also have "... = replace_at (replace_at t ?p2 s2) ?p1 s1" unfolding t
      upd_conv_take_nth_drop[OF j] by simp
    finally show ?case by simp
  qed
qed

lemma parallel_replace_at_subt_at: assumes parallel: "p1 \<bottom> p2"
  and p1: "p1 \<in> poss t"
  and p2: "p2 \<in> poss t"
  shows " (replace_at t p1 s1) |_ p2 = t |_ p2"
proof -
  from parallel_pos_prefix[OF parallel]
  obtain p i j q1 q2 where p1_id: "p1 = p <#> i <# q1" and p2_id: "p2 = p <#> j <# q2"
    and ij: "i \<noteq> j" by blast
  from p1 p2 show ?thesis unfolding p1_id p2_id
  proof (induct p arbitrary: t)
    case (PCons k p)
    from PCons(3) obtain f ts where t: "t = Fun f ts" and k: "k < length ts" by (cases t, auto)
    note PCons = PCons[unfolded t]
    let ?p1 = "p <#> i <# q1" let ?p2 = "p <#> j <# q2"
    from PCons(2) PCons(3) have "?p1 \<in> poss (ts ! k)" "?p2 \<in> poss (ts ! k)" by auto
    from PCons(1)[OF this] have rec: "(replace_at (ts ! k) ?p1 s1) |_ ?p2 = (ts ! k) |_ ?p2" .
    from k have min: "min (length ts) k = k" by simp
    show ?case unfolding t using rec min k
      by (simp add: nth_append)
  next
    case Empty
    from Empty(2) obtain f ts where t: "t = Fun f ts" and j: "j < length ts" by (cases t, auto)
    note Empty = Empty[unfolded t]
    from Empty have i: "i < length ts" by auto
    let ?p1 = "i <# q1"
    let ?p2 = "j <# q2"
    let ?s1 = "replace_at (ts ! i) q1 s1"
    let ?ts1 = "ts[i := ?s1]" 
    from j have j': "j < length ?ts1" by simp
    have "(replace_at t ?p1 s1) |_ ?p2 = (Fun f ?ts1) |_ ?p2" unfolding t upd_conv_take_nth_drop[OF i] by simp
    also have "... = ts ! j |_ q2" using ij by simp
    finally show ?case unfolding t by simp
  qed
qed

lemma parallel_poss_replace_at: assumes parallel: "p1 \<bottom> p2"
  and p1: "p1 \<in> poss t"
  shows "(p2 \<in> poss (replace_at t p1 s1)) = (p2 \<in> poss t)"
proof -
  from parallel_pos_prefix[OF parallel]
  obtain p i j q1 q2 where p1_id: "p1 = p <#> i <# q1" and p2_id: "p2 = p <#> j <# q2"
    and ij: "i \<noteq> j" by blast
  from p1 show ?thesis unfolding p1_id p2_id
  proof (induct p arbitrary: t)
    case (PCons k p)
    from PCons(2) obtain f ts where t: "t = Fun f ts" and k: "k < length ts" by (cases t, auto)
    note PCons = PCons[unfolded t]
    let ?p1 = "p <#> i <# q1" let ?p2 = "p <#> j <# q2"
    from PCons(2) have "?p1 \<in> poss (ts ! k)" by auto
    from PCons(1)[OF this] have rec: "(?p2 \<in> poss (replace_at (ts ! k) ?p1 s1)) = (?p2 \<in> poss (ts ! k))" .
    from k have min: "min (length ts) k = k" by simp
    show ?case unfolding t using rec min k 
      by (auto simp: nth_append)
  next
    case Empty
    then obtain f ts where t: "t = Fun f ts" and i: "i < length ts" by (cases t, auto)
    let ?p1 = "i <# q1"
    let ?s1 = "replace_at (ts ! i) q1 s1"
    have "replace_at t ?p1 s1 = Fun f (ts[i := ?s1])" unfolding t upd_conv_take_nth_drop[OF i] by simp
    thus ?case unfolding t using ij by auto
  qed
qed

lemma replace_at_subt_at: "p \<in> poss t \<Longrightarrow> (replace_at t p s) |_ p = s"
  by (metis hole_pos_ctxt_of_pos_term subt_at_hole_pos)

lemma replace_at_below_poss: assumes p: "p' \<in> poss t" and le: "p \<le> p'"
  shows "p \<in> poss (replace_at t p' s)"
proof -
  from le obtain p'' where p'': "p' = p <#> p''" unfolding less_eq_pos_def by auto
  from p show ?thesis  unfolding p''
    by (metis hole_pos_ctxt_of_pos_term hole_pos_poss poss_append_poss)
qed

lemma ctxt_of_pos_term_replace_at_below: assumes p: "p \<in> poss t" and le: "p \<le> p'"
  shows "ctxt_of_pos_term p (replace_at t p' u) = ctxt_of_pos_term p t"
proof -
  from le obtain p'' where p': "p' = p <#> p''" unfolding less_eq_pos_def by auto
  from p show ?thesis unfolding p'
  proof (induct p arbitrary: t)
    case (PCons i p)
    from PCons(2) obtain f ts where t: "t = Fun f ts" and i: "i < length ts" and p: "p \<in> poss (ts ! i)"
      by (cases t, auto)
    from i have min: "min (length ts) i = i" by simp
    show ?case unfolding t using PCons(1)[OF p] i by (auto simp: nth_append min)
  next
    case Empty show ?case by simp
  qed 
qed

fun ctxt_to_term_list :: "('f,'v)ctxt \<Rightarrow> ('f,'v)term list"
  where "ctxt_to_term_list Hole = []"
     |  "ctxt_to_term_list (More f bef C aft) = ctxt_to_term_list C @ bef @ aft"

lemma ctxt_to_term_list_supt: "t \<in> set (ctxt_to_term_list C) \<Longrightarrow> C\<langle>s\<rangle> \<rhd> t"
proof (induct C)
  case Hole thus ?case by auto
next
  case (More f bef C aft)
  from More(2) have choice: "t \<in> set (ctxt_to_term_list C) \<or> t \<in> set bef \<or> t \<in> set aft" by simp
  {
    assume "t \<in> set bef \<or> t \<in> set aft"
    hence "t \<in> set (bef @ C\<langle>s\<rangle> # aft)" by auto
    hence ?case by auto
  } 
  moreover
  {
    assume t: "t \<in> set (ctxt_to_term_list C)"
    have "(More f bef C aft)\<langle>s\<rangle> \<rhd> C\<langle>s\<rangle>" by auto
    moreover have "C\<langle>s\<rangle> \<rhd> t" 
      by (rule More(1)[OF t])
    ultimately have ?case
      by (rule supt_trans)
  }
  ultimately show ?case using choice by auto
qed

lemma var_subt_imp_subt_in_vars: "r \<unrhd> Var x \<Longrightarrow> x \<in> vars_term r"
proof (induct r rule: term_induct)
 case (Var y) hence "x = y" by (cases rule: supteq.cases) auto
 thus ?case by simp
next
 case (Fun f ss)
 from `Fun f ss \<unrhd> Var x` have "(Fun f ss = Var x) \<or> (Fun f ss \<rhd> Var x)" by auto
 thus ?case
 proof
  assume "Fun f ss = Var x" thus ?thesis by auto
 next
  assume "Fun f ss \<rhd> Var x"
  then obtain s where "s \<in> set ss" and "s \<unrhd> Var x" using supt_Fun_imp_arg_supteq by best
  with Fun have "s \<unrhd> Var x \<Longrightarrow> x \<in> vars_term s" by best
  with `s \<unrhd> Var x` have "x \<in> vars_term s" by simp
  with `s \<in> set ss` show ?thesis by auto
 qed
qed

lemma linear_term_replace_in_subst: assumes lin: "linear_term t"
  and p: "p \<in> poss t"
  and x: "t |_ p = Var x"
  and \<tau>y: "\<And> y. y \<in> vars_term t \<Longrightarrow> y \<noteq> x \<Longrightarrow> \<sigma> y = \<tau> y"
  and \<tau>x: "\<tau> x = s"
  shows "replace_at (t \<cdot> \<sigma>) p s = t \<cdot> \<tau>"
  using p lin x \<tau>y
proof (induct p arbitrary: t)
  case Empty
  thus ?case using \<tau>x by auto
next
  case (PCons i p t)
  from PCons(2) obtain f ts where t: "t = Fun f ts" and i: "i < length ts" 
    and p: "p \<in> poss (ts ! i)" by (cases t, auto)
  note PCons = PCons[unfolded t]
  from PCons have rec: "linear_term (ts ! i)" "ts ! i |_ p = Var x" by auto
  note IH = PCons(1)[OF p rec PCons(5)]
  have id: "replace_at (ts ! i \<cdot> \<sigma>) p s = ts ! i \<cdot> \<tau>" 
    by (rule IH, insert i, auto)
  let ?l = " (take i (map (\<lambda>t. t \<cdot> \<sigma>) ts) @
    (ts ! i \<cdot> \<tau>) # drop (Suc i) (map (\<lambda>t. t \<cdot> \<sigma>) ts))"
  from i have len: "length ts = length ?l" by auto
  show ?case unfolding t subst_apply_term.simps ctxt_of_pos_term.simps ctxt_apply.simps term.simps nth_map[OF i] id
  proof (rule conjI[OF refl], rule sym, unfold map_nth_eq_conv[OF len] len[symmetric], intro allI impI)
    fix j
    assume j: "j < length ts"
    show "ts ! j \<cdot> \<tau> = ?l ! j"
    proof (cases "j = i")
      case True
      with i show ?thesis by (simp add: nth_append)
    next
      case False
      let ?ts = "map (\<lambda> t. t \<cdot> \<sigma>) ts"
      from i j have le: "j \<le> length ?ts" "i \<le> length ?ts" by auto
      from nth_append_take_drop_is_nth_conv[OF le False]
      have "?l ! j = ?ts ! j" by simp
      also have "... = ts ! j \<cdot> \<sigma>" using j by simp
      also have "... = ts ! j \<cdot> \<tau>"
      proof (rule term_subst_eq)
        fix y
        assume y: "y \<in> vars_term (ts ! j)"
        from p have "ts ! i \<unrhd> ts ! i |_ p" by (rule subt_at_imp_supteq)
        also have "ts ! i |_ p = Var x" unfolding rec(2) ..
        finally have x: "x \<in> vars_term (ts ! i)" by (rule var_subt_imp_subt_in_vars)
        from PCons(3) have "is_partition (map vars_term ts)" by simp
        from this[unfolded is_partition_alt is_partition_alt_def, rule_format, of j i] j i False have "vars_term (ts ! i) \<inter> vars_term (ts ! j) = {}" by auto
        with y x have yx: "y \<noteq> x" by auto
        show "\<sigma> y = \<tau> y" 
          by (rule PCons(5)[OF _ yx], insert y j, auto)
      qed
      finally show ?thesis by simp
    qed
  qed
qed



fun instance_term :: "('f,'v)term \<Rightarrow> ('f set,'v)term \<Rightarrow> bool"
  where "instance_term (Var x) (Var y) = (x = y)"
      | "instance_term (Fun f ts) (Fun fs ss) = (f \<in> fs \<and> length ts = length ss \<and> (\<forall> i < length ts. instance_term (ts ! i) (ss ! i)))"
      | "instance_term _ _ = False"

fun
  subt_at_ctxt :: "('f, 'v)ctxt \<Rightarrow> pos \<Rightarrow> ('f, 'v)ctxt" ("(_ |'_c _)" [60])
where
  "C |_c \<epsilon> = C"
| "More f bef C aft |_c i<#p = C |_c p"

lemma subt_at_subt_at_ctxt: assumes "hole_pos C = p <#> q"
  shows "C\<langle>t\<rangle> |_ p = (C |_c p)\<langle>t\<rangle>"
  using assms 
proof (induct p arbitrary: C)
  case Empty
  thus ?case by simp
next
  case (PCons i p)
  then obtain f bef D aft where C: "C = More f bef D aft" by (cases C, auto)
  from PCons(2) have "hole_pos D = p <#> q" unfolding C by simp
  from PCons(1)[OF this] have id: "D |_c p \<langle>t\<rangle> = D\<langle>t\<rangle> |_ p" by simp
  show ?case unfolding C subt_at_ctxt.simps id using PCons(2) C by auto
qed
  
lemma hole_pos_subt_at_ctxt: assumes "hole_pos C = p <#> q"
  shows "hole_pos (C |_c p) = q"
  using assms 
proof (induct p arbitrary: C)
  case Empty
  thus ?case by simp
next
  case (PCons i p)
  then obtain f bef D aft where C: "C = More f bef D aft" by (cases C, auto)
  show ?case unfolding C subt_at_ctxt.simps
    by (rule PCons(1), insert PCons(2) C, auto) 
qed

lemma subt_at_ctxt_compose[simp]: "(C \<circ>c D) |_c hole_pos C = D"
  by (induct C, auto)

lemma split_ctxt: assumes "hole_pos C = p <#> q"
  shows "\<exists> D E. C = D \<circ>c E \<and> hole_pos D = p \<and> hole_pos E = q \<and> E = C |_c p"
  using assms
proof (induct p arbitrary: C)
  case Empty
  show ?case
    by (rule exI[of _ \<box>], rule exI[of _ C], insert Empty, auto)
next
  case (PCons i p)
  then obtain f bef C' aft where C: "C = More f bef C' aft" by (cases C, auto)
  from PCons(2) have "hole_pos C' = p <#> q" unfolding C by simp
  from PCons(1)[OF this] obtain D E where C': "C' = D \<circ>c E" 
    and p: "hole_pos D = p" and q: "hole_pos E = q" and E: "E = C' |_c p"
    by auto
  show ?case 
    by (rule exI[of _ "More f bef D aft"], rule exI[of _ E], unfold C C',
      insert p[symmetric] q E PCons(2) C, simp)
qed

lemma ctxt_subst_id[simp]: "C \<cdot>c Var = C" by (induct C, auto)

(* the strict subterm relation between contexts and terms *)
inductive_set
  suptc :: "(('f, 'v) ctxt \<times> ('f, 'v) term) set"
where
    arg: "s \<in> set bef \<union> set aft \<Longrightarrow> s \<unrhd> t \<Longrightarrow> (More f bef C aft, t) \<in> suptc"
 |  ctxt: "(C,s) \<in> suptc \<Longrightarrow> (D \<circ>c C, s) \<in> suptc"

hide_const suptcp 

abbreviation suptc_pred where "suptc_pred C t \<equiv> (C, t) \<in> suptc"

notation (xsymbols)
  suptc ("op \<rhd>c") and
  suptc_pred ("(_/ \<rhd>c _)" [56, 56] 55) 

lemma suptc_stable: "C \<rhd>c s \<Longrightarrow> C \<cdot>c \<sigma> \<rhd>c s \<cdot> \<sigma>"
proof (induct rule: suptc.induct)
  case (arg s bef aft t f C)
  let ?s = "\<lambda> t. t \<cdot> \<sigma>"
  let ?m = "map ?s"
  have id: "More f bef C aft \<cdot>c \<sigma> = More f (?m bef) (C \<cdot>c \<sigma>) (?m aft)" by simp
  show ?case unfolding id
    by (rule suptc.arg[OF _ supteq_stable[OF arg(2)]],
    insert arg(1), auto)
next
  case (ctxt C s D)
  have id: "D \<circ>c C \<cdot>c \<sigma> = (D \<cdot>c \<sigma>) \<circ>c (C \<cdot>c \<sigma>)" by simp
  show ?case unfolding id
    by (rule suptc.ctxt[OF ctxt(2)])
qed

lemma suptc_imp_supt: "C \<rhd>c s \<Longrightarrow> C\<langle>t\<rangle> \<rhd> s"
proof (induct rule: suptc.induct)
  case (arg s bef aft u f C)
  let ?C = "(More f bef C aft)"
  from arg(1) have "s \<in> set (args (?C\<langle>t\<rangle>))" by auto
  hence "?C\<langle>t\<rangle> \<rhd> s" by auto
  from supt_supteq_trans[OF this arg(2)]
  show ?case .
next
  case (ctxt C s D)
  have supteq: "(D \<circ>c C)\<langle>t\<rangle> \<unrhd> C\<langle>t\<rangle>" by auto
  from supteq_supt_trans[OF this ctxt(2)]
  show ?case .
qed 

lemma suptc_supteq_trans: "C \<rhd>c s \<Longrightarrow> s \<unrhd> t \<Longrightarrow> C \<rhd>c t"
proof (induct rule: suptc.induct)
  case (arg s bef aft u f C)
  show ?case
    by (rule suptc.arg[OF arg(1) supteq_trans[OF arg(2) arg(3)]])
next
  case (ctxt C s D)
  hence supt: "C \<rhd>c t" by auto
  thus ?case by (rule suptc.ctxt)
qed

lemma supteq_suptc_trans: "C = D \<circ>c E \<Longrightarrow> E \<rhd>c s \<Longrightarrow> C \<rhd>c s"
  by (auto intro: suptc.ctxt)

hide_fact (open)
  suptcp.arg suptcp.cases suptcp.induct suptcp.intros suptc.arg suptc.ctxt 

definition vars_subst :: "('f,'v)subst \<Rightarrow> 'v set"
  where "vars_subst \<sigma> \<equiv> subst_domain \<sigma> \<union> \<Union>{vars_term (\<sigma> x) | x. x \<in> subst_domain \<sigma>}" 

definition subst_compose' :: "('f,'v)subst \<Rightarrow> ('f,'v)subst \<Rightarrow> ('f,'v)subst"
  where "subst_compose' \<sigma> \<tau> \<equiv> \<lambda> x. if (x \<in> subst_domain \<sigma>) then \<sigma> x \<cdot> \<tau> else Var x"

lemma vars_subst_compose': assumes "vars_subst \<tau> \<inter> subst_domain \<sigma> = {}"
  shows "\<sigma> \<circ>s \<tau> = \<tau> \<circ>s (subst_compose' \<sigma> \<tau>)" (is "?l = ?r")
proof
  fix x
  show "?l x = ?r x"
  proof (cases "x \<in> subst_domain \<sigma>")
    case True
    with assms have nmem: "x \<notin> vars_subst \<tau>" by auto
    hence nmem: "x \<notin> subst_domain \<tau>" unfolding vars_subst_def by auto
    hence id: "\<tau> x = Var x" unfolding subst_domain_def by auto
    have "?l x = \<sigma> x \<cdot> \<tau>" unfolding subst_compose_def by simp
    also have  "... = ?r x" unfolding subst_compose'_def subst_compose_def using True unfolding id by simp
    finally show ?thesis .
  next
    case False 
    hence l: "?l x = \<tau> x \<cdot> Var" unfolding subst_domain_def subst_compose_def by auto
    let ?\<sigma>\<tau> = "(\<lambda> x. if x \<in> subst_domain \<sigma> then \<sigma> x \<cdot> \<tau> else Var x)"
    have r: "?r x = \<tau> x \<cdot> ?\<sigma>\<tau>" 
      unfolding subst_compose'_def subst_compose_def ..
    show "?l x = ?r x" unfolding l r
    proof (rule term_subst_eq)
      fix y
      assume y: "y \<in> vars_term (\<tau> x)"
      have "y \<in> vars_subst \<tau> \<or> \<tau> x = Var x"
      proof (cases "x \<in> subst_domain \<tau>")
        case True thus ?thesis using y unfolding vars_subst_def by auto
      next
        case False
        thus ?thesis  unfolding subst_domain_def by auto
      qed
      thus "Var y = ?\<sigma>\<tau> y"
      proof
        assume "y \<in> vars_subst \<tau>"
        with assms have "y \<notin> subst_domain \<sigma>" by auto
        thus ?thesis by simp
      next
        assume "\<tau> x = Var x"
        with y have y: "y = x" by simp
        show ?thesis unfolding y using False by auto
      qed
    qed
  qed
qed

lemma vars_subst_compose'_pow: assumes "vars_subst \<tau> \<inter> subst_domain \<sigma> = {}"
  shows "\<sigma> ^^ n \<circ>s \<tau> = \<tau> \<circ>s (subst_compose' \<sigma> \<tau>) ^^ n" 
proof (induct n)
  case 0 thus ?case by auto
next
  case (Suc n)
  let ?\<sigma>\<tau> = "subst_compose' \<sigma> \<tau>"
  have "\<sigma> ^^ Suc n \<circ>s \<tau> = \<sigma> \<circ>s (\<sigma> ^^ n \<circ>s \<tau>)" by simp
  also have "... = \<sigma> \<circ>s (\<tau> \<circ>s ?\<sigma>\<tau>^^n)" unfolding Suc ..
  also have "... = (\<sigma> \<circ>s \<tau>) \<circ>s ?\<sigma>\<tau> ^^ n" by auto
  also have "... = (\<tau> \<circ>s ?\<sigma>\<tau>) \<circ>s ?\<sigma>\<tau> ^^ n" unfolding vars_subst_compose'[OF assms] ..
  finally show ?case by simp
qed

lemma subst_pow_commute: assumes "\<sigma> \<circ>s \<tau> = \<tau> \<circ>s \<sigma>"
  shows "\<sigma> \<circ>s (\<tau> ^^ n) = \<tau> ^^ n \<circ>s \<sigma>"
proof (induct n)
  case (Suc n)
  have "\<sigma> \<circ>s \<tau> ^^ Suc n = (\<sigma> \<circ>s \<tau>) \<circ>s \<tau> ^^ n" by simp
  also have "... = \<tau> \<circ>s (\<sigma> \<circ>s \<tau> ^^ n)" unfolding assms by simp
  also have "... = \<tau> ^^ Suc n \<circ>s \<sigma>" unfolding Suc by simp
  finally show ?case .
qed simp

lemma subst_power_commute: assumes "\<sigma> \<circ>s \<tau> = \<tau> \<circ>s \<sigma>"
  shows "(\<sigma> ^^ n) \<circ>s (\<tau> ^^ n) = (\<sigma> \<circ>s \<tau>)^^n"
proof (induct n)
  case (Suc n)
  have "(\<sigma> ^^ Suc n) \<circ>s (\<tau> ^^ Suc n) = (\<sigma>^^n \<circ>s (\<sigma> \<circ>s \<tau> ^^ n) \<circ>s \<tau>)" 
    unfolding subst_power_Suc by simp
  also have "... = (\<sigma>^^n \<circ>s \<tau> ^^ n) \<circ>s (\<sigma> \<circ>s \<tau>)" 
    unfolding subst_pow_commute[OF assms] by simp
  also have "... = (\<sigma> \<circ>s \<tau>)^^Suc n" unfolding Suc
    unfolding subst_power_Suc ..
  finally show ?case .
qed simp

lemma vars_term_ctxt_apply: "vars_term (C\<langle>t\<rangle>) = vars_ctxt C \<union> vars_term t"
  by (induct C, auto)

lemma vars_ctxt_pos_term: assumes p: "p \<in> poss t"
  shows "vars_term t = vars_ctxt (ctxt_of_pos_term p t) \<union> vars_term (t |_ p)"
proof -
  let ?C = "ctxt_of_pos_term p t"
  have "t = ?C\<langle>t |_ p\<rangle>" using ctxt_supt_id[OF p] by simp
  hence "vars_term t = vars_term (?C\<langle>t |_ p\<rangle>)" by simp
  thus ?thesis unfolding vars_term_ctxt_apply .
qed

lemma Var_pow_Var[simp]: "Var ^^ n = Var"
  by (rule, induct n, auto)

definition var_renaming :: "('f,'v)subst \<Rightarrow> bool"
  where "var_renaming \<sigma> \<equiv> (\<forall> x. is_Var (\<sigma> x)) \<and> inj_on \<sigma> (subst_domain \<sigma>)"

definition inverse_var_renaming :: "('f,'v)subst \<Rightarrow> ('f,'v)subst"
  where "inverse_var_renaming \<sigma> y \<equiv> if Var y \<in> subst_range \<sigma> then Var (the_inv_into (subst_domain \<sigma>) \<sigma> (Var y)) else Var y"

lemma var_renaming_inverse_domain: 
  assumes ren: "var_renaming \<sigma>"
  and x: "x \<in> subst_domain \<sigma>"
  shows "Var x \<cdot> \<sigma> \<cdot> inverse_var_renaming \<sigma> = Var x" (is "_ \<cdot> ?\<sigma> = _")
proof -
  note ren = ren[unfolded var_renaming_def]
  from ren obtain y where \<sigma>x: "\<sigma> x = Var y" by force
  from ren have inj: "inj_on \<sigma> (subst_domain \<sigma>)" by auto
  note inj = the_inv_into_f_eq[OF inj, OF \<sigma>x]
  note d = inverse_var_renaming_def
  from x have "Var y \<in> subst_range \<sigma>" using \<sigma>x by force
  hence "?\<sigma> y = Var (the_inv_into (subst_domain \<sigma>) \<sigma> (Var y))" unfolding d by simp
  also have "... = Var x" using inj[OF x] by simp
  finally show ?thesis using \<sigma>x by simp
qed

lemma var_renaming_inverse_range: 
  assumes varren: "var_renaming \<sigma>"
  and x: "Var x \<notin> subst_range \<sigma>"
  shows "Var x \<cdot> \<sigma> \<cdot> inverse_var_renaming \<sigma> = Var x" (is "_ \<cdot> ?\<sigma> = _")
proof (cases "x \<in> subst_domain \<sigma>")
  case True
  from var_renaming_inverse_domain[OF varren True] 
  show ?thesis .
next
  case False
  hence \<sigma>x: "\<sigma> x = Var x" unfolding subst_domain_def by auto
  note ren = varren[unfolded var_renaming_def]
  note d = inverse_var_renaming_def
  have "Var x \<cdot> \<sigma> \<cdot> ?\<sigma> = ?\<sigma> x" using \<sigma>x by auto
  also have "... = Var x" 
    unfolding d using x by simp
  finally show ?thesis .
qed

end
